#ifndef __ABSTRACTCODECCLASS_H_INCLUDED_SW__
#define __ABSTRACTCODECCLASS_H_INCLUDED_SW__

/** \file AbstractCodecClass.h
 * \brief Declares abstract codec (compression/decompression) interface class. 
    
	\author Stefan Weisse / DESY
	\date   2008
 
*/

#include "tine.h"

/** \class CAbstractCodecClass
	\brief Declares an abstract codec (compression/decompression) interface. 

	\author Stefan Weisse / DESY
	\date   2008

	This abstract class defines all methods necessary for a proper codec interface which can easily be used 
	inside VSv3 C++ sourcecode.
	
	\remark Any implementation of an image compression/decompression algorithm <b>must</b> be a child of this class.
  */


// one class might have more than one possible input compression formats
// one class must have only one output compression format
class CAbstractCodecClass
{
public:
	
	/**	\brief Destruction of class instance.

		Each interface needs a destructor (does cleanup etc.pp.).
		In principle, the only task of a child destructor is to call Deinitialize() method.
	*/
	virtual ~CAbstractCodecClass(){};

	//
	/**
		\brief Initialization of codec interface (so that one is able to compress and/or decompress).
		
		This method is like a placeholder for codec initialisation. For some codecs there is no kind of
		initialization functionality so this method implementation might stay empty. It should return true in 
		this case.

		Usually some buffers might be allocated and/or precalculation tables are generated.

		After this method returned true, it can be assumed by the user that the instance is ready for 
		immediate call(s) to Compress() or Decompress().

		\retval true in case of proper initialization
		\retval false some error happened, no proper initialization done
	*/
	virtual bool Initialize() = 0;
	
	/**
		\brief Proper deinitialization of codec instance.
		
		This method should implement all necessary cleanup. After it returned success (true), the codec 
		instance is supposed to be 100% uninitialized. This method must be robustly implemented by the 
		subclasser, because it might also be called without a successful Initialize() call before.
		
		\retval true in case of proper Deinitialization
		\retval false some error at Deinitialization
	*/
	virtual bool Deinitialize() = 0;

	// ie. will it work?

	/**
		\brief Is raw input format suitable for compression?
		
		This method checks whether the passed input format is valid(adequate) for this compression 
		technique. This function may work at any time but is guaranteed to work only after a 
		successful call to Initialize() method.
		
		\retval true in case format is adequate for compression
		\retval false format is not adequate and thus, cannot be compressed using this codec
	*/
	virtual bool isAllowedCompressionInputFormat( const DIMAGE &aCheckFormat ) = 0;

	/**
		\brief Is input format suitable for decompression?
		
		This method checks whether the passed input format is valid(adequate) for decompression
		using this codec instance. This function may work at any time but is guaranteed to work 
		only after a successful call to Initialize() method.
		
		\retval true in case format is adequate for decompression
		\retval false format is not adequate and thus, cannot be decompressed using this codec
	*/
	virtual bool isAllowedDecompressionInputFormat( const DIMAGE &aCheckFormat ) = 0;

	// might do without it?
	//bool getCompressionOutputFormatForInputFormat( const DIMAGE &aInputFormat, INT32 &aOutputFormat) = 0;

	/**
		\brief Compresses a raw image.
		
		This method checks whether the passed (raw) input format is valid(adequate) for compression
		using this codec instance. Naturally, this function performs work only after a successful call 
		to Initialize() method. With the optional parameter aUndoIfBiggerThanUncompressed one has the
		possibility to keep compressed data if it has become bigger than raw size (failed compression) or
		restore uncompressed raw bits (do not blow up data).
		
		  \param aImageReference input and output image reference
		  \param aUndoIfBiggerThanUncompressed restore raw image if compression's intention (reduce size)
		         was not fulfilled

		\retval true if image was properly compressed (or undone), referenced image parameter was changed
		\retval false image was not properly compressed, referenced image parameter was not changed
	*/
	virtual bool Compress( DIMAGE& aImageReference, bool aUndoIfBiggerThanUncompressed=true ) = 0;

	/**
		\brief Decompresses a (previously compressed) image.
		
		This method decompresses a DIMAGE in place of its reference. Naturally, this function performs 
		work only after a successful call to Initialize() method.
		
		\param aImageReference input and output image reference
		
		\retval true if image was properly decompressed, referenced image was touched
		\retval false image was not properly decompressed, referenced image was not touched
	*/
	virtual bool Decompress( DIMAGE& aImageReference ) = 0;
};

#endif // __ABSTRACTCODECCLASS_H_INCLUDED_SW__
