#ifndef __INIEX_XML_H_INCLUDED_SW__
#define __INIEX_XML_H_INCLUDED_SW__

/** \file IniEx_XML.h
 * \brief Declaration of XML configuration file handling class.
	\author Stefan Weisse (SW) / DESY  (creation of XML wrapper)
	\author Stefan Weisse (SW) / DESY  (adaption, refactoring, extending)
	\author Koray Gecici (SW downloaded basic implementation from WWW in 2002 - codeguru.com)
	\date   2002-2008
*/

/** \class CIniEx
    \brief Provides XML configuration file handling. 
	\author Stefan Weisse (SW) / DESY  (creation of XML wrapper)
	\author Stefan Weisse (SW) / DESY  (adaption, refactoring, extending)
	\author Koray Gecici (SW downloaded basic implementation from WWW in 2002 - codeguru.com)
	\date   2002-2008

	Originally taken from the WWW, supported no comments and only CString/MFC/win32 api. Later on it 
	was partly rewritten for platform independent projects and comment handling was added ('#'). It 
	implements handling for multiple-purpose configuration files based on section/key/value-schema. 
	It has proven to work fine in a lot of projects made by the adoptor.

	Current implementation is an XML-wrapper (based on TinyXML class) of CIniEx functionality with
	almost the same public interface. Due to the same public interface, error handling is left out 
	almost completely because it was so in the old version. As no errors were observed because of 
	this, no changes were done up to now.

    */


class CIniEx  
{
public:

	//! Construction of yet empty instance.
	//! \param aUnusedParameter now unused, but left in for compatibility reason
	CIniEx(int aUnusedParameter=4);
	
	//! Destruction of class. Writes back xml file if necessary and cleans up memory.
	virtual ~CIniEx();

	/** Initializes the instance properly, opens the configuration file and reads its contents 
		into memory buffers.

		\param pFileName		filename (may include full path) of file to open
		\param writeWhenChange	do write-back to file at any content change
		\param createIfNotExist create a blank file if it does not exist
		\param noCaseSensitive	no effect in XML, checking always <b>case sensitive!</b>
		\param makeBackup		do make a backup file if saving to disk
		\retval	true success 
		\retval false failure (mostly file not found)

		\remark As XML files are always case-sensitive, the parameter noCaseSensitive is ignored and not
		taken into account when getting values! <b>Comparing/Matching is always case sensitive!</b>
	*/
	bool Open(const char *pFileName,
			  bool writeWhenChange=true,
			  bool createIfNotExist=true,
			  bool noCaseSensitive=true,
			  bool makeBackup=false);

	/** Initializes the instance properly, opens the configuration file at executable directory
		and reads its contents into memory buffers. To avoid redundancy, after querying the 
		executable path, the method	calls overloaded Open() method.

		\param pFileName		filename (must not include full path) of file to open
		\param writeWhenChange	do write-back to file at any content change
		\param createIfNotExist create a blank file if it does not exist
		\param noCaseSensitive	no effect in XML, checking always <b>case sensitive!</b>
		\param makeBackup		do make a backup file if saving to disk
		\retval	true success
		\retval false failure (mostly file not found)

		\remark As XML files are always case-sensitive, the parameter noCaseSensitive is ignored and not
		taken into account when getting values! <b>Comparing/Matching is always case sensitive!</b>
	*/
	bool OpenAtExeDirectory(const char *pFileName,
			  bool writeWhenChange=true,
			  bool createIfNotExist=true,
			  bool noCaseSensitive=true,
			  bool makeBackup=false);

	/** Resets the configuration file memory buffer contents to empty (no sections, no keys, no values).
	*/
	void ResetContent();

	/** Writes back the contents of the memory structure to initialization file on disk. There is nearly
		no error checking inside the method!
	
		\param makeBackup		true: do make a backup file, false: don't
		\return	backup file name or "" if no backup file was created
	*/
	char *WriteFile(bool makeBackup=false);


	/** Simple GetValue method for obtaining the value of the first key identified by match to 
		input string buffer 'Key'. This method is a simple overload of GetValue(section,key,defaultvalue).

		\param Section section-id to search for
		\param Key key to search for (must be found inside dedicated section)
		\param DefaultValue default string value to return in case no match to section+key was found
		
		\return Value that was found as static string pointer reference(!).
		
		\remark As this method returns static string pointer reference, any subsequent call to this (set 
		of) overloaded method(s) will alter the return value!
	*/
	char *GetValue(const char *Section,const char *Key,const char *DefaultValue="");


	/** Overloaded GetValue for directly obtaining integer value.

		\param Section see "See also:"
		\param Key see "See also:"
		\param defval default integer value to be returned in case section and key could not be found
		\return integer value obtained or default integer value on error (not found or conversion error)

		\sa GetValue(const char *Section, const char *Key, const char *DefaultValue="");
	*/
	int     GetValueInt(const char *Section,const char *Key,int defval=0);

	/** Overloaded GetValue for directly obtaining double value.

		\param Section see "See also:"
		\param Key see "See also:"
		\param defval default double value to be returned in case section and key could not be found
		\return double value obtained or default double value on error (not found or conversion error)
		
		\sa GetValue(const char *Section, const char *Key, const char *DefaultValue="");
	*/
	double  GetValueDouble(const char *Section,const char *Key,double defval=0.0);


	/** Method for directly setting or updating a value on a key inside a dedicated section.

		\param Section section-name to search for
		\param Key key-name to search for (must be found inside dedicated section found before)
		\param Value value to store or add at section-key position
	*/
	void SetValue(const char *Section,const char *Key,const char *Value);
	


	/** Overloaded SetValue for directly setting integer value.

		\param Section see "See also:"
		\param Key see "See also:"
		\param value integer value to be stored to memory-mapped file contents

		\sa SetValue(const char *Section, const char *Key, const char *Value)

	*/
	void SetValueInt(const char *Section, const char *Key, int value );
    
	/** Overloaded SetValue for directly setting double value.

		\param Section see "See also:"
		\param Key see "See also:"
		\param value double value to be stored to memory-mapped file contents

		\sa SetValue(const char *Section, const char *Key, const char *Value)
	*/
	void SetValueDouble(const char *Section, const char *Key, double value );

	/** Gets the current flag whether to write-back to file on disk on any change 
		of memory-mapped contents.

		\return true: write on any change, false: <b>only</b> write on demand (user programmer 
		has to call WriteFile() by hand)
	*/
	bool GetWriteWhenChange();

	/** Sets the current flag whether to write-back to file on disk on any change 
		of memory-mapped contents. No automatic writeback is done in case of true parameter.

		\param WriteWhenChange true: write on any change, false: <b>only</b>write on demand (user programmer 
		has to call WriteFile() by hand)
	*/
	void SetWriteWhenChange(bool WriteWhenChange);


	/** Returns the internal TinyXML TiXmlDocument class instance as pointer reference. Be careful
	    and never delete this pointer!

		\return pointer reference to TiXMLDocument
	*/
	TiXmlDocument *GetInternalXMLDocumentHandle();

	//! top-level element in XML configuration files, necessary for detection of proper XML file
	static  const char WrapElement[];

private:

	/** Simple GetValue method for obtaining the value of the first key identified by match to 
		input string buffer 'Key'. 

		\param Key key to search for
		\return value that was found as string pointer reference(!)
		\sa GetValue(const char *Section, const char *Key, const char *DefaultValue="");
	*/
	char *GetValue(const char *Key);

	/** Simple SetValue method for directly setting or updating a value on a key.
		This method is a simple overload of SetValue(section,key,tvalue) where an
		empty section is passed.

		\param Key see "See also:"
		\param Value see "See also:"

		\sa SetValue(const char *Section, const char *Key, const char *Value)
	*/
	void SetValue(const char *Key,const char *Value);

	/* Removes a named 'Key' (and thus, its value, if any) out of a named Section inside the 
		memory-mapped file buffer.

		\param Section named section where to look into for named 'Key'
		\param Key named key to delete if found
		\return true in case of successful removal, false otherwise
	*/
	//bool RemoveKey(const char *Section,const char *Key);

	/* Overloaded RemoveKey method to remove a key from the only (empty="") section.

		\param Key see "See also:"
		\return see "See also:"

		\sa RemoveKey(const char *Section, const char *Key)
	*/
	//bool RemoveKey(const char *Key);

	/* Removes whole named 'Section' (and thus, its keys and values, if any) out of the 
		memory-mapped file buffer.

		\param Section named section to delete if found
		\return true in case of successful removal, false otherwise
	*/
	//bool RemoveSection(const char *Section);

	/*	Sets/Changes the currently used backup file name.

		\param backupFile	new backup file name
	*/
	//void SetBackupFileName(const char *backupFile);

	
	//! true: valid structure, false: invalid substructure
	bool	ValidateWrappedIniExInXML();

	//Variables
	
	//! internal XML document class (of tinyxml library)
	TiXmlDocument m_Doc;

		//! path to executable file of application
	char				m_ExePath[MAX_PATH_LENGTH];
	
	
	// true: matching of configuration file contents vs. lookup functions is case insensitive, false: case-sensitive
	//bool	m_NoCaseSensitive;
	
	//! true: perform write-back to file on disk on any change, false: only write on demand
	bool	m_writeWhenChange;

	//! current backup filename (plus full path maybe)
	char	m_BackupFileName[MAX_STRING_LENGTH];
	
	//! current initialization file name (plus full path maybe)
	char    m_FileName[MAX_STRING_LENGTH];
	
	//! true: do make a backup copy before saving, false: no backup file should be created
	bool	m_makeBackup;
	
	//! true: memory duplicates of file were changed in between, false: contents not changed
	bool	m_Changed;	


	//!		finds a suitable backup file name, rudimentary implementation
	void	FindBackupFile(void);
};

#endif // __INIEX_XML_H_INCLUDED_SW__
