#ifndef __SGP_BASE_INCLUDED_SW__
#define __SGP_BASE_INCLUDED_SW__

/** \file SGP_Base.h
 * \brief Declares abstract grabbing interface class. 
    
	\author Stefan Weisse / DESY
	\date   2008
 
*/


/** \class CAbstractGrabberInterface
	\brief Declares an abstract grabbing interface. 

	\author Stefan Weisse / DESY
	\date   2008

	In order to support the same set of portable sourcecode with different (also: non-portable) grabbing interfaces, 
	this abstract class defines all functions necessary for proper interfacing inside tine_server.cpp. In basic 
	terms, one just has to create a child class of CAbstractGrabberInterface, fill it with logic and compile and 
	link the whole stuff. The result should be a working "Small Grabber Part" TINE server without the need of 
	changing any additional code.

	\remark Any implementation of an image source ('grabbing') interface <b>must</b> be a child of this class.
  */

// FltGain and FltShutter

#define ABSTRACT_NUM_GENERIC_SLOW_CONTROL_PROPERTIES		2
#define ABSTRACT_NUM_GENERIC_HARDWARE_CONTROL_PROPERTIES	ABSTRACT_NUM_GENERIC_SLOW_CONTROL_PROPERTIES*9


class CAbstractGrabberInterface
{
public:
	/**	\brief Destruction of hardware interface.

		Each grabber interface needs a destructor (does cleanup etc.pp.).
		In principle, the only task of a child destructor is to call Deinitialize() method.
	*/
	virtual ~CAbstractGrabberInterface(){};

	/**
		\brief Initialization and start-up of hardware interface.
		
		This method sets up the grabbing subsystem and should start acquiring frames from image source 
		immediately if it returns true. On false return, one should check GetLastError() method to find out
		what happened.
		
		\param aCfg configuration file that contains hardware-dependent settings
		\param notifyFunction notify function for each new successful acquired frame
		\param aBuffers pointer to aNumBuffers source image buffers (flat 1-dim memory space, one after another)
		\param aNumBuffers number of source image buffers
		\retval true in case of proper initialization
		\retval false some error happened, no proper initialization done
	*/
	virtual bool Initialize( CIniEx *aCfg, void(*notifyFunction)(UINT32 aBufferNumber, UINT32 aBufferLength, DOUBLE aTineFrameTimestamp), void *aBuffers, UINT32 aNumBuffers ) = 0;

	/**
		\brief Proper deinitialization of hardware interface.
		
		This method should implement all necessary cleanup. After it returned success (true), the grabbing 
		interface is supposed to be 100% uninitialized.
		
		\retval true in case of proper Deinitialization
		\retval false some error at Deinitialization
	*/
	virtual bool Deinitialize() = 0;

	/**
		\brief Obtain last error.

		This method should return the last error that happened inside the hardware class.
	
		\return last error code 
		\remark For developers: pass your last error code here!
	*/
	virtual INT32 GetLastError() = 0;
	
	/**
		\brief Obtain last error as human-readable string representation.

		This method should return the last error that happened inside the hardware class
		as descriptive string for at least medium experts.

		\return internal const string buffer with information regarding last error code 
	*/
	
	virtual const char *GetLastErrorAsString()=0; // use built-in buffer

	/**
		\brief Validates hardware section of configuration file.
	  
		Before calling Initialize(), the hardware configuration file will be checked for physical validity. 
		This means that required keys must exist and contain a meaningful value. Relations between keys and
		values which makes no sense from logical point of view should be handled inside Initialize() method.
		Any code which checks physical validity must be put in here.
		
		\param aCfg reference to the configuration file instance
		\retval true contents of the file are valid
		\retval false validation failed!
	*/
	virtual bool ValidateHardwareConfiguration(CIniEx &aCfg) = 0;

	/**	
		\brief Returns current frame rate.
	
		To be more specific, it returns the current calculated average framerate in frames per second [Hz].
		\return frames per second 
	*/
	virtual FLOAT getFramerate()=0;

	/**
		\brief Indicates whether hardware control is provided.

		This method's return value indicates whether hardware control (e.g. gain, shutter) is 
		provided by the image source implementation.
	
		\return true hardware control is provided
		\retval false no hardware control is provided

		\sa EquipmentFunctionHardwareDeviceTineControl()
	*/
	virtual bool IsHardwareControlProvided() = 0;
	
	virtual int GetNumHardwareControlProperties() = 0;

	// left out, not needed, comments left in as memorial ;-)
	//virtual bool RegisterHardwareDeviceTineControl() = 0;
	//virtual bool UnregisterHardwareDeviceTineControl() = 0;

	/**	\brief TINE Equipment Function part for hardware control.
		
		In case hardware control is provided, queries for hardware control properties will be downcalled to
		this sub-equipment function which must deal with everything regarding hardware control. The parameters
		are all like any general Equipment Function for TINE protocol.
		
		\param devName device name that was queried
		\param devProperty property name that was queried
		\param dout output value(s) in order to be filled and sent to client
		\param din input value(s) delivered by client
		\param access access mode delivered by client request (usual CA_READ and/or CA_WRITE)
		
		\retval 0 success
		\retval 1-512 TINE error code
		\retval >512 TINE user error code 

		\sa IsHardwareControlProvided()
	*/
	virtual int EquipmentFunctionHardwareDeviceTineControl(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access) = 0;

	virtual bool RegisterAndPrepareHardwareXQueryStruct( SWXPropertyQueryStruct *hardwareStruct, XPropertyQueryStruct *aXPQS) = 0;

	/** \brief Indicates temporal unavailability of hardware.

		Some grabbing APIs allow that cameras temporarily vanish from the API. This is true for Prosilica 
		GigE API. By examining the return code of this function from inside TINE server, one is able to get 
		to know this state.

		\retval true image source hardware temporarily unavailable
		\retval false image source hardware is up and running
	*/
	virtual bool IsImageSourceHardwareTemporarilyUnavailable() = 0;

	/** \brief Provides ability to perform asynchronous events.

		For asynchronous background events, one can enter the logic inside this method.
		This method will be called in any TINE background function at an average rate of about
		TINE_BACKGROUNDFUNCTION_EXECUTERATE_MS defined in tine_server.cpp.
	 */
	virtual void AsyncronousBackgroundEvents() = 0;

	// if no hardware control is provided, these functions should return false in any case and
	// don't touch the parameters

	// gain as a.u. as float value
	virtual bool ReadbackGain( float &aValue ) = 0;
	virtual bool SetGain( float aValue ) = 0;
	virtual bool GetGainSetpoint( float &aValue ) = 0;
	virtual bool GetGainMinMaxAsFloat( float &aMin, float &aMax) = 0;

	// shutter as us as float value
	virtual bool ReadbackExposureTime( float &aValue ) = 0;
	virtual bool SetExposureTime( float aValue ) = 0;
	virtual bool GetExposureTimeSetpoint( float &aValue ) = 0;
	virtual bool GetExposureTimeMinMaxAsFloat( float &aMin, float &aMax) = 0;

};

#endif // __SGP_BASE_INCLUDED_SW__
