#ifndef __GLOBALHEADERS_H_INCLUDED_SW__
#define __GLOBALHEADERS_H_INCLUDED_SW__

#include "StdAfx.h"

/** \file globalheaders.h
    \author Stefan Weisse / DESY 
	\date   2008

	\brief Declares global structures and definitions for VSv3.
	\sa platform.h
*/

//! maximum frames per second for framerate calculation
const INT32 SGP_MAX_FPS			= 100;

//! measure time of FPS (seconds, integer)
const INT32 FPS_MEASURE_TIME	= 5;

//! must be exactly 188 bytes
const INT32 DIMAGE_HEADER_SIZE		= sizeof(hdrIMAGE);
const INT32 DIMAGE_HEADER_SIZE_COUNTED_BY_HUMAN = 188;


/** \brief maximum transport size for TINE in bytes
	
	  This constant defines the maximum transport size for TINE. One whole uncompressed image plus 
	  header must fit inside. As it can be seen, currently it is set to exact 6 Megabytes so it is 
	  sufficent for e.g. ~1.5 Megapixel x 4 bytes per pixel (RGBA).
	  The Header is NOT part of the max transport size!
 */

const INT32 MAX_TRANSPORT_SIZE	= 1024*1024*6;

/** \brief maximum buffer size for TINE image + FULL header in bytes
	
	  This constant defines the maximum transport size for TINE. One whole uncompressed image plus 
	  header must fit inside. As it can be seen, currently it is set to 6 Megabytes plus sizeof(DIMAGE)
	  structure so it is sufficent for e.g. ~1.5 Megapixel x 4 bytes per pixel (RGBA) plus Header.
	  This constant is used to allocate and pass buffers inside of software! It is not used on the 
	  network (TINE) where the header size is not taken into account!
*/
const INT32 MAX_BUFFER_SIZE = MAX_TRANSPORT_SIZE+sizeof(DIMAGE);

const UINT32 CF_IMAGE_INITIAL_VERSION	= 0x1;
const UINT32 CF_IMAGE_VERSION			= CF_IMAGE_INITIAL_VERSION;

const UINT32 CF_IMAGE_NO_CAMERA_PORT_ID = 0;

//! 'vsv3'
const UINT32 CF_IMAGE_FORMAT_BASETAG = 0x33565356; 

//! image format: grayscale data (single plane)
const INT32 CF_IMAGE_FORMAT_GRAY		= 0;
//! image format: rgb data (three planes interwoven as 24 bits)
const INT32 CF_IMAGE_FORMAT_RGB			= 1;
//! image format: rgb data (three planes plus alpha interwoven as 32 bits)
const INT32 CF_IMAGE_FORMAT_RGBA		= 2;
//! image format: rgb data (three planes interwoven as 15 bits)
const INT32 CF_IMAGE_FORMAT_RGB15		= 3;
//! image format: rgb data (three planes interwoven as 16 bits; green=6)
const INT32 CF_IMAGE_FORMAT_RGB16		= 4;
//! image format: jpeg data bits follow
const INT32 CF_IMAGE_FORMAT_JPEG		= 5;
//! image format: tiff data bits follow
const INT32 CF_IMAGE_FORMAT_TIFF		= 6;
//! image format: YUV411 encoded raw bits 
const INT32 CF_IMAGE_FORMAT_YUV411		= 7;
//! image format: YUV422 encoded raw bits 
const INT32 CF_IMAGE_FORMAT_YUV422		= 8;
//! image format: YUV444 encoded raw bits 
const INT32 CF_IMAGE_FORMAT_YUV444		= 9;
//! image format: HUFFYUV-compressed image bits follow
const INT32 CF_IMAGE_FORMAT_HUFFYUV		= 10;
//! image format: Windows BMP (BitMaP) bits follow
const INT32 CF_IMAGE_FORMAT_BMP			= 11;
//! image format: Bayer-packed image bits 
const INT32 CF_IMAGE_FORMAT_BAYER		= 12;

const INT32 CF_IMAGE_FORMAT_MAX		= CF_IMAGE_FORMAT_BAYER;

//! image flag: bits are in little endian byte order, if not set, it is big endian order
const UINT32 CF_IMAGE_FLAG_LITTLE_ENDIAN_BYTE_ORDER	= 0x1;
//! image flag: image bits have not yet undergone lossy transformation
const UINT32 CF_IMAGE_FLAG_IMAGE_LOSSLESS			= 0x2;
//! image flag: image bits are flipped around the horizontal axis by 180 degress on the way
const UINT32 CF_IMAGE_FLAG_HORIZONTAL_FLIP_180		= 0x4;
//! image flag: image bits are flipped around the vertical axis by 180 degress on the way
const UINT32 CF_IMAGE_FLAG_VERTICAL_FLIP_180		= 0x8;

/** \struct SWXPropertyQueryStruct

	\brief Property details definition structure type

	In order to support different device names with different properties each at 
	TINE server level, a special, dedicated XPropertyQueryStruct needs to be defined. 
	As directly defining this structure would result in a lot of redundancies and would
	interfere with clean source code paradigm, a special structure was invented to only
	define the required members for purposes in SGP projects.

	All fields are mandatory.

*/
typedef struct
{
	//! property name (allowed up to PROPERTY_NAME_SHORTSIZE chars (Tine R4: 32)
	char	prpName[PROPERTY_NAME_SHORTSIZE];
 
	//! property description (not 100% in)formal field (up to 64 chars)
	char	prpDescription[64];
  
	//! output tagged structure data type: up to TAG_NAME_SHORTSIZE chars (Tine R4: 8)
	char	prpTag[TAG_NAME_SHORTSIZE];
  
	//! input tagged structure data type: up to TAG_NAME_SHORTSIZE chars (Tine R4: 8)
	char	prpTagIn[TAG_NAME_SHORTSIZE];
  
	//! (in?)formal unit field in human-readable format as string, if any
	char	prpUnits[16];
  
	//! smalles property value
	float	prpMinValue;
  
	//! biggest property value
	float	prpMaxValue;
  
	//! size in elements of prpFormat for output data
	UINT32	prpSize;
  
	//! size in elements of prpInFormat for input data
	UINT32	prpSizeIn;
  
	//! data type format for output
	short	prpFormat;
  
	//! data type format for input
	short	prpFormatIn;
  
	//! allowed acces mode(s) for the property
	BYTE	prpAccess;

	//! unique id of that property
	UINT16	property_id;

} SWXPropertyQueryStruct;

/** \def PRP_BASEID_HARDWARECONTROL
 first property id for hardware control
*/
#define PRP_BASEID_HARDWARECONTROL 100

/** \def RETCODE_HWC_READBACKFAILED_DEFINE 
returncode of tine-server in case a readback of hardware control property failed
*/
#define RETCODE_HWC_READBACKFAILED_DEFINE			518

/** \def RETCODE_HWC_PUTTOHARDWAREFAILED_DEFINE 
returncode of tine-server in case a set-to-hardware command failed
*/
#define RETCODE_HWC_PUTTOHARDWAREFAILED_DEFINE		519

#define RETCODE_HWC_PARAMETERSHADOWED_DEFINE		520

#define RETCODE_HWC_SLOWCONTROLPROPERTY_IS_READONLY_DEFINE	521

/** \def MAKE_COMPILER_HAPPY(a) 
remove warning: unused argument for compiler
*/
#define MAKE_COMPILER_HAPPY(a) {a=a;};

enum ePropertyType {
	PT_Readback,
	PT_SetAndReadback
};

enum eDataType {
//	DT_Determine,
	DT_Integer,
	DT_Float,
	DT_String,
	DT_Enumeration
};

enum eSetType {
	ST_UserAndExpert,
	ST_Expert
};

enum eVisibilityType {
	VT_UserAndExpert,
	VT_Expert
};


typedef struct
{
	char sBaseName[16];
	char sCamApiPropertyName[64];
	eDataType dataType;
	ePropertyType propertyType; // readback or set_and_readback
	eVisibilityType visibilityType;
	eSetType setType;
	char sMinOverride[16];
	char sMaxOverride[16];
	char sUnit[16];
	char sDescription[44];
	// 44 instead of 64 because
	// for each subtype a special string could be prepended:
	//
	// 1234567890123456789
	// Setpoint of 
	// Readback of
	// Increment of 
	// Enum values of 
	// not implemented
	// Visibility type of 
	// Set type of

	// for Integer and Float type: to be filled in later
	// will be copied directly to XPQS from SWXPQS
	float prpMinValue;
	float prpMaxValue;
} GSC_PROPERTY;

#endif // __GLOBALHEADERS_H_INCLUDED_SW__
