#include "StdAfx.h"

/** \file platform.cpp
	\brief Implements platform-dependent functionality.

	All platform-dependent wrapper functions and/or structures are defined in this file.

	\sa platform.h
    \author Stefan Weisse / DESY 
	\date   2008
*/


void platformSleepMilliseconds(UINT32 msec)
{
#if defined(PLF_SYSTEM_LINUX)

	struct timespec req={0};
		
	req.tv_sec=(int)(msec/1000);
	req.tv_nsec=(msec%1000)*1000000L;
		
	while(nanosleep(&req,&req)==-1)
	{
	}
#endif
	
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	::Sleep( msec );
#endif 
}


bool  platformGetTime( PLF_TIME &aTimeStruct )
{
#ifdef PLF_SYSTEM_WIN2005
	__timeb32 tmpTime;

	::_ftime32( &tmpTime );

	aTimeStruct.time = tmpTime.time;
	aTimeStruct.millitm = (UINT32) tmpTime.millitm;
#endif
	
#ifdef PLF_SYSTEM_WINNT
	_timeb tmpTime;

	::_ftime( &tmpTime);

	aTimeStruct.time = tmpTime.time;
	aTimeStruct.millitm = (UINT32) tmpTime.millitm;
#endif

#ifdef PLF_SYSTEM_LINUX
	struct timeb tmpTime; 
	  
	ftime( &tmpTime);

	aTimeStruct.time = tmpTime.time;
	aTimeStruct.millitm = (UINT32) tmpTime.millitm;
#endif

	return true;
}


bool platformGetFullExecutablePath( char *aPathBuffer, UINT32 aMaxBufferLength)
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	if (GetModuleFileName(NULL, aPathBuffer, aMaxBufferLength) == 0) return false;
		
	return true;
#endif
	
#if defined(PLF_SYSTEM_LINUX)
	if (readlink( "/proc/self/exe", aPathBuffer, aMaxBufferLength ) == -1)
	{
		fprintf( stderr, "(WW) readlink() /proc/self/exe failed: %s\n", strerror( errno ) );
		return false;
	}
#endif
		
	return true;
}


bool 	platformCopyFile( const char *aSourceFilename, const char *aCloneFilenameToCreate, bool aFailIfExist)
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)

	BOOL inBool = TRUE;
	if (aFailIfExist == false) inBool = FALSE;
	
	BOOL ret = CopyFile(aSourceFilename, aCloneFilenameToCreate, inBool);
		
	if (ret == TRUE) return true;
		
	return false;

#endif


#if defined(PLF_SYSTEM_LINUX)
	// 16k static for copying
	const long buffersize = 16384;
	static char buffer[buffersize];
		
	// check whether clone file already exist
	// simple, not good error prone
	if (aFailIfExist == true)
	{
		FILE *fp = fopen(aCloneFilenameToCreate, "r");
		if (fp != NULL)
		{
			fclose(fp);
			return false;
		}
	}
		
	FILE *fpIn = fopen(aSourceFilename, "rb");
	if (fpIn == NULL)
	{
		// input file could not be opened
		return false;
	}
		
	// get size of input file
	fseek( fpIn, 0, SEEK_END);
	long inputfilesize = ftell(fpIn);
	fseek( fpIn, 0, SEEK_SET);

	FILE *fpOut = fopen(aCloneFilenameToCreate, "wb");
	if (fpOut == NULL)
	{
		// output file could not be opened for writing...
		fclose(fpIn);
		return false;
	}

	long stilltocopy = inputfilesize;
	while( stilltocopy > 0 )
	{
		long portion = 0;
		if (stilltocopy>buffersize) portion=buffersize; else portion = stilltocopy;

		if (fread(buffer, portion, 1, fpIn) != 1) break;
		if (fwrite(buffer, portion, 1, fpOut) != 1) break;

		stilltocopy -= portion;
	}

	// not necessary, but for clarity and safety
	fseek( fpOut, 0, SEEK_END);
	long outputfilesize = ftell(fpOut);
			
	fclose(fpOut);
	fclose(fpIn);
		
	if (outputfilesize!=inputfilesize) return false;
		
	return true;
#endif
}

#if defined(PLF_USING_PTHREAD)
	typedef struct 
	{
		pthread_mutex_t	mutex;
		pthread_cond_t		condition;
		int			semCount;	
		int 		semMaxCount;
	}
	sem_private_struct, *sem_private;
#endif

HANDLE platformCreateSemaphore( UINT16 initial_count, UINT16 maximum_count )
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	HANDLE sem;
  
  	sem = CreateSemaphore(NULL, initial_count, maximum_count, NULL);
  	if (sem == NULL) return( NULL );
  
  	return sem;
#endif

#if defined(PLF_USING_PTHREAD)
	HANDLE sem;
	int rc;
	
	sem_private token;
	
	assert( initial_count <= maximum_count );
	
	token = (sem_private) malloc(sizeof(sem_private_struct));

	if ((rc = pthread_mutex_init(&(token->mutex), NULL)))
	{
		free(token);
		
		return NULL; //RC_OBJECT_NOT_CREATED;
	}

	if ((rc = pthread_cond_init(&(token->condition), NULL)))
	{
		pthread_mutex_destroy( &(token->mutex) );
		free(token);

		return NULL; //RC_OBJECT_NOT_CREATED;
	}

	token->semCount = initial_count;
	token->semMaxCount = maximum_count;

	sem = token;

	return sem;
#endif
}
  
bool platformDeleteSemaphore( HANDLE semaph )
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	if (semaph == NULL) return false;
  
	if (CloseHandle( semaph ) == TRUE) return true;
  
	return false;
#endif

#if defined(PLF_USING_PTHREAD)
	sem_private token = (sem_private) semaph;
	
	if (token == NULL) return false;
	
	pthread_mutex_destroy(&(token->mutex));

	pthread_cond_destroy(&(token->condition));

	free (token);
		
	return true;
#endif
}
  

bool platformWaitForSemaphore( HANDLE semaph, UINT32 timeout_msec )
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	assert( semaph != NULL);
  
  	if (timeout_msec == 0) timeout_msec = INFINITE;
  
  	DWORD ret = WaitForSingleObject(semaph, timeout_msec); 
  	if (ret == WAIT_OBJECT_0) return true;
  
  	return false;
#endif

#if defined(PLF_USING_PTHREAD)	
	int rc;
	struct timespec tm;
	struct timeb tp;
	long sec, millisec;
	
	sem_private token = (sem_private) semaph;

	if ((rc = pthread_mutex_lock(&(token->mutex)))) return false; //RC_SEM_WAIT_ERROR;

	if (timeout_msec > 0)
	{
		sec = timeout_msec / 1000;
		millisec = timeout_msec % 1000;
		ftime( &tp );
		tp.time += sec;
		tp.millitm += millisec;
	
		if( tp.millitm > 999 )
		{
			tp.millitm -= 1000;
			tp.time++;
		}
	
		tm.tv_sec = tp.time;
		tm.tv_nsec = tp.millitm * 1000000 ;
	}

	while (token->semCount <= 0)
	{
		if (timeout_msec == 0)
		{
			rc = pthread_cond_wait(&(token->condition), &(token->mutex));
		}
		else
		{
			rc = pthread_cond_timedwait(&(token->condition), &(token->mutex), &tm);
		}
		if (rc && (errno != EINTR) )
		break;
	}

	if (rc)
	{
		if ( pthread_mutex_unlock(&(token->mutex)) ) return false; //RC_SEM_WAIT_ERROR );

		if ( rc == ETIMEDOUT ) 
		{
			//printf("Sema timed out.\n");
			/* we have a time out */ return false; //RC_TIMEOUT );
		}

		return false; //RC_SEM_WAIT_ERROR );
	}

	token->semCount--;

	if ((rc = pthread_mutex_unlock(&(token->mutex)))) return false; //RC_SEM_WAIT_ERROR;

	return true;
#endif  
}
  
  
bool platformIsSemaphoreSignalled( HANDLE semaph )
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
  	assert( semaph != NULL);
  
  	DWORD ret = WaitForSingleObject(semaph, 0); // 0 == immediate return
  	if (ret == WAIT_OBJECT_0) return true;
  
  	return false;
#endif

#if defined(PLF_USING_PTHREAD)	
	int rc;
	bool retval = true;
	
	assert( semaph != NULL );
	
	sem_private token = (sem_private) semaph;

	if ((rc = pthread_mutex_lock(&(token->mutex)))) return false; //RC_SEM_WAIT_ERROR;

	// async under posix is (by lack of knowledge) implemented as 1 msec wait and return

	if (token->semCount <= 0) retval = false;
	if (retval == true) token->semCount--;

	if ((rc = pthread_mutex_unlock(&(token->mutex)))) return false; //RC_SEM_WAIT_ERROR;

	return retval;
#endif
}
  
  
bool platformReleaseSemaphore( HANDLE semaphore )
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	assert( semaphore != NULL );
  
  	BOOL ret = ReleaseSemaphore( semaphore, 1, NULL); 
  
  	if (ret == TRUE) return true;
  
  	return false;
#endif

#if defined(PLF_USING_PTHREAD)	
	bool retval = true;
	int rc;
	
	sem_private token = (sem_private) semaphore;
	
	if (token == NULL) return false;
	
	if ((rc = pthread_mutex_lock(&(token->mutex))))	return false;

	if (token->semCount < token->semMaxCount)
	{
		token->semCount ++;
	}
	else retval = false;
	
	if ((rc = pthread_mutex_unlock(&(token->mutex))))	return false;

	if (retval == true) if ((rc = pthread_cond_signal(&(token->condition)))) return false;	
		
	return retval;
#endif
}

#if defined(PLF_USING_PTHREAD)
	typedef struct 
	{
		pthread_t             threadId1;
		pthread_attr_t        pthread_attr;
	} 
	thread_private_struct, *thread_private;
#endif

HANDLE	platformCreateThread( PLF_SW_THREAD_START_ROUTINE aThreadFunction, void *aThreadParameter)
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	ULONG hThreadID;
	  
	return CreateThread(NULL, 0, aThreadFunction, aThreadParameter, 0, &hThreadID);
#endif

#if defined(PLF_USING_PTHREAD)	
	HANDLE plf_thread;
	thread_private hThread;
	int rc;
	int plf_prio;
	
	plf_prio = 0;
	
	hThread = (thread_private) malloc(sizeof(thread_private_struct));
	if (hThread == NULL) return NULL;
	
    if ( (rc = pthread_attr_init(&(hThread->pthread_attr))) )
    {
           fprintf( stderr, "Platform: pthread_attr_init ERROR.\n" );
           free( hThread );
           return NULL;
    }

    if ( (rc = pthread_create(&(hThread->threadId1), &(hThread->pthread_attr), aThreadFunction, aThreadParameter )) )
		{
			fprintf( stderr, "Platform: pthread_create ERROR.\n" );
			pthread_attr_destroy(&(hThread->pthread_attr));
			free( hThread );
			
			return NULL;
   	}
	
	plf_thread = hThread;
	return plf_thread;	
#endif
}


bool		platformDeleteThread(HANDLE aThreadHandle)
{
#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	if (CloseHandle( aThreadHandle ) == FALSE) return false;
  
	return true;
#endif

#if defined(PLF_USING_PTHREAD)	
	if (aThreadHandle == NULL) return false;
	
	thread_private hThread = (thread_private) aThreadHandle;
	
	pthread_attr_destroy(&(hThread->pthread_attr));
	
	free( hThread );

	return true;
#endif
}




	#if defined(PLF_SYSTEM_LINUX)
		// constants required by source code on unix and lwip
		const	INT32	SOCKET_ERROR	= -1;
		const	INT32	INVALID_SOCKET	= -1; 
	#endif

	const INT32 platformMaxFdSets = PLF_MAX_FDSETS;

	// create a human-readable IP-adress out of an host-encoded ip address

	char *platformIPAddrConvert( UINT32 ip_hostorder )
	{
		static char buffer[16];
		UINT32 networkorder = htonl(ip_hostorder);
		unsigned char *b = (unsigned char *) &networkorder;
	
		sprintf(buffer, "%d.%d.%d.%d", b[0], b[1], b[2], b[3]);
	
		return (char *) buffer;
	}

	// four decimals separated by three dots
	// other forms are not supported!
	// returns INADDR_NONE in error case (syntactical, not logical!)
	UINT32 platformIPAddrConvert( const char * const ip_addr_dotted_string)
	{
		// returned ip address will be already in network byte order
	
		INT32 ip1, ip2, ip3, ip4;
		UINT32 ip_addr;
	
		int ret = sscanf((const char *) ip_addr_dotted_string, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4); 
		ret = 1; // make compiler happy
	
		if (ip1 < 0 || ip1 > 255) return INADDR_NONE;
		if (ip2 < 0 || ip2 > 255) return INADDR_NONE;
		if (ip2 < 0 || ip2 > 255) return INADDR_NONE;
		if (ip2 < 0 || ip2 > 255) return INADDR_NONE;

		#ifndef PLF_ENDIANCHANGE
			ip_addr = (ip1&0xff) + ((ip2&0xff)<<8) + ((ip3&0xff)<<16) + ((((UINT32)ip4)&0xff)<<24); 
		#else
			ip_addr = (ip4&0xff) + ((ip3&0xff)<<8) + ((ip2&0xff)<<16) + ((((UINT32)ip1)&0xff)<<24);
		#endif
	
		return ip_addr;
	}

	void		platformInitializeSocketSubsystem()
	{
		#ifdef PLF_WINNT_TCPIPSTACK
			WSADATA wsad;

			int ret = WSAStartup( 0x0202, &wsad);
			if (ret != 0) assert( 1==2 );
		#endif
	}

	void		platformDeinitializeSocketSubsystem()
	{
		#ifdef PLF_WINNT_TCPIPSTACK
			WSACleanup();
		#endif	
	}

	//
	/** \brief create a socket for use with TCP or UDP wrapper
	 * 
	 * 	type:	SOCK_STREAM -	 tcp socket
	 * 			SOCK_DGRAM	-	 udp socket
	 * 
	 * 	return:
	 * 		(socktype) SOCKET_ERROR	on error
	 * 								an appropriate socket otherwise 
	 * 
	 * 
	 *  This function is part of the platform-independence wrapping and will be accessed by 
	 *  platform-independent parts of the Network Queue.
	 */ 
	socktype platformSocketCreateInet( INT32 type )
	{
		#ifdef PLF_UNIX_TCPIPSTACK
			return( socket( PF_INET, type, 0 ) );
		#endif

		#ifdef PLF_LWIP_TCPIPSTACK
			return( lwip_socket( PF_INET, type, 0 ) ); 
		#endif

		#ifdef PLF_WINNT_TCPIPSTACK
			socktype sockret;
			int cc1,cc2;
			int optval;
    
			sockret = socket( AF_INET, type, 0 );
    
			if (sockret == INVALID_SOCKET) return INVALID_SOCKET;
    
			optval = SO_RECVBUF_SIZE;
			cc1 = setsockopt(sockret,SOL_SOCKET,SO_RCVBUF,(const char *)&optval,sizeof(int));
    
			optval = SO_SENDBUF_SIZE;
			cc2 = setsockopt(sockret,SOL_SOCKET,SO_SNDBUF,(const char *)&optval,sizeof(int));

			if (cc1 != 0 || cc2 != 0) 
			{
				closesocket( sockret );
				return INVALID_SOCKET;
			}
    
			return sockret;
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            socktype sockret;
            int cc1,cc2;
            int optval;
    
            sockret = bsd_socket( PF_INET, type, 0 );
    
            if (sockret == INVALID_SOCKET) return INVALID_SOCKET;
    
            optval = NICHE_SO_RECVBUF_SIZE;
            cc1 = t_setsockopt(sockret,SOL_SOCKET,SO_RCVBUF,(void *)&optval,sizeof(int));
    
            optval = NICHE_SO_SENDBUF_SIZE;
            cc2 = t_setsockopt(sockret,SOL_SOCKET,SO_SNDBUF,(void*)&optval,sizeof(int));

            if (cc1 != 0 || cc2 != 0) 
            {
                bsd_close( sockret );
                return INVALID_SOCKET;
            }

            return sockret;
        #endif
	}

	//
	// connect a previously created socket to the remote endpoint defined via Saddr 
	// used inside Network Queue for UDP connects
	//
	//		s		socket to be connected	
	//		Saddr	remote peer address to connect to
	//
	//	return:
	//			SOCKET_ERROR	on error
	//		0				successful connect
	//
	//
	// This function is part of the platform-independence wrapping and will be accessed by 
	// platform-independent parts of the Network Queue.
	//
	INT32 platformSocketConnect(socktype s, const struct sockaddr * Saddr, size_t size_sockaddr)
	{	
		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_WINNT_TCPIPSTACK)
			return( connect( s, Saddr, size_sockaddr) );
		#endif

		#ifdef PLF_LWIP_TCPIPSTACK
			return( lwip_connect( s, (sockaddr *) Saddr, size_sockaddr) );
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            return( bsd_connect( s, (sockaddr *) Saddr, size_sockaddr) );
        #endif

	}

	//
	// bind a previously created socket to a local address defined via Saddr 
	// used inside Network Queue for UDP bind
	//
	//		s		socket to be bound
	//		Saddr	local address to bind socket to
	//
	//	return:
	//		SOCKET_ERROR	on error
	//		0				successful bind
	//
	//
	// This function is part of the platform-independence wrapping and will be accessed by 
	// platform-independent parts of the Network Queue.
	//
	INT32 platformSocketBind( socktype s, const struct sockaddr * Saddr, size_t size_sockaddr)
	{
		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_WINNT_TCPIPSTACK)
			return( bind( s, Saddr, size_sockaddr) );
		#endif

		#ifdef PLF_LWIP_TCPIPSTACK
			return( lwip_bind( s, (sockaddr *) Saddr, size_sockaddr) );
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            return( bsd_bind( s, (sockaddr *) Saddr, size_sockaddr) );
        #endif
	}

	//
	// close or cleanup a socket (connected, but must work on unconnected sockets, too)
	// used inside Network Queue for UDP close()
	//
	//		s		socket to be cleaned up and closed
	//
	//	return:
	//		SOCKET_ERROR	on error
	//		0				successful close
	//
	//
	// This function is part of the platform-independence wrapping and will be accessed by 
	// platform-independent parts of the Network Queue.
	//
	INT32	platformSocketClose( socktype s )
	{
		#ifdef PLF_UNIX_TCPIPSTACK
			//shutdown(s, SHUT_WR);
	
			return( close( s ) ); 
		#endif
	
		#ifdef PLF_WINNT_TCPIPSTACK
			const int SD_SEND_WIN = 0x01;
			shutdown(s, SD_SEND_WIN);
	
			return( closesocket( s ) );
		#endif
  
		#ifdef PLF_LWIP_TCPIPSTACK
			return( lwip_close( s ) );
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            return( bsd_close( s ) );
        #endif

	}

	INT32 platformSocketSelect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout)
	{
		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_WINNT_TCPIPSTACK)
			return( select( n, readfds, writefds, exceptfds, timeout) );
		#endif

		#ifdef PLF_LWIP_TCPIPSTACK
			return( lwip_select( n, readfds, writefds, exceptfds, timeout) );
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            return( bsd_select( n, readfds, writefds, exceptfds, timeout) );
        #endif

	}


// wrapper for receive data on a connected socket
INT32 platformSocketReceiveFunction( socktype s, void *buf, UINT32 buflen)
{
	return( recv(s, (char *) buf, (INT32) buflen, 0) );
}

// wrapper for receive data on a connected socket
INT32 platformSocketSendFunction( socktype s, const void *buf, UINT32 buflen)
{
	return( send(s, (const char *) buf, (INT32) buflen, 0) );
}


	INT32		platformSocketReceiveFromFunction( socktype s, void *buf, UINT32 buflen, UINT32 *ip_hostorder, UINT16 *port_hostorder)
	{
		INT32 retval;

		struct sockaddr_in fromaddr;
	
		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_LWIP_TCPIPSTACK) 
			socklen_t addrlen;
		#endif
		#if defined(PLF_WINNT_TCPIPSTACK) || defined(PLF_NICHE_TCPIPSTACK)
			INT32 addrlen;
		#endif

		addrlen = sizeof(fromaddr);

		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_WINNT_TCPIPSTACK)
			retval = recvfrom(s, (char *) buf, (INT32) buflen, 0, (sockaddr *) &fromaddr, &addrlen );
		#endif

		#ifdef PLF_LWIP_TCPIPSTACK
			retval = lwip_recvfrom(s, (char *) buf, (INT32) buflen, 0, (sockaddr *) &fromaddr, &addrlen );
		#endif
  
        #ifdef PLF_NICHE_TCPIPSTACK
            retval = bsd_recvfrom(s, (char *) buf, (INT32) buflen, 0, (sockaddr *) &fromaddr, &addrlen );
        #endif
  
		if (retval >= 0)
		{
			if (ip_hostorder != NULL)
				*ip_hostorder =  ntohl(fromaddr.sin_addr.s_addr); 
			if (port_hostorder != NULL)
				*port_hostorder = ntohs(fromaddr.sin_port);
		}

		return retval;
	}

	INT32		platformSocketSendToFunction( socktype s, const void *buf, UINT32 buflen, UINT32 ip_hostorder, UINT16 port_hostorder )
	{
		struct sockaddr_in toaddr;

		toaddr.sin_family = AF_INET;
		toaddr.sin_port = htons(port_hostorder);
		toaddr.sin_addr.s_addr = htonl(ip_hostorder);

		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_WINNT_TCPIPSTACK) 
			return( sendto(s, (const char *) buf, (INT32) buflen, 0, (const sockaddr *) &toaddr, sizeof(toaddr)) );
		#endif
  
		#ifdef PLF_LWIP_TCPIPSTACK
			return( lwip_sendto(s, (void *) buf, (INT32) buflen, 0, (sockaddr *) &toaddr, sizeof(toaddr)) );  
		#endif 

        #ifdef PLF_NICHE_TCPIPSTACK
            return( bsd_sendto(s, (char *) buf, (INT32) buflen, 0, (sockaddr *) &toaddr, sizeof(toaddr)) );  
        #endif 
	}

	INT32   platformSocketDisconnectUDP( socktype s)
	{
		#if defined(PLF_UNIX_TCPIPSTACK) || defined(PLF_WINNT_TCPIPSTACK) 
			struct sockaddr disconn;

			memset(&disconn, 0, sizeof (disconn));
			disconn.sa_family = AF_UNSPEC;      // not really necessary 
			INT32 ret = connect(s, &disconn, sizeof(disconn) );
			if (ret == SOCKET_ERROR) return SOCKET_ERROR;

			return 0; // success
			// -1 failure
		#endif
  
		#ifdef PLF_LWIP_TCPIPSTACK
			// SW! afaik on LWIP side there is no udp disconnect socket function!
			return 0; // success
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            struct sockaddr disconn;

            memset(&disconn, 0, sizeof (disconn));
            disconn.sa_family = AF_UNSPEC;      // not really necessary 
            INT32 ret = bsd_connect(s, &disconn, sizeof(disconn) );
            if (ret == SOCKET_ERROR) return SOCKET_ERROR;

            return 0; // success
            // -1 failure
        #endif

	}

	//
	// wrapper in order to put the socket to nonblocking mode 
	//
	// 0 == success
	// or SOCKET_ERROR on error
	//

	INT32 platformSocketSetNonblocking ( socktype s )
	{
		#ifdef PLF_UNIX_TCPIPSTACK
			return fcntl( s, F_SETFL, FNONBLOCK | FASYNC );
		#endif

		#ifdef PLF_WINNT_TCPIPSTACK
			unsigned long nonblock = 1;
	
			return( ioctlsocket( s, FIONBIO, &nonblock) );
		#endif
	
		#ifdef PLF_LWIP_TCPIPSTACK
			unsigned long nonblock = 1;
  
			return( lwip_ioctl( s, FIONBIO, &nonblock) );
		#endif

        #ifdef PLF_NICHE_TCPIPSTACK
            unsigned long nonblock = 1;
  
            return( bsd_ioctl( s, FIONBIO, &nonblock) );
        #endif

	} 


hostent*			platformSocketGetHostByName( char *hostname )
{
	#ifdef PLF_WINNT_TCPIPSTACK
		return gethostbyname(hostname);
	#endif

	// SW! all others than windows unimplemented at the moment

	return NULL;
}

char	   *platformSocketGetOSErrorString( UINT32 intOSLastError, char *lasterrorstr, UINT32 length_string)
{
#ifdef PLF_WINNT_TCPIPSTACK
	
	LPSTR MessageBuffer;
	DWORD dwBufferLength;

	DWORD dwFormatFlags =	FORMAT_MESSAGE_ALLOCATE_BUFFER |
							FORMAT_MESSAGE_IGNORE_INSERTS |
							FORMAT_MESSAGE_FROM_SYSTEM ;

	//
	// Call FormatMessage() to allow for message 
	//  text to be acquired from the system 
	//  or from the supplied module handle.
	//

	strcpy(lasterrorstr, "Unknown error.");

	dwBufferLength = FormatMessageA(dwFormatFlags, 
									NULL, // module to get message from (NULL == system)
									intOSLastError,
									MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // default language
									(LPSTR) &MessageBuffer,
									0,
									NULL);
	if (dwBufferLength > 0)
	{
		// remove CRLF from string if it exists!
		if (MessageBuffer[strlen(MessageBuffer)-1] == 10) MessageBuffer[strlen(MessageBuffer)-1] = 0;
		if (MessageBuffer[strlen(MessageBuffer)-1] == 13) MessageBuffer[strlen(MessageBuffer)-1] = 0;
			

		unsigned int len = strlen(MessageBuffer);
			
		if (len >= length_string) len = length_string - 1;
		memcpy(lasterrorstr, MessageBuffer, len);
		lasterrorstr[len] = 0;

	    LocalFree(MessageBuffer);
	}

	return lasterrorstr;

#endif

	// SW! only WINNT stack implemented by now

	return "";
}


bool		platformSocketIsBlockingError( UINT32 errcode )
{

#ifdef PLF_WINNT_TCPIPSTACK
	if (errcode == WSAEWOULDBLOCK) return true;
#endif

#ifdef PLF_UNIX_TCPIPSTACK
	if (errcode == EWOULDBLOCK) return true;
#endif

	return false;
}
