#ifndef __PLATFORM_H_INCLUDED_SW__
#define __PLATFORM_H_INCLUDED_SW__

/** \file platform.h
 * \brief Declares platform-specifics.

	\author Stefan Weisse / DESY 
	\date   2008
    
	Does neccessary platform-based includes, definitions, data type typedefs, platform-based structure typedefs 
	and forward declaration for platform-wrapped functions. Works in conjunction with platform.cpp, which does 
	necessary definition of function based on defines done in here.
 
	In order to tie to a special platform, a general include file named "system.h" needs to be created inside the
	project. These file should not include much, but just \#define the desired platform type and \#undef all other 
	defined platform types. It is needed for proper branching inside platform.h.

	For abstract data types, some typedefs were invented:
	INT32		signed 32bit integer value
	UINT32		unsigned 32bit integer value

	DOUBLE		double precision float value (64 bits in size)
	FLOAT		single precision float value (32 bits in size)

	INT8		signed 8bit integer value
	UINT8		unsigned 8bit integer value

	INT16		signed 16bit integer value
	UINT16		unsigned 16bit integer value

	INT64		signed 64bit integer value
	UINT64		unsigned 64bit integer value

	HANDLE		pointer to memory location (e.g. mapped to void *)

	In addition to that, certain defines and typedefs will be made for
	thread creation compatibility, for function-mapping like snprintf and for
	path separator(s) and wildcards in platform-dependent manner which is needed 
	for configuration file access, naming lookup and support.
*/

#ifdef SYSTEM_LINUX
	#include <stdio.h>	
	#include <stdlib.h>
	#include <errno.h>
	#include <string.h>
	#include <ctype.h>
	#include <termios.h>
	#include <assert.h>
	#include <unistd.h>
	
	#include <sys/timeb.h>
	#include <sys/times.h>
	#include <sys/types.h>
	#include <pthread.h> 
	#define  PLF_USING_PTHREAD

	#define  PLF_SYSTEM_LINUX
	#define  PLF_UNIX_TCPIPSTACK


	#define  PLF_PATH_SEPARATOR_CHAR '/'
	#define  PLF_PATH_SEPARATOR_CHAR_AS_STRING "/"
	#define  PLF_ALLFILES_WILDCARD	"*"
	#define  MAX_STRING_LEN		1024
	#define  platformvsnprintf vsnprintf
	#define  platformsnprintf snprintf
	#define  platformstrcasecmp strcasecmp
	#define  platformstrncasecmp strncasecmp

	#define  PLF_MAX_FDSETS 		64
	#undef PLF_ENDIANCHANGE // SW! not properly implemented
#endif

#ifdef SYSTEM_WINDOWS_VC6
	#include <stdio.h>
	#include <windows.h>
	#include <conio.h>
	#include <time.h>
	#include <assert.h>
	
	#define  PLF_SYSTEM_WINNT
	#define  PLF_WINNT_TCPIPSTACK
	#define  PLF_PATH_SEPARATOR_CHAR '\\'
	#define  PLF_PATH_SEPARATOR_CHAR_AS_STRING "\\"
	#define  PLF_ALLFILES_WILDCARD	"*.*"	
	#define  platformvsnprintf _vsnprintf
	#define  platformsnprintf _snprintf
	#define  platformstrcasecmp _stricmp
	#define  platformstrncasecmp _strnicmp

	#define  PLF_MAX_FDSETS 		FD_SETSIZE
	#define PLF_WINNT_TCPIPSTACK

	// SW! check whether really needed
	#define SO_SENDBUF_SIZE 40960
	#define SO_RECVBUF_SIZE 81920 

	#undef PLF_ENDIANCHANGE // SW! not properly implemented
#endif
 
#ifdef SYSTEM_WINDOWS_VS2005
	#include <stdio.h>
	#include <windows.h>
	#include <conio.h>
	#include <time.h>
	#include <assert.h>

	#pragma warning(disable : 4996) // disable deprecated warnings
	
	#define  PLF_SYSTEM_WIN2005
	#define  PLF_WINNT_TCPIPSTACK
	#define  PLF_PATH_SEPARATOR_CHAR '\\'
	#define  PLF_PATH_SEPARATOR_CHAR_AS_STRING "\\"	
	#define  PLF_ALLFILES_WILDCARD	"*.*"	
	#define  platformvsnprintf _vsnprintf
	#define  platformsnprintf _snprintf
	#define  platformstrcasecmp _stricmp
	#define  platformstrncasecmp _strnicmp
	
	#define  PLF_MAX_FDSETS 		FD_SETSIZE
	#define PLF_WINNT_TCPIPSTACK	
	#undef PLF_ENDIANCHANGE // SW! not properly implemented

	// SW! check whether really needed
	#define SO_SENDBUF_SIZE 40960
	#define SO_RECVBUF_SIZE 81920 

#endif

#if defined(PLF_SYSTEM_LINUX)
	#ifndef INT32
		typedef		int				INT32;
	#endif

	#ifndef UINT32
		typedef		unsigned int	UINT32;
	#endif

	#ifndef DOUBLE
		typedef		double			DOUBLE;
	#endif

	#ifndef INT8
		typedef		signed char		INT8;
	#endif

	#ifndef UINT8
		typedef		unsigned char	UINT8; 
	#endif

	#ifndef UINT16
		typedef		unsigned short	UINT16;
	#endif

	#ifndef INT16
		typedef		signed short	INT16;
	#endif

	#ifndef INT64
		typedef		long long			INT64;
	#endif

	#ifndef UINT64
		typedef		unsigned long long UINT64;
	#endif

	#ifndef FLOAT
		typedef		float			FLOAT;
	#endif

	#ifndef HANDLE
		typedef		void *	HANDLE;
	#endif

	typedef void * PLF_SW_THREADRETURN;
	#define PLF_SW_THREADMARKER
	typedef void * PLF_SW_THREADPARAMETER;
	#define PLF_SW_THREADRETURN_VALUE NULL
	
	typedef struct {
		UINT32 time;
		UINT32 millitm; // make it 32bit for pairing/aligning
	} PLF_TIME;
	

#endif



#if defined(PLF_SYSTEM_WINNT) || defined(PLF_SYSTEM_WIN2005)
	#ifndef INT32
		typedef		int				INT32;
	#endif

	#ifndef UINT32
		typedef		unsigned int	UINT32;
	#endif

	#ifndef DOUBLE
		typedef		double			DOUBLE;
	#endif

	#ifndef INT8
		typedef		signed char		INT8;
	#endif

	#ifndef UINT8
		typedef		unsigned char	UINT8; 
	#endif

	#ifndef UINT16
		typedef		unsigned short	UINT16;
	#endif

	#ifndef INT16
		typedef		signed short	INT16;
	#endif

	#ifndef INT64
		typedef		__int64			INT64;
	#endif

	#ifndef UINT64
		typedef		unsigned __int64 UINT64;
	#endif

	#ifdef socktype
		#error socktype already defined! see Types.h  
	#else
		typedef		int				socktype;
	#endif

	typedef DWORD PLF_SW_THREADRETURN;
	#define PLF_SW_THREADMARKER WINAPI
	typedef LPVOID PLF_SW_THREADPARAMETER;
	#define PLF_SW_THREADRETURN_VALUE 0

	typedef struct {
		UINT32 time;
		UINT32 millitm; // make it 32bit for pairing/aligning
	} PLF_TIME;

#endif


//#ifdef PLF_SYSTEM_LINUX
//#endif

/**	
	\brief Platform-independent type definition of thread startup function.
	
	Type definition in a platform-independent manner for a thread which can be started.
	Tested at least for posix style and for Win32 API.
	
	\remark This is pure internal type and subject to change if other platforms require change.
*/ 
typedef PLF_SW_THREADRETURN (PLF_SW_THREADMARKER *PLF_SW_THREAD_START_ROUTINE)(
    PLF_SW_THREADPARAMETER aThreadParameter
    );


/**	\brief Creates a counting semaphore with initial and maximum count settings.
	
	\remark This function must support binary semaphore (0,1) or (1,1) also.

	\param initial_count	The initial count of this semaphore, must be less than or equal to maximum count
	\param maximum_count	The maximum count of this semaphore, must be less than or equal max(UINT16)
	\return proper HANDLE in success or NULL in case of error
 */
extern HANDLE	platformCreateSemaphore( UINT16 initial_count, UINT16 maximum_count );

/** \brief Deletes a previously created counting semaphore.
 
	\param semaph	The semaphore handle to be deleted (must not be used after true return)
	\retval true successful deletion
	\retval false unsuccessful deletion or error case
 */ 
extern bool		platformDeleteSemaphore( HANDLE semaph );

/**
	\brief Waits for semaphore 'free' in order to take it (decrement by 1 if not 0).

	\param semaph			The semaphore handle to be obtained (must not be NULL)
	\param timeout_msec		Timeout in milliseconds before giving up waiting if semaphore is already taken.
	\retval	true successful obtaining of Semaphore
	\retval false unsuccessful obtain or error
 */ 
extern bool		platformWaitForSemaphore( HANDLE semaph, UINT32 timeout_msec );

/**
	\brief Releases/signals semaphore count (increment by 1 if not max).

	\param semaphore	The semaphore handle to be released (must not be NULL)
	\retval true successful release
	\retval false unsuccessful release or error 
 */ 
extern bool		platformReleaseSemaphore( HANDLE semaphore );

/**
	\brief Probes a semaphore for being possible to take (count > 0).
	
	\param semaph		The semaphore handle to be probed (must not be NULL)
	\retval	true successful probing
	\retval false unsuccessful or error case
 */ 
extern bool		platformIsSemaphoreSignalled( HANDLE semaph ) ;

/**	\brief Wrapper for non-signalling sleep function.

	Pauses the calling thread by a discrete amount of time. 

	\param msec		The time in milliseconds to pause
	\remark Take special care on e.g. UNIX platforms so that signals do not improperly interrupt the sleeping.
 */ 
extern void		platformSleepMilliseconds(UINT32 msec); 

/** \brief Creates, initializes and starts up a thread (passed as arguments).

	\param	aThreadFunction	The thread function defined somewhere in your code.
	\param	aThreadParameter	The parameter will be passed to the thread. Very useful for passing context information.
	\return proper HANDLE in success or NULL in case of error
 */ 
extern HANDLE	platformCreateThread( PLF_SW_THREAD_START_ROUTINE aThreadFunction, void *aThreadParameter); 

/**
	\brief Deletes a previously created thread HANDLE.
	
	  The thread must be gracefully shut down before this function is called. It will clean up memory and 
	  possibly unregisters it from the API / OS.

	\param aThreadHandle		thread handle to be cleaned up / deleted
	\retval	true success
	\retval false unsuccessful, error case
 */ 
extern bool		platformDeleteThread(HANDLE aThreadHandle);

/** \brief Gets the current time of the system clock and puts it into aTimeStruct
	
	\param aTimeStruct		on true return code, the current time will be put in there
	\retval	true success
	\retval false returned in error case

	\remark On false return, the parameter aTimeStruct is left untouched!
 */ 
extern bool		platformGetTime( PLF_TIME &aTimeStruct);

/** \brief Copies (clones) a file from aSourceFilename to aCloneFilenameToCreate.
	
	\param aSourceFilename			name of the source file (should include full path), must exist
	\param aCloneFilenameToCreate	name of the destionation (clone) file, may or must not exist
	\param aFailIfExist				fail in case the destination file (clone) does already exist, false: overwrite it!
	\retval	true successfully cloned file contents
	\retval false unsuccessful, error case
 */ 
extern bool 	platformCopyFile( const char *aSourceFilename, const char *aCloneFilenameToCreate, bool aFailIfExist);

/**	\brief Returns full path to executable file (including filename).

	\param aPathBuffer			path buffer memory location
	\param aMaxBufferLength		maximum length (in chars) of aPathBuffer NOT including mandatory nullbyte
	\retval	true successful put into argument buffer
	\retval false unsuccessful or error case
 */ 
extern 	bool platformGetFullExecutablePath( char *aPathBuffer, UINT32 aMaxBufferLength);


////////////////////////////////////////////////////////////////////////
// socket function(s)
////////////////////////////////////////////////////////////////////////

extern 		void		platformInitializeSocketSubsystem();
extern 		void		platformDeinitializeSocketSubsystem();

extern		INT32		platformSocketGetLastError();
extern		socktype	platformSocketCreateInet( INT32 type );

extern		INT32		platformSocketReceiveFunction( socktype s, void *buf, UINT32 buflen);
extern		INT32		platformSocketSendFunction( socktype s, const void *buf, UINT32 buflen);

extern		INT32		platformSocketReceiveFromFunction( socktype s, void *buf, UINT32 buflen, UINT32 *ip_hostorder, UINT16 *port_hostorder);
extern		INT32		platformSocketSendToFunction( socktype s, const void *buf, UINT32 buflen, UINT32 ip_hostorder, UINT16 port_hostorder );

extern		INT32		platformSocketClose( socktype s);
extern		INT32		platformSocketConnect(socktype s, const struct sockaddr * Saddr, size_t size_sockaddr);
extern		INT32		platformSocketBind( socktype s, const struct sockaddr * Saddr, size_t size_sockaddr);
extern		INT32		platformSocketSetNonblocking ( socktype s );
extern		UINT32		platformIPAddrConvert( const char * const ip_addr_dotted_string);
extern 		char		*platformIPAddrConvert( UINT32 ip_hostorder );
extern		INT32		platformSocketDisconnectUDP( socktype s);
extern		INT32		platformSocketSelect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout);
extern hostent*			platformSocketGetHostByName( char *hostname );
extern char				*platformSocketGetOSErrorString( UINT32 intOSLastError, char *lasterrorstr, UINT32 length_string);
extern		bool		platformSocketIsBlockingError( UINT32 errcode );
extern const INT32		platformMaxFdSets; 



#endif // __PLATFORM_H_INCLUDED_SW__
