#ifndef __STRINGFUNC_H_INCLUDED__
#define __STRINGFUNC_H_INCLUDED__

/** \file stringfunc.h
 * \brief Basic string handling.
    \author Stefan Weisse / DESY
	\date   2005-2008

	Reimplemented MFC CString methods for use with platform-independent CIniEx.cpp and CIniEx.h were put to
	stringfunc.h and stringfunc.cpp for easier reuse. 
    
	The initialization file handling class CIniEx was initially written (taken from whe WWW) for 
	MFC/win32 API. In order to use it in platform-independend projects, certain functions had to be wrapped or
	changed. These functions were extracted from CIniEx.cpp and CIniEx.h and put into files stringfunc.h and 
	stringfunc.cpp for easier reuse. 

	Some functions contain little bugs which were found later but have not yet been fixed as they did not harm at
	any time the functionality of CIniEx. In future, the file handling is planned to be superseded by xml files.
	It is not decided whether these little bugs, which are noted as found, will be fixed until then.

	In certain functions a stringarray is referenced. A stringarray in that case is a memory buffer where strings
	(including single(!) nullbyte) are concatenated together. The end of the stringarray is defined by a double 
	nullbyte. e.g. "First String\0Second String\0Third and last String\0\0"


*/

//! \def MAX_STRING_LENGTH
//! Defines maximum string buffer length (including nullbyte terminator) for certain functions.
	#define MAX_STRING_LENGTH   600

//! \def MAX_PATH_LENGTH 
//! Defines maximum path buffer length (including nullbyte terminator) for certain functions.
	#define MAX_PATH_LENGTH     255

//!	Copies (clones) a string from src to dest.
//! \param dest	destination string buffer (must be long enough for strlen(src)+1
//! \param src source string buffer
	void	copywholestring(char *dest, const char *src);

//!	Finds occurence of string 'search' inside string 'str'.
//! \param str		string buffer of string where to search
//! \param search	string buffer that contains substring (string to search for)
//! \retval	position (index) of first occurence
//!	\retval -1 if not found or on error case
	int		find( const char *str, const char *search);

//!	Finds last occurence of character 'c' inside string 'str'.
//! \param str		string buffer in which to search for last occurence of 'c'
//! \param c		character to search for
//! \retval	position (index) of last occurence
//! \retval -1 if not found or on error case
	int		reversefind( const char *str, int c );

/**	Copies up to 'num' characters from left side of 'src' string buffer and 
    puts it to destination string buffer 'dest'.
 
	\param dest		destionation string buffer (must be at least num+1 chars big)
	\param src		source string buffer
	\param num		number of characters to copy
*/
	void	Left(char *dest, const char *src, int num);
	
/**	Copies up to 'count' characters starting from left position index 'left 
	of 'src' string buffer and puts it to destination string buffer 'dest'.
   
	\param dest	destination string buffer (must be long enough for count+1)
	\param src source string buffer
	\param left		left position index (where to start copy)
	\param count	number of characters to copy (at most)
*/
	void	Mid( char *dest, const char *src, int left, int count);
	
/**	Copies up to 'count' characters starting from left position index 'left'
	of 'src' string buffer and puts it to destination string buffer 'dest'.
   
	\param dest	destination string buffer (must be long enough for strlen(src)-left+1)
	\param src	source string buffer
	\param left	left position index (where to start copy)
*/
	void	Mid( char *dest, const char *src, int left);
	
/**	Copies up to 'num' last characters of 'src' string buffer and puts it to destination 
	string buffer 'dest'.
   
	\param dest	destination string buffer (must be long enough for num+1)
	\param src	source string buffer
	\param num	number (count) of right-hand-side characters to copy (at most)
*/
	void	Right( char *dest, const char *src, int num);

	
/**	Checks whether 'c' is the leftmost (first) character inside string buffer 'str'.
   
	\param str	string buffer
	\param c	character to look for
	\return		true if 'c' is the leftmost character, false otherwise
*/
	bool	isLeftmostChar(const char *str, int c);

/**	Checks whether 'c' is the rightmost (last) character inside string buffer 'str'.
   
	\param str	string buffer
	\param c	character to look for
	\return		true if 'c' is the rightmost character, false otherwise
*/
	bool	isRightmostChar( const char *str, int c);

/**	Adds the contents of string buffer 'add' to the string buffer 'dest'. It basically wraps strcat.
   
	\param dest	destination string buffer to cat to
	\param add	source string buffer to add to dest
*/
	void	addstr( char *dest, const char *add);

	
/**	Checks whether the string buffer 'str' is empty (nullbyte is first content).
   
	\param	str		string buffer to check for being empty
	\retval	1 str is empty
	\retval 0 str is not empty
*/
	int		IsEmpty( char *str );

/**	Reads a single line with fgets from file fp. The resulting string buffer line must be at
    least MAX_STRING_LENGTH chars long. The maximum number of chars this function will return is
	MAX_STRING_LENGTH-1 chars, because a nullbyte is appended automatically.
   
	\param fp	file pointer to already opened file for reading
	\param line	destination string buffer (at least MAX_STRING_LENGTH size)
	\retval	>0	number of characters that have been successfully read
	\retval 0	on end of file 
	\retval -1	on error
*/
	int		ReadStringInt(FILE *fp, char *line);
	
/**	Reads a single line with fgets from file fp. This function wraps 
	ReadStringInt in order to avoid redundancy in code.
   
	\param fp	file pointer to already opened file for reading
	\param line	destination string buffer (at least MAX_STRING_LENGTH size)
	\return		true if at least one character was read, false otherwise
*/
	bool	ReadString(FILE *fp, char *line);
	

/**	Trims (removes) any of the characters in defaultset until there are no more characters in line or a 
	character that is not member of 'defaultset' has been found
   
	\param line	in/out string buffer
	\param defaultset set of characters to be removed from the right-hand-side of string, default is space,
	newline and tabulator (3 chars)
*/
	void	TrimRight(char *line, char *defaultset=" \n\t");

/**	Deletes 'nCount' characters from string buffer 'string', starting from index position 'nIndex'. If
	starting from nIndex 'nCount' characters to delete are more than remaining string length, nCount is
	modified internally so that everything up to the end is deleted.
   
	\param string	in/out string buffer
	\param nIndex	index position (starting from left) where to start deleting
	\param nCount	number of characters to delete (at most), defaults to 1 character

	\return		remaining string length
*/
	int		DeleteString( char *string, int nIndex, int nCount=1);



/**	Adds 'string' string buffer contents (copy) to the open end of string array 'stringarray'. Stringarray
    will be altered and newly allocated. 
	
	\bug stringfunc.h Add(): it might be that it does not work properly for empty value
   
	\param stringarray	stringarray to add contents to
	\param string		string to add to stringarray

	\return		true on success, false on failure/error
*/
	bool	Add( char **stringarray, const char *string	);

/**	Gets the upper bound (maximum element; counting starts from zero) of the passed string array. To avoid
    redundancy, this function is implemented as a wrapper to GetSize().

   	\param stringarray	stringarray to browse for upper bound
	\retval	index	of maximum element (zero-based)
	\retval -1		stringarray is empty
*/
	int		GetUpperBound( const char *stringarray );

/**	Gets the number of elements of the passed string array. 

   	\param stringarray	stringarray to obtain number of elements contained within
	\retval	number of elements	in string array
	\retval 0					stringarray is empty
*/
	int		GetSize( const char *stringarray );

/**	Gets the number of elements of the passed string array.  

	\bug stringfunc.h GetAt(): it might be that it does not work properly for empty value

   	\param stringarray	stringarray 
 	\param index	index position of element to return direct pointer reference(!) to 
	\retval	reference (direct pointer!) to 'index'th element (zero based)
	\retval NULL error
*/
	char   *GetAt( const char *stringarray, int index);
	
/**	Sets/Changes a certain stringarray element. Although the string array is reallocated, a new element
	whose index was not existing cannot be appended by this function. No error code is reserved for that manner!
	In order to 'grow' the string array, use SetAtGrow function. 
	
	\bug stringfunc.h SetAt(): it might be that it does not work properly for empty value

   	\param stringarray	stringarray to update 
 	\param index	index position of element to change (zero-based)
	\param value	new string value of 'index'-th element (zero-based)
*/
	void	SetAt( char **stringarray, int index, const char *value);

/**	Sets/Changes a certain stringarray element. No error code is reserved for any error!
	
	\bug stringfunc.h SetAtGrow(): it might be that it does not work properly for empty value

   	\param stringarray	stringarray to update
 	\param index	index position of element to change (zero-based)
	\param value	new string value of 'index'-th element (zero-based)
*/
	void    SetAtGrow( char **stringarray, int index, const char *value);
	

/**	Removes a certain stringarray element. Reallocation will happen internally.
	
	\bug stringfunc.h RemoveAt(): it might be that it does not work properly for empty value

   	\param stringarray	stringarray to update
 	\param index	index position of element to remove (zero-based)
	
	\return true on successful removal, false otherwise
*/
	bool	RemoveAt(char **stringarray, int index);

#endif
