package de.desy.tine.client;
import java.net.InetAddress;
import java.util.Date;
import java.util.LinkedList;
import de.desy.tine.addrUtils.*;
import de.desy.tine.bitfieldUtils.TBitfield;
import de.desy.tine.bitfieldUtils.TBitfieldRegistry;
import de.desy.tine.dataUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.headers.*;
import de.desy.tine.io.TBucket;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.properties.TPropertyList;
import de.desy.tine.server.properties.TStockProperties;

/** 
 * TLink.java
 *
 * @brief User interface calls to access tine data links.
 *  
 * The methods presented here deal with linking local application data with
 * remote data sources.  Primarily you will need to create a TLink object,
 * passing the necessary identifying parameters of the remote server to which 
 * the link is to be established.  When the TLink object is instantiated, 
 * you will then need to retrieve the data using one of the retrieval 
 * methods, e.g. execute() for synchronous acquisition, or attach() for  
 * asynchronous acquisition, or receive() for the acquisition of network
 * globals.
 */
public final class TLink
{
  protected void finalize() throws Throwable
  {
    if (active) 
    {
      MsgLog.log("finalize TLink","closing active link to /"+cntName+"/"+expName+"/"+devName+"["+devProperty+"] in finalizer",0,null,1);
      close();
    }
    super.finalize();
  }
  public String cntName;
  public String expName;
  public String devName;
  public String devProperty;
  public TDataType dOutput;
  public TDataType dInput;
  public TDataType dError;
  private static int maxOutputBufferSize = 0xffff;
  private String arrayDelimiter;
  public short devAccess;
  public short devTimeout;
  public static int defaultTimeout = 500;  // milliseconds
  public short srvId; // for network subscriptions
  private InetAddress mcastGrp = null;
  public InetAddress getMulticastGroup() { return mcastGrp; }
  public void setMulticastGroup(InetAddress group) { mcastGrp = group; }
  private int tineProtocol = TLinkFactory.DEFAULT_PROTOCOL_LEVEL;
  boolean active;
  boolean terminate;
  boolean delayEstablishLink = false;
  boolean needsNotification = false;
  boolean isInsideCallback = false;
  boolean isWildcardLink = false;
  boolean isGlobalsLink = false;
  boolean isQuery = false;
  boolean useErrValue = false;
  boolean useErrObject = false;
  boolean isRedirected = false;
  TLinkHook hook = null;
  Object hookObj = null;
  public boolean adjustDefaultValues = false;
  public boolean retryOnTimeoutError = true;
  public boolean removeOnClose = false; // was true
  public boolean hasRenewed = false;
  public boolean hasNotifiedOnce = false;
  public boolean isBeingWatched = false;
  public boolean needsToStartLinkWatchdog = false;
  public TSrvEntry srvAddr;
  public String toString()
  {
    String s = getFullDeviceName() + " ["+getProperty()+"] ";
    s += TMode.toString(sub.mode) + " @" + sub.pollingRate + " msec";
    s += " <"+linkStatus+"> ";    
    return s;
  }
  TContract con;
  TSubscription sub;
  public TSubscription getSubscription() { return sub; }
  TPHdr pHdr;
  TReqHdr reqHdr;
  TLink boundTo = null;
  boolean cancelledWithDependencies = false;
  private TLinkGroup grp = null;
  public boolean isGrouped() {return grp == null ? false : true; }
  public TLinkGroup getGroup() { return grp; }
  LinkedList<TLink> dependencies = null;
  public LinkedList<TLink> getDependencies() { return dependencies; }
  public boolean hasDependencies() 
  { 
    if (dependencies != null && dependencies.size() > 0) return true;
    return false;
  }
  public boolean isBound() { return boundTo == null ? false : true; }
  public TLink getBoundLink() { return boundTo; }
  public boolean isCancelledWithDependencies() { return cancelledWithDependencies; }
  public void addDependency(TLink tlink)
  {
    if (dependencies == null) dependencies = new LinkedList<TLink>();
    dependencies.add(tlink);
  }
  public void rmvDependency(TLink tlink)
  {
    if (dependencies == null) return;
    dependencies.remove(tlink);
  }
  public int linkStatusLastNotification = 0;
  public int linkStatus;
  public int linkCounter;
  public boolean linkStale;
  public double linkErrValue;
  public String linkErrString;
  public byte[] linkStatusStringBuffer = new byte[TStrings.STATUS_SIZE];
  public InetAddress linkPeer;
  TCallback tcb;
  int callbackId;
  TLinkCallback tlcb;
  TWildcardLink twcl;
  private TDataTolerance dtf = null;
  public TDataTolerance getNotificationTolerance() { return dtf; }
  /**
   * @brief sets the tolerance level for asynchrnous event notification
   *
   * Link callback notification can be suppressed if changes in the link
   * data do not exceed the tolerance applied in this call.  The tolerance
   * can have an absolute part (in the units of the data) and/or a 
   * percent part. 
   * 
   * @param absoluteValue is the absolute tolerance any particular data element
   * is allowed to have before an event notification is called.
   * @param percentValue is the percent tolerance any particular data element
   * is allowed to have before an event notification is called.
   */
  public void setNotificationTolerance(float absoluteValue, float percentValue)
  {
    dtf = new TDataTolerance(absoluteValue,percentValue);
  }
  public void setNotificationTolerance(double absoluteValue, double percentValue)
  {
    dtf = new TDataTolerance((float)absoluteValue,(float)percentValue);
  }
  public void setNotificationTolerance(int absoluteValue, int percentValue)
  {
    dtf = new TDataTolerance(absoluteValue,percentValue);
  }
  public boolean isWithinTolerance()
  {
    if (dtf == null) return false; // no tolerance specified => always true
    if (dOutput == null) return false; // no output data object
    if (isGrouped()) return false; // no tolerances for grouped calls
    TDataType dRef = dtf.getReference();
    if (dRef == null) // first call ?
    {
      dRef = new TDataType(dOutput.getArrayLength(),dOutput.getFormat());
      dRef.dataCopy(dOutput);
      dtf.setReference(dRef);
      return false;
    }
    int len = dOutput.getArrayLength();
    float abs = dtf.absolute;
    float pcn = (float)(dtf.percent/100.0);
    switch (dOutput.getFormat())
    {
      case TFormat.CF_BYTE:
      {
        byte[] athis = (byte[])dOutput.getDataObject();
        byte[] alast = (byte[])dRef.getDataObject(); 
        for (int i=0; i<len; i++)
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_SHORT:
      {
        short[] athis = (short[])dOutput.getDataObject();
        short[] alast = (short[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_LONG:
      {
        int[] athis = (int[])dOutput.getDataObject();
        int[] alast = (int[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_FLOAT:
      {
        float[] athis = (float[])dOutput.getDataObject();
        float[] alast = (float[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      case TFormat.CF_DOUBLE:
      {
        double[] athis = (double[])dOutput.getDataObject();
        double[] alast = (double[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      
      case TFormat.CF_DLONG:
      {
        long[] athis = (long[])dOutput.getDataObject();
        long[] alast = (long[])dRef.getDataObject(); 
        for (int i=0; i<len; i++) 
          if (Math.abs(athis[i]-alast[i]) > abs + pcn*(float)alast[i])
          {
            dRef.dataCopy(dOutput);
            return false;
          }
        break;
      }
      
      default:
        return false;
    }
    return true;
  }
  public int getLinkAccessMode()
  {
    if (sub != null) return sub.mode;
    return TMode.CM_CANCEL;
  }
  public int getLinkPollingInterval()
  {
    if (sub != null) return sub.pollingRate;
    return 0;    
  }
  /**
   * @brief Returns the current status for this link
   * 
   * @return The Link status as tine error code
   */
  public int getLinkStatus()
  {
    return linkStatus < 0 ? 0 : linkStatus;
  }
  /**
   * @brief Returns the context of this link
   * 
   * @return The Context of the current link
   */
  public String getContext()
  {
    return cntName;
  }
  /**
   * @brief Returns the device server addressed in this link
   * 
   * @return The device server addressed in the current link
   */
  public String getDeviceServer()
  {
    return expName;
  }
  /**
   * @brief Returns the device (module) name addressed in this link
   * 
   * @return The device (module) name addressed in the current link
   */
  public String getDeviceName()
  {
    return devName;
  }
  /**
   * @brief Returns the device property addressed in this link
   * 
   * @return The device property addressed in the current link
   */
  public String getProperty()
  {
    return devProperty;
  }
  /**
   * @brief Returns the fully qualified device name addressed in this link
   * 
   * @return The the fully qualified device name addressed in the current link
   */
  public String getFullDeviceName()
  {
    return new String("/"+cntName+"/"+expName+"/"+devName);
  }
  /**
   * @brief Returns the output data object as a TDataType
   * 
   * @return The output data type 
   */
  public TDataType getOutputDataObject()
  {
    return dOutput;
  }
  /**
   * @brief Returns the output data input as a TDataType
   * 
   * @return The input data type 
   */
  public TDataType getInputDataObject()
  {
    return dInput;
  }
  /**
   * @brief The last link timestamp
   * 
   * @return The last link timestamp as a UTC java long value (milliseconds
   * since 1.1.1970)
   */
  public long getLastTimeStamp()
  {
    return (long)(dOutput.timestamp) * 1000 + (long)(dOutput.timestampMSEC);
  }

  /**
   * @brief The last link timestamp as a TINE double timestamp value
   * 
   * @return The last link timestamp as a UTC java long value (seconds
   * since 1.1.1970 with decimal fractional) 
   */
  public double getLastDataTimeStamp()
  {
    return (double)(dOutput.timestamp) + ((double)(dOutput.timestampMSEC)/1000);
  }
  /**
   * @brief The last link status in plain text
   * 
   * @return The last link status in plain text
   */
  public String getLastError()
  {
	  return getError(linkStatus);
  }      
  /**
   * @brief The specified error code in plain text
   * 
   * @param code Is the tine error code for which the error text is desired
   * 
   * @return The specified error code in plain text
   */
  public String getError(int code) 
  {
    if (code >= 512)
      return new String(linkStatusStringBuffer).trim();
    else
	    return TErrorList.getErrorString(code);
  }
  Date  date = new Date();
  /**
   * @brief Specifies an error value to fill in associated data buffers 
   * in case of link error.
   * 
   * @param errValue Is the error value to use in case of link errors
   * 
   * 0 if successful
   */
  public int setErrorValue(double errValue)
  {
    if (sub.mode == TMode.CM_CANCEL) return TErrorList.illegal_mode;
    linkErrValue = errValue;
    useErrValue = true;
    return 0;
  }
  /**
   * @brief Specifies an error string to fill in associated data buffers 
   * in case of link error.
   * 
   * @param errString Is the string to use in case of link errors
   * 
   * 0 if successful
   */
  public int setErrorValue(String errString)
  {
    if (sub.mode == TMode.CM_CANCEL) return TErrorList.illegal_mode;
    linkErrString = errString;
    useErrValue = true;
    return 0;
  }
  public int setDebugLevel(int level)
  {
    return tf.setDebugLevel(level);
  }
  public int getDebugLevel()
  {
    return tf.getDebugLevel();
  }
  public String getModuleAddressInfo()
  {
    return srvAddr.toString(expName,cntName);
  }
  private TLink getExistingTLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    return tf.getExistingLink(devname,devproperty,dout,din,devaccess);    
  }
  private void redirectLink()
  {
    tf.redirectLink(this);
  }
  private void relinkLink()
  {
    tf.reLinkLink(this);
  }
  private boolean isBlackListed()
  {
    return tf.isLinkBlackListed(this);
  }
  // TODO:  why is this so complicated ?
  // a name can be /<context> only (a global)
  // a name can be /<context>/<server> (no device name : e.g. some stock properties)
  // a name can be /<context>/<server>/<device> (most cases)
  // a name can be <server>  (bad idea, but historically allowed) (no context, no device)
  // a name can be <server>/<device> (bad idea, but historically allowed) (no context)
  private void splitDeviceName(String fullname)
  {
    int indx = fullname.indexOf('/');
    cntName = "DEFAULT";
    devName = "";
    if (indx == 0)  
    { // context given
      fullname = fullname.substring(1);
      if ((indx=fullname.indexOf('/')) == -1)
      { // no server name given -> assume globals server
        cntName = fullname;
        expName = "GLOBALS";
        return;
      }
      else
      { // set the context and move along
        cntName = fullname.substring(0,indx);
        fullname = fullname.substring(indx+1);
        indx = fullname.indexOf('/');
      }
    }
    else
    { // no context specified (discouraged!)
      expName = fullname;
      if (indx > 0)
      { // device name given
        expName = fullname.substring(0,indx);
        devName = fullname.substring(indx+1);        
      }
      return;
    }
    if (indx == -1)
    { // no device name
      expName = fullname;
      devName = "";
    }
    else
    { // normal case
      expName = fullname.substring(0,indx);
      devName = fullname.substring(indx+1);
    }
  }
  private void remapStringArrayToStringBuffer(TDataType d, boolean hasInputData)
  {
    try
    {
      if (Integer.parseInt(d.getTag()) > 0) return; // already done ?
    }
    catch (NumberFormatException e) {};
    d.setTag(""+d.dArrayLength);
//    int dlen = d.dArrayLength * 128; // best guess
//    if (hasInputData)
//    { // find the length in bytes ...
//      
//    }
//    d.dArrayLength = dlen;
  }
  private int makeLink(String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  {
    if (devproperty == null) return TErrorList.argument_list_error;
    if (devname == null) 
    { // assume a SITE global is desired
      devname = "/SITE"; 
    }
    String fullname;
    //tf = TLinkFactory.getInstance();
    //setDebugLevel(tf.getDebugLevel());
    if (hook == null) hook = tf.getTLinkHook(); 
    if (hook != null && (fullname=hook.getLinkName(devname)) != null)
    {
      Object objOut = dout == null ? null : dout.getDataObject();
      Object objIn = din == null ? null : din.getDataObject();
      if (dout != null) dOutput = dout; else dOutput = new TDataType();
      if (din != null) dInput = din; else dInput = new TDataType();
      hookObj = hook.create(fullname,devproperty,objOut,objIn);
      devName = new String(fullname);
      devProperty = new String(devproperty);
      expName = "cdi";
      cntName = "localhost";
      return 0;
    }
    fullname = new String(devname);
    active = true;
    if (din != null)
    { // in case of 'late additions' ...
      din.putData();
    }
    if (dout != null) dOutput = dout; else dOutput = new TDataType();
    if (din != null) dInput = din; else dInput = new TDataType();
    if (arrayDelimiter != null) dOutput.setArrayDelimiter(arrayDelimiter);

    splitDeviceName(fullname);

    devProperty = new String(devproperty);
    devAccess = devaccess;
    devTimeout = (short)1000;

    if (dOutput.dFormat == TFormat.CF_STRING) remapStringArrayToStringBuffer(dOutput, false);
    if (dInput.dFormat == TFormat.CF_STRING) remapStringArrayToStringBuffer(dInput, true);
    if (dOutput.dArrayLength == 0)
    { // strip the read access off if no data returned
      devAccess &= ~(TAccess.CA_READ);
      dOutput.dFormat = TFormat.CF_NULL;
    }
    TPropertyList pl = TStockProperties.getPropertyList();
    isQuery = pl != null ? TStockProperties.getPropertyList().hasProperty(devProperty) : false;
    if (isBlackListed())
    { // link parameters have been black listed -> don't send out
      if (TLinkFactory.debugLevel > 0) DbgLog.log("makeLink","link " + getFullDeviceName() + " " + getProperty() + " has been black listed");
      linkStatus = TErrorList.link_blacklisted;
      return linkStatus;
    }
    redirectLink();
    relinkLink();
    remapIfBitfield();

    isGlobalsLink = (expName.compareToIgnoreCase("GLOBALS") == 0) && !TQuery.isStockProperty(devProperty);
    if (expName.compareToIgnoreCase("CYCLER") == 0 && devProperty.compareToIgnoreCase("CycleNumber") == 0)
    { // a very special "GLOBAL"
      isGlobalsLink = true;
    }
    srvAddr = new TSrvEntry(expName,cntName);
    if (srvAddr.fecAddr == null && !isGlobalsLink)
    {
      linkStatus = TErrorList.address_unknown;
      tf.addLinkToBlackList(this);
      String msg = "Could not resolve address for "+getFullDeviceName()+" : "+TErrorList.getErrorString(linkStatus);
      RuntimeException e = new RuntimeException(msg);
      MsgLog.log("makeLink",msg,TErrorList.address_unknown,e,1);
      throw e;
    }
    // redirecLink() or relinkLink() may have altered the original deviceName
    TLink xlnk = getExistingTLink(this.getFullDeviceName(),this.devProperty,dout,din,devaccess);
    if (xlnk != null)
    { // will NOT add this TLink to the link table !
      synchronized (this) 
      {
        boundTo = xlnk;
        sub = xlnk.sub;
        con = xlnk.con;
        reqHdr = xlnk.reqHdr;
        linkId = xlnk.linkId;
        TDataType xdin = xlnk.getInputDataObject();
        if (xdin != null && din != null && din != xdin) 
        { // we have to lock the input data !
          din.isLocked = xdin.isLocked = true;
        }
        TDataType xdout = xlnk.getOutputDataObject();
        if (xdout != null && dout != null && xdout.hasBeenUpdated)
        {
          if (dout.acquireDefaultFormat) synchronized (dout) 
          {
            dout.dFormat = xdout.dFormat;
            dout.dArrayLength = xdout.dArrayLength;
            dout.dCompletionLength = xdout.dCompletionLength;
          }
          if (dout.dFormat != TFormat.CF_DEFAULT)
          { // if the parent doesn't know yet either don't do this
            dout.setDataObject(dout.dArrayLength,dout.dFormat);
            dout.dataCopy(xdout);
            dout.setDataTimeStamp(xdout.getDataTimeStamp());
          }
        }                          
        xlnk.addDependency(this);
      }
      return 0;
    }

    linkId = tf.registerLink(this); // adds a placeholder in the link table and returns the link id
    if (linkId < 0)
    {
      linkStatus = TErrorList.resources_exhausted;
      DbgLog.log("makeLink","link table is full, please close some links and try again!");
      return linkStatus;
    }
   
    this.tineProtocol = srvAddr.isLegacy ? TLinkFactory.DEFAULT_PROTOCOL_LEVEL-1 : TLinkFactory.DEFAULT_PROTOCOL_LEVEL; 
    con = new TContract(this);
    sub = new TSubscription(con,this); // going out ...
    reqHdr = new TReqHdr(tf.getUserName(),tineProtocol);
    String defaultTimeoutStr;
    if ((defaultTimeoutStr = System.getProperty("default.timeout")) != null)
    {
      defaultTimeout = Integer.parseInt(defaultTimeoutStr);
    }
    tf.activateLink(this); // sets the placeholder to this TLink !
    return 0;
  }

  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * @throws RunTimeException if deviceName does not point to a valid address
   *
   * \b Example:
   * 
   * \include eg_TLink1.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   * @see de.desy.tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String deviceName, String deviceProperty, TDataType dout, TDataType din, short devaccess)
  {
    makeLink(deviceName,deviceProperty,dout,din,devaccess);
  }
  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * @throws RunTimeException if deviceName does not point to a valid address
   *
   * \b Example:
   * 
   * \include eg_TLink1.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   * @see de.desy.tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String devname, String devproperty, TDataType dout, TDataType din, int devaccess)
  {
    makeLink(devname,devproperty,dout,din,(short)devaccess);
  }
  /** 
   * @brief Defines a tine Data link
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param fullDeviceNameAndProperty is the fully specified, property-appended
   * device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * or &lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * @param dout is a reference to the output data object (which will receive
   * the data from the server)
   * @param din is a reference to the input data object (which will send data
   * to the server)
   * @param devaccess is the requested control access mode (TMode.CA_READ, TMode.CA_WRITE
   * etc.)
   *
   * @return A valid TLink instance if successful or null if not
   *
   * @throws RunTimeException if fullDeviceNameAndProperty does not point to a valid address
   * 
   * \b Example:
   * 
   * \include eg_TLink2.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   * @see de.desy .tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String fullDeviceNameAndProperty, TDataType dout, TDataType din, short devaccess)
  {
    if (fullDeviceNameAndProperty == null) return; // TODO: throw exception
    String devname, devproperty;
    int cnt = 1;
    int indx = fullDeviceNameAndProperty.indexOf('/');
    if (indx == 0) cnt = 3;
    for (int i=0; i<cnt; i++)
    {
      indx +=  fullDeviceNameAndProperty.substring(indx+1).indexOf('/') + 1;
    }
    devname = new String(fullDeviceNameAndProperty.substring(0,indx));
    devproperty = new String(fullDeviceNameAndProperty.substring(indx+1));
    
    makeLink(devname,devproperty,dout,din,devaccess);
  }

  /** 
   * @brief Defines a read-only tine Data link with default behavior.
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param devname is the fully specified device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;
   * or &lt;server&gt;/&lt;device&gt;
   * @param devproperty is the desired property
   *
   * @return A valid TLink instance if successful or null if not
   *
   * @throws RunTimeException if deviceName does not point to a valid address
   * 
   * \b Example:
   * 
   * \include eg_TLink3.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   */
  public TLink(String deviceName, String deviceProperty)
  {
    TDataType dout = new TDataType(TLink.maxOutputBufferSize,TFormat.CF_DEFAULT);
    TDataType din = new TDataType();
    makeLink(deviceName,deviceProperty,dout,din,TAccess.CA_READ);
  }
  public static String[] splitDeviceAndPropertyFromName(String fullDeviceNameAndProperty)
  {
    String[] names = new String[2];
    int cnt = 1;
    int indx;
    
    if (fullDeviceNameAndProperty == null) return names;
    indx = fullDeviceNameAndProperty.indexOf('[');
    if (indx > 0)
    { // "check for /<context>/<server>/<device>[<property>]" form
      names[0] = new String(fullDeviceNameAndProperty.substring(0,indx));
      names[1] = new String(fullDeviceNameAndProperty.substring(indx+1,fullDeviceNameAndProperty.length()));      
    }
    else
    { // "check for /<context>/<server>/<device>/<property>" form
      indx = fullDeviceNameAndProperty.indexOf('/');
      if (indx == 0) cnt = 3;
      for (int i=0; i<cnt; i++)
      {
        indx +=  fullDeviceNameAndProperty.substring(indx+1).indexOf('/') + 1;
      }    
      names[0] = new String(fullDeviceNameAndProperty.substring(0,indx));
      names[1] = new String(fullDeviceNameAndProperty.substring(indx+1));
    }
    return names;
  }
  /** 
   * @brief Defines a read-only tine Data link with default behavior.
   * 
   * This defines a tine Data Link, verifies that the requested deviceName
   * exists and returns a non-null link object if successful.  This does not
   * establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param fullDeviceNameAndProperty is the fully specified, property-appended
   * device name as in /&lt;context&gt;/&lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   * or &lt;server&gt;/&lt;device&gt;/&lt;property&gt;
   *
   * @return A valid TLink instance if successful or null if not
   *
   * @throws RunTimeException if fullDeviceNameAndProperty does not point to a valid address
   *
   * \b Example:
   * 
   * \include eg_TLink4.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#execute
   */
  public TLink(String fullDeviceNameAndProperty)
  {
    if (fullDeviceNameAndProperty == null) return; // TODO: exception
    String[] names = splitDeviceAndPropertyFromName(fullDeviceNameAndProperty);
    String devname = names[0], devproperty = names[1];
    TDataType dout = new TDataType(TLink.maxOutputBufferSize,TFormat.CF_DEFAULT);
    TDataType din = new TDataType();
    makeLink(devname,devproperty,dout,din,TAccess.CA_READ);
  }

  /** 
   * @brief Defines a tine GLobal Data link
   * 
   * This defines a tine Data Link and assumes that the requested parameter
   * is a network global, available on the tine multicast group.
   * This method does not establish the link! You must use either
   * the execute method or the attach method to establish the link.
   *  
   * @param globalKeyword is the desired network global parameter
   * @param dout is a reference to the output data object (which will receive
   * the network global data).
   *
   * @return A valid TLink instance if successful or null if not
   *
   * \b Example:
   * 
   * \include eg_TLink5.java
   *
   * @see de.desy.tine.client.TLink#attach
   * @see de.desy.tine.client.TLink#receive
   * @see de.desy.tine.dataUtils.TDataType#TDataType 
   */
  public TLink(String globalKeyword, TDataType dout)
  {
    makeLink(null,globalKeyword,dout,null,TAccess.CA_READ);
  }

  /** 
   * @brief Attaches the data sets to the created link at the default polling rate (1000 msec)
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   *  
   * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
   * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
   * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_attach1.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#execute
   */
  public int attach(int mode,TCallback f)
  {
    return attach(mode,f,1000,-1);
  }
  public int attach(short mode,TCallback f)
  {
    return attach((int)mode,f); 
  }
  /**
   * Stop thread execution until this link has been signalled.
   */
	   
  public synchronized int Twait()
  {
	 if (sub.id < 0) return -1;
	 try
	 {
		while (!this.terminate)
		{
		  wait();
//		  if (!this.terminate && 
//		       this.needsNotification &&
//		        tcb != null) tcb.callback(sub.id,linkStatus);
//		  this.needsNotification = false;
		}
	 }
	 catch (InterruptedException e) {}
//	 sub.mode = TMode.CM_CANCEL;
	 return -1;
  }            
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for 1000 msec
   * or until the call completes.
   *  
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpreted using getError() or getLastError().
   *
   * \b Example:
   * 
   * \include eg_execute1.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute()
  {
	 return execute(1000);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param appendMode will be appended to the tine Control Mode, which is 
   * TMode.CM_SINGLE. (e.g. TMode.CM_CONNECT).
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpreted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute2.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute(int timeout, short appendMode)
  {
    boolean retry = (devAccess & TAccess.CA_RETRY) == TAccess.CA_RETRY;
    return execute(timeout,appendMode,retry);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param retryOnError if 'true' will instruct the tine engine to issue
   * an automatic retry following a connection timeout. 
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpreted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute3.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute(int timeout, boolean retryOnError)
  {
    return execute(timeout,(short)0,retryOnError);
  }
  private int validLinkStatus(int status)
  {
    return status < 0 ? TErrorList.connection_timeout : status;
  }
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   * @param appendMode will be appended to the tine Control Mode, which is 
   * TMode.CM_SINGLE. (e.g. TMode.CM_CONNECT).
   * @param retryOnError if 'true' will instruct the tine engine to issue
   * an automatic retry following a connection timeout. 
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpreted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute4.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public synchronized int execute(int timeout, short appendMode,boolean retryOnError)
  {
    if (!tf.allowSynchronousLinks)
    {
      if (TLinkFactory.debugLevel > 0) DbgLog.log("execute","execute link ("+linkId+") "+getFullDeviceName()+"["+getProperty()+"] not allowed within link watchdog callback");
      return TErrorList.not_allowed;
    }
    if (TLinkFactory.debugLevel > 0) DbgLog.log("execute","execute link ("+linkId+") "+getFullDeviceName()+"["+getProperty()+"]");
    if (hookObj != null) return hook.execute(hookObj,timeout);
    if (this.isBlackListed())
    {
      MsgLog.log("execute","link execution error "+getFullDeviceName()+" ("+linkId+")",TErrorList.link_blacklisted,null,1);
      return TErrorList.link_blacklisted;
    }
    if (this.sub == null)
    {
      if (linkStatus == TErrorList.resources_exhausted) 
        MsgLog.log("execute","\nLink Table ("+tf.getLinkTable().length+" elements) is full",linkStatus,null,0);
      else
        MsgLog.log("execute",getFullDeviceName()+ " ["+getProperty()+"] has a null subscription <"+linkStatus+">",linkStatus,null,0);
      return linkStatus == 0 ? TErrorList.code_failure : linkStatus;
    }
    short bmode = TMode.getBaseMode(this.sub.mode);
    if (isGlobalsLink)
    {
      if (TLinkFactory.debugLevel > 0) DbgLog.log("execute", "remove globals attribute from synchronous link");
      isGlobalsLink = false;
    }
    if (isBound()) 
    { // different TLinks to the same thing in synchronous mode :
      //this.sub.id = (short)this.linkId;
      bmode = TMode.getBaseMode(this.boundTo.sub.mode);
      if (bmode > TMode.CM_CANCEL && bmode < TMode.CM_REGISTER)
      { // has been or is being used -> just copy the data ...
        //tf.activateLink(this);
        while (this.boundTo.linkStatus == -1)
        {
          try { Thread.sleep(1); }
          catch (InterruptedException e) {}
        }
        this.getOutputDataObject().pushBytes(this.boundTo.getOutputDataObject().getDataBuffer());
        return validLinkStatus(this.boundTo.getLinkStatus());
      }
      this.linkStatus = TErrorList.link_exists;
      this.linkErrString = "link is bound to an inactive link";
      return TErrorList.link_exists;
    }
    else
    { /* allow new data from the input reference under some circumstances */
      if (dInput != null && dInput.dArrayLength > 0 && !dInput.isLocked) 
        dInput.putData();
    }
    int sts = 0;
    if (this.linkId < 0)
    {
      sts = this.linkStatus = TErrorList.out_of_client_memory;
      return validLinkStatus(sts);
    }
    if (timeout < 100) timeout = 100;
    devTimeout = (short)timeout;
    //bmode = TMode.getBaseMode(this.sub.mode);
    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
    if (bmode == TMode.CM_REGISTER || bmode <= TMode.CM_SINGLE)
    { // this could already be an established monitor link ...
      this.sub.mode = TMode.CM_SINGLE;
      devAccess |= TAccess.CA_SYNC;
    }
    if ((appendMode & (TMode.CM_CONNECT|TMode.CM_STREAM))!= 0) this.sub.mode |= appendMode;
    if (retryOnError) devAccess |= TAccess.CA_RETRY; // make sure it's set
    if (tf.alwaysRetry) retryOnError = true;
    if ((devAccess & TAccess.CA_NORETRY) == TAccess.CA_NORETRY)
    { // overrides any global settings for this link
      retryOnError = false;
    }
    if ((devAccess & TAccess.CA_RETRY) == TAccess.CA_RETRY)
    { // overrides any global settings for this link
      retryOnError = true;
    }
    retryOnTimeoutError = retryOnError;
    linkStatus = TErrorList.link_not_open;
    //linkStale = true;  <- don't do this here!

    if (this.devName.indexOf("*") >= 0 && !TQuery.isStockProperty(this.devProperty))
    {
      TWildcardLink wc = tf.getWildcardLink();
      wc.parent = this;
      wc.format = this.dOutput.dFormat;
      /* reset these to the wildcard callback  */
      this.tlcb = wc.scb;
      //wait = TRUE;
      this.twcl = wc;
      this.isWildcardLink = true;
    }
    if ((sts=tf.sendLinkRequest(this)) != 0) // will set the link to active
    {
      MsgLog.log("execute","can't establish link",sts,null,1);
      devAccess &= ~(TAccess.CA_SYNC);
      return validLinkStatus(sts);
    }
   
    TLinkFactory.adjustLinkTable(this,TLinkFactory.adjustLinkTableAdd);
    if (TLinkFactory.debugLevel > 2) DbgLog.log("execute","link (" + linkId + ") is " + (removeOnClose ? "non-persistent" : "persistent"));
    for (int i=0; i<3; i++)
    {
      try
      {
        while (true)
        {
          wait();
          if (this.twcl == null || this.twcl.canNotify) break;
        }
        if (bmode == TMode.CM_SINGLE) this.active = false;
        devAccess &= ~(TAccess.CA_SYNC);
        if (linkStatus != 0 && (linkStatus & TErrorList.CE_SENDDATA) != TErrorList.CE_SENDDATA)
        {
          MsgLog.log("execute","link execution error "+getFullDeviceName()+" ("+linkId+")",linkStatus,null,1);
        }
        return validLinkStatus(linkStatus);
      }
      catch (InterruptedException e) 
      {
        linkStatus = TErrorList.net_read_error;
        MsgLog.log("execute","Interrupted io on "+getFullDeviceName()+" ("+linkId+")",linkStatus,e,1);
        //terminate = true;
      }
    }
    devAccess &= ~(TAccess.CA_SYNC);
    if (linkStatus != 0 && (linkStatus & TErrorList.CE_SENDDATA) != TErrorList.CE_SENDDATA)
    {
      MsgLog.log("execute","link execution error "+getFullDeviceName()+" ("+linkId+")",linkStatus,null,1);
    }
    return validLinkStatus(linkStatus);
  }                    
  /** 
   * @brief Executes a synchronous link using the data sets supplied at  
   * link creation.
   * 
   * Synchronous. This method executes the link using the data sets and 
   * parameters provided at instantiation.  The call will block for the
   * duration of the timeout specified or until the call completes.
   *
   * @param timeout The amount of time in milliseconds to block execution
   * until the call completes.
   *   
   * \note The TLink object created will remain resident until a call to
   * the cancel() method is made. Thus repeated calls to execute() can
   * be made without invoking separate calls to TLink(). 
   *  
   * @return 0 if successful, otherwise a tine return code which can be
   * interpreted using getError() or getLastError().
   * 
   * \b Example:
   * 
   * \include eg_execute5.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#attach
   */
  public int execute(int timeout)
  {
    return execute(timeout,(short)0);
  }                            
  /**
   * Terminates an active link
   */
  public int cancel()
  {
    if (hookObj != null) hook.close(hookObj);
    if (dInput != null) dInput.isLocked = false;
    if (isBeingWatched) 
    {
      TWatchdogLink wdl = tf.getWatchdogLink("/"+cntName+"/"+expName);
      if (wdl != null) wdl.remove(this);
    }
    return tf.cancel(this);
  }              
  /**
   * @brief Terminates an active link
   * 
   * Synonymous with cancel().
   */
  public int close()
  {
    return cancel();
  }              
  public int linkId;
  int linkBlacklists = 0;
  int linkTimeouts = 0;  
  boolean pending = false;
  boolean needsToSendLinkRequest = false;
  //static TErrorList tel;  
  public TLinkFactory tf = TLinkFactory.getInstance();; 
  TBucket tb;
  public TBucket getBucket() { return tb; }
  private void remapIfBitfield()
  {
    if (dOutput == null || !TFormat.isBitfield(dOutput.dFormat)) return;
    String tag = dOutput.getTag();
    if (tag.length() == 0) return;
    TQuery.AcquireAndRegisterBitfieldInfo(cntName, expName, tag, dOutput.dFormat);
    TBitfield bf = TBitfieldRegistry.getBitfield(cntName,expName,tag);
    if (bf == null) return;
    String fld = null;
    String[] parts = devProperty.split("\\.");
    if (parts.length > 1)
    { // field attached to property ?
      fld = parts[parts.length-1];
      if (bf.isField(fld))
      { // so far, so good
        dOutput.setBitField(bf);
        dOutput.setField(fld);
        devProperty = devProperty.substring(0, devProperty.indexOf(fld)-1);
        return; // finished
      }
    }
    parts = devName.split("\\.");
    if (parts.length > 1)
    { // field attached to device name ?
      fld = parts[parts.length-1];
      if (bf.isField(fld))
      { // so far, so good
        dOutput.setBitField(bf);
        dOutput.setField(fld);
        devName = devName.substring(0, devName.indexOf(fld)-1);
        return; // finished
      }
    }
  }
  public boolean isQueryLink()
  {
    if (expName.startsWith("ENS") || expName.startsWith("GENS"))
      return true;
    if (devProperty.indexOf("PROPS") >= 0 || 
        devProperty.indexOf("PROPERTIES") >= 0)
      return true;
    if (devProperty.indexOf("DEVICES") >= 0)
      return true;
    return false;
  }
  public boolean canSendPacked()
  {
    if (tf.isUseConnectedSockets()) return false;
    if (TMode.isConnected(sub.mode)) return false;
    if (isQueryLink()) return false;
    if ((devAccess & TAccess.CA_SYNC) != 0) return false;
    if ((sub.contract.eqmName.compareTo(TSrvEntry.SRVEQM_NAME)) == 0) return false;
    if ((sub.mode & TMode.CM_GLOBAL) == TMode.CM_GLOBAL) return false;
    if ((dInput.numblks > 1)) return false;
    return true;
  }
  TLink() {}   
  protected TLink(int tid, String devname, String devproperty, TDataType dout, TDataType din, short devaccess)
  { // called only by TLinkFactory.createTLink() via TSrvEntry().getSrvAddrFromENS
    String fullname = new String(devname);
    active = true;
    if (dout != null) dOutput = dout; else dOutput = new TDataType();
    dInput = din;
    //tf = TLinkFactory.getInstance();

    splitDeviceName(fullname);
    
    devProperty = new String(devproperty);
    devAccess = devaccess;
    devTimeout = (short)500;
    
    remapIfBitfield();
    
    linkId = tid = tf.registerLink(this); // an ENS link will return 0 here 
    if (TLinkFactory.debugLevel > 0) DbgLog.log("TLink","assigned link id " + linkId + " to /" + cntName + "/" + expName);
    srvAddr = new TSrvEntry(expName,cntName);
    if (srvAddr.fecAddr == null && TLinkFactory.debugLevel > 0) 
      DbgLog.log("TLink","Cannot resolve Server Address for /" + cntName + "/" + expName);
    // this constructor bypasses makeLink()
    this.tineProtocol = srvAddr.isLegacy ? TLinkFactory.DEFAULT_PROTOCOL_LEVEL-1 : TLinkFactory.DEFAULT_PROTOCOL_LEVEL; 
    con = new TContract(this);
    sub = new TSubscription(con,this); // going out ...
    reqHdr = new TReqHdr(System.getProperty("user.name"),tineProtocol);
    tf.activateLink(this);
    if (tid == 0) removeOnClose = false;
  }

/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach2.java
 *
 * @see de.desy.tine.client.TLink#TLink
 * @see de.desy.tine.client.TLink#TLinkCallback
 * @see de.desy.tine.client.TLink#execute
 */
  public int attach(int mode,TCallback f,int pollrate)
  {
	return attach(mode,f,pollrate,-1);
  }
  public int attach(short mode,TCallback f,int pollrate)
  {
    return attach((int)mode,f,pollrate);
  }
/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * and signals the callback routine using the give notification parameter. 
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 * @param notificationId is the desired identifier to be passed to the
 * supplied callback routine when it is fired.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach3.java
 *
 * @see de.desy.tine.client.TLink#TLink
 * @see de.desy.tine.client.TLink#TCallback
 * @see de.desy.tine.client.TLink#execute
 */
  public synchronized int attach(int mode,TCallback f,int pollrate,int notificationId)
  {
    short bmode = TMode.getBaseMode(mode);
    String msg = "attach link ("+linkId +") "+getFullDeviceName()+"["+getProperty()+"]"+" : " + TMode.toString((short)mode);
    if (bmode > TMode.CM_SINGLE)
      MsgLog.log("attach",msg,0,null,1);
    else
      if (TLinkFactory.debugLevel > 0) DbgLog.log("attach", msg);
    if (hookObj != null) return hook.attach(hookObj,pollrate,(short)mode,f,notificationId);
    if (this.isBlackListed())
    {
      this.linkStatus = TErrorList.link_blacklisted;
      return -1;
    }
    if (this.sub == null || linkId < 0)
    {
      if (linkStatus == TErrorList.resources_exhausted) 
        MsgLog.log("attach","\nLink Table ("+tf.getLinkTable().length+" elements) is full",linkStatus,null,0);
      else
        MsgLog.log("attach",getFullDeviceName()+ " ["+getProperty()+"] has a null subscription <"+linkStatus+">",linkStatus,null,0);
      return linkStatus == 0 ? -TErrorList.code_failure : -linkStatus;
    }
    if (dInput != null && bmode > TMode.CM_SINGLE) 
    { // don't let this change from above!
      dInput.isLocked = true; 
    }
    if (isGlobalsLink) mode = ensureGlobalsMode(mode);
    this.sub.mode = (short)mode;
    tcb = f;
    if (notificationId >= 0)
      callbackId = notificationId;
    else if (callbackId <= 0)
      callbackId = sub.id;
    if (pollrate < 100) pollrate = 100;
    devTimeout = (short)pollrate;
    this.sub.pollingRate = devTimeout;
    if (this.isBound()) return sub.id;

    if (this.devName.indexOf("*") >= 0 && !TQuery.isStockProperty(this.devProperty))
    {
      TWildcardLink wc = tf.getWildcardLink();
      wc.tcb = f;
      wc.parent = this;
      wc.format = this.dOutput.dFormat;
      /* reset these to the wildcard callback  */
      this.tlcb = wc.scb;
      //wait = TRUE;
      this.twcl = wc;
      this.isWildcardLink = true;
      this.sub.mode &= ~TMode.CM_GROUPED; // conflicting groups! => turn it off
    }
    if (canSendPacked()) delayEstablishLink = true;
    if (tf.sendLinkRequest(this) != 0) return -1;

    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;
 
    if (tf.alwaysRetry) retryOnTimeoutError = true;
    if ((devAccess & TAccess.CA_NORETRY) == TAccess.CA_NORETRY) 
      retryOnTimeoutError = false;
    if ((devAccess & TAccess.CA_RETRY) == TAccess.CA_RETRY) 
      retryOnTimeoutError = true;
    if ((this.sub.mode & TMode.CM_GROUPED) == TMode.CM_GROUPED) 
    {
      if (grp == null) grp = tf.getGroup(f);
      grp.addMember(this);
    }
    if ((mode & TMode.CM_USE_ON_ERROR) == TMode.CM_USE_ON_ERROR) 
    {
      dError = (TDataType)dOutput.clone();
      if (dError != null) useErrObject = true;
    } 
    if ((mode & TMode.CM_WAIT) == TMode.CM_WAIT) waitForLinkCompletion();
 
    TLinkFactory.adjustLinkTable(this,TLinkFactory.adjustLinkTableAdd);
    if (TLinkFactory.debugLevel > 0) DbgLog.log("attach","link " + linkId + " is " + (removeOnClose ? "non-persistent" : "persistent"));

    return sub.id;
  }
  public int attach(short mode,TCallback f,int pollrate,int notificationId)
  {
    return attach((int)mode,f,pollrate,notificationId);
  }
 private int ensureGlobalsMode(int thisMode)
 {
   int bmode = TMode.getBaseMode((short)thisMode);
   if (bmode == TMode.CM_POLL || bmode == TMode.CM_REFRESH)
   {
     thisMode &= ~(TMode.CM_POLL|TMode.CM_REFRESH);
     thisMode |= TMode.CM_GLOBAL;
   }
   return thisMode;
 }
/** 
 * @brief Attaches the data sets to the created link at the given polling rate
 * and signals the callback routine using a reference to the TLink instance. 
 * 
 * Asynchronous. This method attaches the data sets to the created link using the calling
 * parameters provided at instantiation.  The arriving data will automatically
 * update the output data set and then fire the callback method provided.
 * If the callback is not needed, a null can be passed for this parameter.
 * This is the preferred AttachLink method to use.  The Callback instantiates
 * a TLinkCallback() interface which provides a TLink object as parameter.
 *  
 * @param mode is the tine control mode parameter, that is one of TMode.CM_POLL, 
 * TMode.CM_REFRESH, TMode.SINGLE, etc. in possible combination with
 * TMode.CM_NETWORK or TMode.CM_CONNECT, etc.
 * @param f is the 'callback function' to be called whenever either new data 
 * arrive or a change in link status occurs.
 * The callback function should instantiate a class you create which 
 * implements the TLinkCallback() class interface.
 * @param pollrate is the desired polling rate (at the server) in milliseconds.
 *
 * @return A positive link handle if successful, otherwise -1.  On 
 * failure the linkStatus property of TLink can be examined.
 * 
 * \b Example:
 * 
 * \include eg_attach4.java
 *
 * @see de.desy.tine.client.TLink#TLink
 * @see de.desy.tine.client.TLink#TLinkCallback
 * @see de.desy.tine.client.TLink#execute
 */
  public synchronized int attach(int mode,TLinkCallback f,int pollrate)
  {
    short bmode = TMode.getBaseMode(mode);
    String msg = "attach link ("+linkId +") "+getFullDeviceName()+"["+getProperty()+"]"+" : " + TMode.toString((short)mode);
    if (bmode > TMode.CM_SINGLE)
    {
      MsgLog.log("attach",msg,0,null,1);
      if (bmode == TMode.CM_DATACHANGE && dtf == null)
      {
        setNotificationTolerance(0, 0);
        MsgLog.log("attach","suppressing null-change notification",0,null,1);
      }
    }
    else
      if (TLinkFactory.debugLevel > 0) DbgLog.log("attach", msg);
    if (hookObj != null) return hook.attach(hookObj,pollrate,(short)mode,f,this);
    if (this.isBlackListed())
    {
      this.linkStatus = TErrorList.link_blacklisted;
      return -1;
    }
    if (this.sub == null || linkId < 0)
    {
      if (linkStatus == TErrorList.resources_exhausted) 
        MsgLog.log("attach","\nLink Table ("+tf.getLinkTable().length+" elements) is full",linkStatus,null,0);
      else
        MsgLog.log("attach",getFullDeviceName()+ " ["+getProperty()+"] has a null subscription <"+linkStatus+">",linkStatus,null,0);
      return linkStatus == 0 ? -TErrorList.code_failure : -linkStatus;
    }
    tlcb = f;

    if (isGlobalsLink) mode = ensureGlobalsMode(mode);
    
    if (dInput != null) 
    { // link has input data
      if (bmode > TMode.CM_SINGLE) 
      { // don't let this change from above!
        dInput.isLocked = true; 
      }
      else if (bmode == TMode.CM_SINGLE)
      { /* allow new data from the input reference under some circumstances */
        if (dInput.dArrayLength > 0) dInput.putData();
      }
    }
    if (pollrate < 50) pollrate = 50;
    devTimeout = (short)pollrate;
    if (this.isBound()) 
    { // this link is bound to a parent !
      if (!this.boundTo.active)
      { // the original link is no longer active !
        this.boundTo.sub.mode = (short)mode; // must reset even if a dependent link (current could be SINGLE)
        this.boundTo.sub.pollingRate = pollrate;
        this.boundTo.cancelledWithDependencies = true;
        this.boundTo.active = true;
        this.boundTo.devAccess = this.devAccess;
        long ltime = System.currentTimeMillis();
        tf.sendLinkRequest(this.boundTo);
        dOutput.timestamp = (int)(ltime/1000);
        this.boundTo.sub.linkLastTime = ltime;
        this.boundTo.tlcb = this.tlcb;
        this.boundTo.tcb = this.tcb;
      }
      else 
      { // we're a dependent link to a live parent
        boolean fireevent = false;
        TDataType bdout = this.boundTo.getOutputDataObject();
        if (bdout.getFormat() != TFormat.CF_DEFAULT)
        { // parent data is a known quantity
          TDataType dout = this.getOutputDataObject();
          synchronized (dout)
          {
            if (dout.getFormat() == TFormat.CF_DEFAULT)
            { // made it this far without knowing the type ...
                dout.dFormat = bdout.dFormat;
                dout.dArrayLength = bdout.dArrayLength;
                dout.setDataObject(bdout.dArrayLength,bdout.dFormat);
            }
          }
          if (dout.hasBeenUpdated)
          { // safe to copy data and fire the callback below
            dout.dataCopy(bdout);
            dout.setDataTimeStamp(bdout.getDataTimeStamp());
            fireevent = true;
          }
        }
        bmode = TMode.getBaseMode(this.boundTo.sub.mode);
        if (TMode.getBaseMode(mode) == TMode.CM_TIMER && bmode != TMode.CM_TIMER)
        { // bound link has a higher polling hierarchy !
          long ltime = System.currentTimeMillis();
          int pmode = (this.boundTo.sub.mode & ~(bmode)) | TMode.CM_TIMER;
          this.boundTo.sub.mode = (short)pmode;
          if (pollrate < this.boundTo.sub.pollingRate)
            this.boundTo.sub.pollingRate = pollrate;
          tf.sendLinkRequest(this.boundTo);
          if (!this.boundTo.hasNotifiedOnce) dOutput.timestamp = (int)(ltime/1000);
          this.boundTo.sub.linkLastTime = ltime;          
        }
        else
        { // must reset even if a dependent link (current could be SINGLE)
          this.sub.mode = (short)mode; 
          this.sub.pollingRate = pollrate;
        }
        if (fireevent)
        { // have to fire this once here, otherwise there might be a long wait
          tf.fireCallbackEvent("attach", this,true);
        }
      }
      return this.sub.id;
    }
    this.sub.mode = (short)mode;
    this.sub.pollingRate = pollrate;

    if (this.devName.indexOf("*") >= 0 && !TQuery.isStockProperty(this.devProperty))
    {
      TWildcardLink wc = tf.getWildcardLink();
      wc.tlcb = f;
      wc.parent = this;
      wc.format = this.dOutput.dFormat;
      /* reset these to the wildcard callback  */
      this.tlcb = wc.scb;
      this.twcl = wc;
      this.isWildcardLink = true;
      this.sub.mode &= ~TMode.CM_GROUPED; // conflicting groups! => turn it off
    }
    
    if (canSendPacked()) delayEstablishLink = true;
    if (tf.sendLinkRequest(this) != 0) return -1;

    long ltime = System.currentTimeMillis();
    dOutput.timestamp = (int)(ltime/1000);
    this.sub.linkStartTime = ltime;
    this.sub.linkLastTime = ltime;

    if (tf.alwaysRetry) retryOnTimeoutError = true;
    if ((devAccess & TAccess.CA_NORETRY) == TAccess.CA_NORETRY)
    { // override the global settings with this 
      retryOnTimeoutError = false;
    }
    if ((devAccess & TAccess.CA_RETRY) == TAccess.CA_RETRY)
    { // override the global settings with this
      retryOnTimeoutError = true;
    }
    if ((this.sub.mode & TMode.CM_GROUPED) == TMode.CM_GROUPED) 
    {
      if (grp == null) grp = tf.getGroup(f);
      grp.addMember(this);
    }
    if ((mode & TMode.CM_USE_ON_ERROR) == TMode.CM_USE_ON_ERROR) 
    {
      dError = (TDataType)dOutput.clone();
      if (dError != null) useErrObject = true;
    } 
    if ((mode & TMode.CM_WAIT) == TMode.CM_WAIT) waitForLinkCompletion();

    TLinkFactory.adjustLinkTable(this,TLinkFactory.adjustLinkTableAdd);
    if (TLinkFactory.debugLevel > 0) DbgLog.log("attach","link " + linkId + " is " + (removeOnClose ? "non-persistent" : "persistent"));
    return sub.id;
  }
  public int attach(short mode,TLinkCallback f,int pollrate)
  {
    return attach((int)mode,f,pollrate);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using the assigned link id. 
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive1.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#receive(TCallback f,int notificationId)
   * @see de.desy.tine.client.TLink#receive(TLinkCallback f)
   */
  public synchronized int receive(TCallback f)
  {
    return attach(TMode.CM_GLOBAL,f,1000);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using a reference to the
   * given notifcation ID in the callback
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TCallback() class interface.
   * @param notificationId is the desired identifier to be passed to the
   * supplied callback routine when it is fired.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive2.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#receive(TCallback f)
   */
  public synchronized int receive(TCallback f,int notificationId)
  {
    return attach(TMode.CM_GLOBAL,f,1000,notificationId);
  }
  /** 
   * @brief Attaches the data sets to the global link
   * and signals the callback routine using a reference to the
   * referenced TLink object.
   * 
   * Asynchronous. This method attaches the data sets to the created link using the calling
   * parameters provided at instantiation.  The arriving data will automatically
   * update the output data set and then fire the callback method provided.
   * If the callback is not needed, a null can be passed for this parameter.
   * Using the receive() method assumes that the data requested are global available.
   * This is the preferred receive method to use.  The Callback instantiates
   * a TLinkCallback() interface which provides a TLink object as parameter.
   *  
   * @param f is the 'callback function' to be called whenever either new data 
   * arrive or a change in link status occurs.
   * The callback function should instantiate a class you create which 
   * implements the TLinkCallback() class interface.
   *
   * @return A positive link handle if successful, otherwise -1.  On 
   * failure the linkStatus property of TLink can be examined.
   * 
   * \b Example:
   * 
   * \include eg_receive3.java
   *
   * @see de.desy.tine.client.TLink#TLink
   * @see de.desy.tine.client.TLink#TLinkCallback
   * @see de.desy.tine.client.TLink#receive(TCallback f)
   */
  public synchronized int receive(TLinkCallback f)
  {
    return attach(TMode.CM_GLOBAL,f,1000);
  }
  /**
   * @brief Stops thread execution until this link has completed.
   * Useful for grouped asynchronous links.
   */
  public void waitForLinkCompletion() 
  {
    try
    {
  	 while (this.linkStatus < 0)
  	 {
  	   Thread.sleep(100);
  	 }
    }
    catch (InterruptedException e) 
    {
      DbgLog.log("waitForLinkCompletion"," waitForLinkCompletion : " + e);
    }
  }
  /**
   * @brief Gets the array delimiter for use in the toString() method 
   * of associated TDataType objects.
   * 
   * @return The delimiter in use in separating data
   * variables in a string dump (default ",")
   */
  public String getArrayDelimiter()
  {
    return arrayDelimiter;
  }
  /**
   * @brief Sets the array delimiter for use in the toString() method 
   * of associated TDataType objects.
   * 
   * @param delimiter is the string delimiter to use in separating data
   * variables in a string dump (default ",")
   */
  public void setArrayDelimiter(String delimiter)
  {
    if (arrayDelimiter != null && arrayDelimiter.compareTo(delimiter) == 0) return;
    arrayDelimiter = new String(delimiter);
    this.dOutput.setArrayDelimiter(arrayDelimiter);
  }
  public int getTineProtocol()
  {
    return tineProtocol;
  }
  public void setTineProtocol(int protocolLevel)
  {
    tineProtocol = protocolLevel;
    if (reqHdr != null) reqHdr.setTineProtocol(protocolLevel);
    if (dInput != null) dInput.resetCounters(protocolLevel);
    if (dOutput != null) dOutput.resetCounters(protocolLevel);
    if (sub != null && tineProtocol < TLinkFactory.DEFAULT_PROTOCOL_LEVEL) sub.isLegacy = true;
  }
}