package de.desy.tine.dataUtils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Vector;
import de.desy.tine.bitfieldUtils.TBitfield;
import de.desy.tine.bitfieldUtils.TBitfieldRegistry;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.headers.*;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.structUtils.*;
import de.desy.tine.types.*;

/**
 * @brief tine Data Type object All data transfered client <-> server must be
 *        contained in a TDataType object. Furthermore all constructors of
 *        TDataType take an array of tine data types as input argument, i.e. a
 *        single element should be passed as an array of length 1
 */
public final class TDataType implements Cloneable
{
  /**
   * @brief converts a value transmitted as an 'unsigned' integer to a java signed long
   * @param ival an encoded unsigned integer value
   * @return the corresponding java signed long
   */
  public static long UnsignedInt(int ival)
  {
    long lval = (long) ival;
    if (lval < 0) lval += 4294967296L;
    return lval;
  }
  /**
   * @brief forms an 'equivalent' unsigned integer for transmission
   * @param lval a java long value (signed) to be encoded as an unsigned integer
   * @return the encoded unsigned integer
   */
  public static int UnsignedInt(long lval)
  {
    if (lval > 2147483647) lval -= 4294967296L;
    return (int)lval;
  }
  /**
   * @brief converts a value transmitted as an 'unsigned' short to a java signed integer
   * @param sval an encoded unsigned short value
   * @return the corresponding java signed integer
   */
  public static int UnsignedShort(short sval)
  {
    int ival = (int)sval;
    if (ival < 0) ival += 65536;
    return ival;
  }
  /**
   * @brief forms an 'equivalent' unsigned short for transmission
   * @param ival a java integer value (signed) to be encoded as an unsigned short
   * @return the encoded unsigned short
   */
  public static short UnsignedShort(int ival)
  {
    if (ival > 32767) ival -= 65536;
    return (short)ival;
  }
  public boolean acquireDefaultFormat = false;
  public boolean isLocked = false;
  public static int RPCERR_SIZE = TStrings.STATUS_SIZE;
  private static int dBufferSegSize = TTransport.UDP_BUFFER_SIZE;
  public int dCompletionLength;
  public int dArrayLength;
  public short dFormat;
  protected Vector<Object> dVector;
  private boolean removeTagAtLinkTime = false;
  private String arrayDelimiter = new String("\n");
  private String dTag = new String(new StringBuffer(TStrings.TAG_NAME_SIZE));
  private TBitfield dBitField = null;
  private String dField = null;
  private int dSizeInBytes;
  private ByteArrayOutputStream dBuffer;
  private ByteArrayOutputStream[] dBuffers;
  public void pushBytes(byte[] bytestream)
  {
    try 
    {
      dBuffers[0].reset();
      dBuffers[0].write(bytestream);
    }
    catch (Exception e)
    {
    }
    return;
  }
  public long dTimestamp;
  public int timestamp;
  public int timestampMSEC;
  public int usrDataStamp;
  public int getUserDataStamp() { return usrDataStamp; }
  public void setUserDataStamp(int dataStamp) { usrDataStamp = dataStamp; }
  public int sysDataStamp;
  public int getSystemDataStamp() { return sysDataStamp; }
  public void setSystemDataStamp(int dataStamp) { sysDataStamp = dataStamp; }
  public int xferReason;
  public int numblks;
  public int blknum;
  public int blkid;
  public int blkcnt;
  public int mtu;
  public int blksin;
  public int bytesin;
  public boolean hasBeenUpdated = false;
  private Object hDataObject;
  private Object hStructObject;
  
  @Override
  public Object clone()
  {
    Object o = null;
    try
    {
      o = super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      System.err.println("TDataType cannot clone");
    }
    return o;
  }
  public void resetCounters(int tineProtocol)
  {
    switch (tineProtocol)
    {
      case 6:
        mtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubscription.hdrSizeInBytes - TReqHdr.hdrSizeInBytes);
        initCounters();
        break;    
      case 5:
        mtu = (TTransport.UDP_BUFFER_SIZE - TContractP5.hdrSizeInBytes - TSubscription.hdrSizeInBytesLegacy - TReqHdr.hdrSizeInBytes);
        initCounters();
        break;    
      default:
        break;    
    }
  }
  private void initCounters()
  {
    //TODO: set the mtu from underneath (either a p5 or p6)
    mtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubscription.hdrSizeInBytes - TReqHdr.hdrSizeInBytes);
    if (dFormat != TFormat.CF_NULL) mtu -= (mtu % TFormat.formatSizeOf(dFormat));
    numblks = dSizeInBytes / mtu;
    if (dSizeInBytes == 0 || (dSizeInBytes % mtu) > 0) numblks += 1;
    blkid = 1;
    blknum = 1;
    blkcnt = 0;
  }
  public boolean equals(TDataType dt)
  {
    return equals(true,dt);
  }
  public boolean equals(boolean compareContents,TDataType dt)
  {
    if (dt == null) 
    {
      return dFormat == TFormat.CF_NULL ? true : false;
    }
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return true;
    if (dt.dFormat == TFormat.CF_DEFAULT && acquireDefaultFormat) return true; 
    if (dt.dFormat != dFormat) return false;
    if (dt.dArrayLength == 0) return false;
    if (dt.dFormat != TFormat.CF_STRING &&
        dt.dArrayLength != dArrayLength) return false;
    if (dt.dBuffers == null || dBuffers == null) return false;
    if (dt.getTag().compareTo(getTag()) != 0) return false;
    if (compareContents) 
    {
      for (int i = 0; i < dt.dBuffers.length; i++)
      {
        if (dBufferSegSize * i > dArrayLength) break;
        if (dBuffers[i] == null || dt.dBuffers[i] == null) return false;
        if (!Arrays.equals(dt.dBuffers[i].toByteArray(), dBuffers[i].toByteArray())) return false;
      }
    }
    return true;
  }
  public int copy(TDataType dt)
  {
    if (dt == null) return TErrorList.argument_list_error;
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return 0;
    if (dt.dFormat != dFormat) return TErrorList.illegal_format;
    if (dt.dFormat == TFormat.CF_STRING)
    {
      dSizeInBytes = dt.dSizeInBytes;
    }
    else
    {
      if (dt.dArrayLength > dArrayLength) return TErrorList.dimension_error;
    }
    if (dt.dBuffers == null || dBuffers == null) return TErrorList.un_allocated;
    dCompletionLength = dt.dCompletionLength;
    if (dBuffers == dt.dBuffers) return 0; // no need to do anything
    try
    {
      for (int i = 0; i < dt.dBuffers.length; i++)
      {
        dBuffers[i].reset();
        dt.dBuffers[i].writeTo(dBuffers[i]);
      }
    }
    catch (Exception e)
    {
    }
    return 0;
  }
  /**
   * @brief Sets the delimiter used in a string representation of an array
   */
  public void setArrayDelimiter(String delimiter)
  {
    if (arrayDelimiter.compareTo(delimiter) == 0) return;
    arrayDelimiter = new String(delimiter);
  }
  /**
   * @brief Get the delimiter used in a string representation of an array
   * @return the delimiter used in a string representation of an array
   */
  public String getArrayDelimiter()
  {
    return arrayDelimiter;
  }
  /**
   * @brief NULL element
   */
  public TDataType()
  {
    dFormat = TFormat.CF_NULL;
    dArrayLength = 0;
    dSizeInBytes = 0;
    initCounters();
  }
  private void makeDataObject(int dataSize, short dataFormat)
  {
    dFormat = dataFormat >= 0 ? dataFormat : (short) (256 + dataFormat);
    switch (dFormat%256)
    {
      case TFormat.CF_DOUBLE:
        hDataObject = new double[dataSize];
        break;
      case TFormat.CF_FLOAT:
        hDataObject = new float[dataSize];
        break;
      case TFormat.CF_BITFIELD16:
      case TFormat.CF_SHORT:
        hDataObject = new short[dataSize];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_BITFIELD32:
      case TFormat.CF_LONG:
        hDataObject = new int[dataSize];
        break;
     case TFormat.CF_DLONG:
        hDataObject = new long[dataSize];
        break;   
      case TFormat.CF_NAME16:
        hDataObject = new NAME16[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16[])hDataObject)[i] = new NAME16();
        break;
      case TFormat.CF_NAME32:
        hDataObject = new NAME32[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32[])hDataObject)[i] = new NAME32();
        break;
      case TFormat.CF_NAME48:
        hDataObject = new NAME48[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME48[])hDataObject)[i] = new NAME48();
        break;
      case TFormat.CF_NAME64:
        hDataObject = new NAME64[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64[])hDataObject)[i] = new NAME64();
        break;
      case TFormat.CF_NAME8:
        hDataObject = new NAME8[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME8[])hDataObject)[i] = new NAME8();
        break;
      case TFormat.CF_HISTORY:
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
      case TFormat.CF_BITFIELD8:
        hDataObject = new byte[dataSize];
        break;
      case TFormat.CF_DBLINT:
        hDataObject = new DBLINT[dataSize];
        break;
      case TFormat.CF_DBLDBL:
        hDataObject = new DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((DBLDBL[])hDataObject)[i] = new DBLDBL();
        break;
      case TFormat.CF_DBLDBLDBL:
        hDataObject = new DBLDBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((DBLDBLDBL[])hDataObject)[i] = new DBLDBLDBL();
        break;
      case TFormat.CF_FLTINT:
        hDataObject = new FLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINT[])hDataObject)[i] = new FLTINT();
        break;
      case TFormat.CF_LNGINT:
        hDataObject = new INTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINT[])hDataObject)[i] = new INTINT();
        break;
      case TFormat.CF_NAME16I:
        hDataObject = new NAME16I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16I[])hDataObject)[i] = new NAME16I();
        break;
      case TFormat.CF_NAME16FI:
        hDataObject = new NAME16FI[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16FI[])hDataObject)[i] = new NAME16FI();
        break;
      case TFormat.CF_NAME16II:
        hDataObject = new NAME16II[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16II[])hDataObject)[i] = new NAME16II();
        break;
      case TFormat.CF_NAME32I:
        hDataObject = new NAME32I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32I[])hDataObject)[i] = new NAME32I();
        break;
      case TFormat.CF_NAME48I:
        hDataObject = new NAME48I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME48I[])hDataObject)[i] = new NAME48I();
        break;
      case TFormat.CF_NAME64I:
        hDataObject = new NAME64I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64I[])hDataObject)[i] = new NAME64I();
        break;
      case TFormat.CF_NAME16DBLDBL:
        hDataObject = new NAME16DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16DBLDBL[])hDataObject)[i] = new NAME16DBLDBL();
        break;
      case TFormat.CF_NAME32DBLDBL:
        hDataObject = new NAME32DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32DBLDBL[])hDataObject)[i] = new NAME32DBLDBL();
        break;
      case TFormat.CF_NAME64DBL:
        hDataObject = new NAME64DBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBL[])hDataObject)[i] = new NAME64DBL();
        break;
      case TFormat.CF_NAME64DBLDBL:
        hDataObject = new NAME64DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBLDBL[])hDataObject)[i] = new NAME64DBLDBL();
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        hDataObject = new NAME64DBLDBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBLDBLDBL[])hDataObject)[i] = new NAME64DBLDBLDBL();
        break;
      case TFormat.CF_SPECTRUM:
        hDataObject = new SPECTRUM(dataSize);
        break;
      case TFormat.CF_INTFLTINT:
        hDataObject = new INTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTFLTINT[])hDataObject)[i] = new INTFLTINT();
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        hDataObject = new char[dataSize];
        break;
      case TFormat.CF_STRING:
        hDataObject = new String[dataSize];
        break;
      case TFormat.CF_USTRING:
        hDataObject = new USTRING[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((USTRING[])hDataObject)[i] = new USTRING();
        break;
      case TFormat.CF_UNAME:
        hDataObject = new UNAME[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((UNAME[])hDataObject)[i] = new UNAME();
        break;
      case TFormat.CF_ADDRESS:
        hDataObject = new INTINTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTINTINT[])hDataObject)[i] = new INTINTINTINT();
        break;
      case TFormat.CF_INTINTINT:
        hDataObject = new INTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTINT[])hDataObject)[i] = new INTINTINT();
        break;
      case TFormat.CF_POINT:
        hDataObject = new FLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTFLT[])hDataObject)[i] = new FLTFLT();
        break;
      case TFormat.CF_FILTER:
        hDataObject = new INTFLTFLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTFLTFLTFLT[])hDataObject)[i] = new INTFLTFLTFLT();
        break;
      case TFormat.CF_FLTFLTINT:
        hDataObject = new FLTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTFLTINT[])hDataObject)[i] = new FLTFLTINT();
        break;
      case TFormat.CF_FLTINTINT:
        hDataObject = new FLTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINTINT[])hDataObject)[i] = new FLTINTINT();
        break;
      case TFormat.CF_FWINDOW:
        hDataObject = new INTINTFLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTFLTFLT[])hDataObject)[i] = new INTINTFLTFLT();
        break;
    }    
  }
  /**
   * @brief An unbound DataType object of the specified size and format.
   * @param dataSize is the array size of the given format.
   * @param dataFormat is the tine data type of the object to create.
   * @note A call to getData() must be made in order to retrieve the incoming
   *       data into the application array parameters.
   */
  public TDataType(int dataSize, short dataFormat)
  {
    dFormat = dataFormat >= 0 ? dataFormat : (short) (256 + dataFormat);
    switch (dFormat%256)
    {
      case TFormat.CF_DEFAULT:
        acquireDefaultFormat = true;
        dSizeInBytes = dataSize;
        dArrayLength = dataSize;
        blknum = 1;
        blkcnt = 0;
        return; // no allocation
      default:
        makeDataObject(dataSize, dataFormat);
        break;
    }
    dArrayLength = dataSize;
    dSizeInBytes = dataSize * TFormat.formatSizeOf(dFormat) + TFormat.getFormatHeaderSize(dFormat);
    if (dFormat == TFormat.CF_HISTORY) 
    { 
      int hsiz = TFormat.getCarriedFormatSize(dTag);
      if (hsiz > 0) 
      {
        hsiz += TFormat.getHistoryHeaderSize();
        dArrayLength = dataSize / hsiz;
      }
    }
    initCounters();
    dBuffers = new ByteArrayOutputStream[numblks * 3 / 2]; // room for NIOS calls
    for (int n = 0, i = 0; n < numblks; n++)
    {
      dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
      DataOutputStream ds = new DataOutputStream(dBuffers[n]);
      try
      {
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++) ds.write(0);
      }
      catch (Exception e)
      {
      }
    }
  }
  /**
   * @brief Returns the associated data timestamp as traditional UNIX timestamp
   *        (seconds since 1.1.1970).
   */
  public int getTimeStamp()
  {
    return timestamp;
  }
  /**
   * @brief Returns the associated data timestamp as JAVA timestamp
   *        (milliseconds since 1.1.1970).
   */
  public long getDataTimeStamp()
  {
    if (dTimestamp > 0) return dTimestamp;
    return (long) timestamp * 1000 + (long) timestampMSEC;
  }
  /**
   * @brief Sets the associated data timestamp from a JAVA timestamp
   *        (milliseconds since 1.1.1970).
   */
  public void setDataTimeStamp(long TimeStamp)
  {
    dTimestamp = TimeStamp;
    timestamp = (int) (TimeStamp / 1000);
    timestampMSEC = (int) (TimeStamp % 1000);
  }
  /**
   * @brief Sets the associated data timestamp from a TINE timestamp
   *        (seconds since 1.1.1970 + fractional seconds).
   */
  public void setDataTimeStamp(double TimeStamp)
  {
    long t = ((long)TimeStamp)*1000 + ((long)(TimeStamp * 1000) % 1000);
    setDataTimeStamp(t);
  }
  /**
   * @brief Returns the associated data completion length
   */
  public int getLastDataSize()
  {
    return dCompletionLength;
  }
  /**
   * @brief A subset of an array of booleans. This corresponds to tine format
   *        CF_BOOLEAN
   * @param data
   *          is a reference to the boolean array used to create and bind to the
   *          TDataType object
   */
  public TDataType(byte[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of bytes
   * @param data
   *          is a reference to the byte array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(byte[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(byte[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_BYTE;
      dArrayLength = length;
      dSizeInBytes = length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(data[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of 'structs' with the given tag or untagged structs if
   *        tag.length == 0 User-defined structures are allowed in tine and not
   *        too difficult to use. It is strongly encouraged to make use of
   *        'tagged' structures whereby the server and client both register the
   *        structure contents with the tine engine and apply a tag. Only then
   *        can issues of 'byte-swapping' and structure boundaries be handled
   *        correctly. As java does not have structures, the associated class is
   *        responsible for providing a toStruct() method which returns a byte
   *        stream containing the contents of the desired structure.
   * @param data
   *          is a reference to the byte array stream containing the 'struct'
   *          used to create and bind to the TDataType object.
   * @param tag
   *          is the structure tag associated with the 'struct.'
   */
  public TDataType(byte[] data, String tag)
  {
    dTag = tag;
    try
    {
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = data.length;
      dSizeInBytes = data.length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A Tagged Structure User-defined structures are allowed in tine
   *        and not too difficult to use. It is strongly encouraged to make use
   *        of 'tagged' structures whereby the server and client both register
   *        the structure contents with the tine engine and apply a tag. Only
   *        then can issues of 'byte-swapping' and structure boundaries be
   *        handled correctly. As java does not have structures, the associated
   *        class is responsible for providing a toStruct() method which returns
   *        a byte stream containing the contents of the desired structure.
   * @param data
   *          is a reference to a TTaggedStructure object.
   */
  public TDataType(TTaggedStructure data)
  {
    dTag = data.getName();
    try
    {
      byte[] hb = data.toByteArray();
      dFormat = TFormat.CF_STRUCT;
      dSizeInBytes = data.getSizeInBytes();      
      dArrayLength = dSizeInBytes; // ?? 
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(hb[i]);
        ds.close();
      }
      hDataObject = data.getByteArray();
      hStructObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A array of Tagged Structures User-defined structures are
   *        allowed in tine and not too difficult to use. It is strongly
   *        encouraged to make use of 'tagged' structures whereby the server and
   *        client both register the structure contents with the tine engine and
   *        apply a tag. Only then can issues of 'byte-swapping' and structure
   *        boundaries be handled correctly. As java does not have structures,
   *        the associated class is responsible for providing a toStruct()
   *        method which returns a byte stream containing the contents of the
   *        desired structure.
   * @param data
   *          is a reference to an array of TTaggedStructure object.
   */
  public TDataType(TTaggedStructure[] data)
  {
    if (data[0] == null)
    {
      MsgLog.log("TDataType","Tagged Structure data type unallocated !",TErrorList.invalid_structure_tag,null,0);
      return;
    }
    for (int k = 1; k < data.length; k++)
      if (data[k] == null) data[k] = (TTaggedStructure) data[0].clone();
    dTag = data[0].getName();
    byte[] hbytes = new byte[data[0].getSizeInBytes() * data.length];
    try
    {
      int dsiz = data[0].getSizeInBytes();
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = dsiz * data.length;      
      dSizeInBytes = dsiz * data.length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int k = 0; k < data.length; k++)
      {
        System.arraycopy(data[k].toByteArray(), 0, hbytes, k * dsiz, dsiz);
      }
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(hbytes[i]);
        ds.close();
      }
      hDataObject = hbytes;
      hStructObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of shorts (2-byte integers) This corresponds to tine format
   *        CF_SHORT
   * @param data
   *          is a reference to the short array used to create and bind to the
   *          TDataType object
   */
  public TDataType(short[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of shorts (2-byte integers) This corresponds to
   *        tine format CF_SHORT
   * @param data
   *          is a reference to the short array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(short[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(short[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_SHORT;
      dArrayLength = length;
      dSizeInBytes = length * 2;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 2; i++)
          ds.writeShort(Swap.Short(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A single short value This method is only useful for quick creation
   *        of data 'down' objects (i.e. data sets sent to the server). As there
   *        is no reference to the argument passed, it cannot be used to bind
   *        returning data to the object passed as the argument.
   * @param data
   *          is a single short value used to create the TDataType object.
   */
  public TDataType(short data)
  {
    try
    {
      dFormat = TFormat.CF_SHORT;
      dArrayLength = 1;
      dSizeInBytes = 2;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffers[0] = new ByteArrayOutputStream(dBufferSegSize);
      DataOutputStream ds = new DataOutputStream(dBuffers[0]);
      ds.writeShort(data);
      byte[] b = dBuffers[0].toByteArray();
      Swap.Bytes(b, 2);
      dBuffers[0].reset();
      ds.write(b, 0, b.length);
      ds.close();
      hDataObject = null;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of ints (4-byte integers) This corresponds to tine format
   *        CF_LONG
   * @param data
   *          is a reference to the int array used to create and bind to the
   *          TDataType object
   */
  public TDataType(int[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of ints (4-byte integers) This corresponds to
   *        tine format CF_LONG
   * @param data
   *          is a reference to the int array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(int[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(int[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_LONG;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.writeInt(Swap.Long(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A subset of an array of booleans. This corresponds to tine format
   *        CF_BOOLEAN
   * @param data
   *          is a reference to the boolean array used to create and bind to the
   *          TDataType object
   */
  public TDataType(boolean[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of booleans. This corresponds to tine format
   *        CF_BOOLEAN
   * @param data
   *          is a reference to the boolean array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(boolean[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(boolean[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_BOOLEAN;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      int[] idata = new int[data.length];
      for (int i=0; i < data.length; i++) idata[i] = data[i] ? 1 : 0;
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.writeInt(Swap.Long(idata[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of floats This corresponds to tine format CF_FLOAT
   * @param data
   *          is a reference to the float array used to create and bind to the
   *          TDataType object
   */
  public TDataType(float[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of floats This corresponds to tine format
   *        CF_FLOAT
   * @param data
   *          is a reference to the float array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(float[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(float[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_FLOAT;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks * 3 / 2]; // NIOS has mtu =
                                                              // 1024
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.write(Swap.Float((data[i])));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of tine data objects (e.g. FLTINTs, FLTFLTs, NAME16s, etc.)
   * can be bound to a tine data type object.  Tine data type objects are used
   * in data links between client and server.
   * @param data
   *          is a reference to the data objects used to create and bind to
   *          the TDataType object
   */
  public TDataType(TCompoundDataObject[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief An array of tine data objects (e.g. FLTINTs, FLTFLTs, NAME16s, etc.)
   * can be bound to a tine data type object.  Tine data type objects are used
   * in data links between client and server.
   * @param data
   *          is a reference to the data objects used to create and bind to
   *          the TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(TCompoundDataObject[] data, int length)
  {
    initTDataType(data, length);
  }
  /**
   * @brief An array of FLTINTs (float + integer pairs) This data type is
   *        frequently used to retrieve archived parameters (float value +
   *        timestamp)
   * @param data
   *          is a reference to the FLTINT objects used to create and bind to
   *          the TDataType object
   */
  private void initTDataType(TCompoundDataObject[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      if (data instanceof TBitfield[])
      { // Bitfields are tagged !
        if (((TBitfield[])data)[0] == null) return;
        dTag = ((TBitfield[])data)[0].getName(); 
      }
      Object tdo = data[0] != null ? 
        data[0] : data.getClass().getComponentType().newInstance();
      dFormat = ((TCompoundDataObject) tdo).getFormat();
      int elementSize = TFormat.formatSizeOf(dFormat);
      dArrayLength = length;
      dSizeInBytes = length * elementSize;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < length && i < (n + 1) * mtu / elementSize; i++)
          setObjectContents(ds, data, i); //
        ds.close();
      }
      setObjectHandle(data); //
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }        
    catch (Exception e)
    {
      e.printStackTrace();      
    }
  }
  private void initTDataType(Object data, int length)
  {
    try
    {
      if (length > getObjectSize(data)) length = getObjectSize(data);
      dFormat = getObjectFormat(data);
      int elementSize = TFormat.formatSizeOf(dFormat);
      dArrayLength = length;
      dSizeInBytes = length * elementSize;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < length && i < (n + 1) * mtu / elementSize; i++)
          setObjectContents(ds, data, i);
        ds.close();
      }
      setObjectHandle(data);
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A SPECTRUM object. This corresponds to tine format CF_SPECTRUM
   * @param data
   *          is a reference to the SPECTRUM object used to create and bind to the
   *          TDataType object
   */
  public TDataType(SPECTRUM data)
  {
    try
    {
      dFormat = TFormat.CF_SPECTRUM;
      dArrayLength = data.length;
      dSizeInBytes = data.length * 4 + 96;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An IMAGE object. This corresponds to tine format CF_IMAGE
   * @param data
   *          is a reference to the IMAGE object used to create and bind to the
   *          TDataType object
   */
  public TDataType(IMAGE data)
  {
    try
    {
      dFormat = TFormat.CF_IMAGE;
      dArrayLength = data.getImageFrameSize();
      dSizeInBytes = data.getSizeInBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  private void repackageDataForOutput()
  {
    if (dFormat == TFormat.CF_NULL) return;
    try
    {
      int fsiz = TFormat.formatSizeOf(dFormat);
      initCounters();
      for (int i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        dBuffers[n].reset();
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < m + dBufferSegSize / fsiz; i++)
        {
          setObjectContents(ds, getDataObject(), i);
          if (dFormat == TFormat.CF_TEXT ||
              dFormat == TFormat.CF_XML ||
              dFormat == TFormat.CF_STRING) i = dArrayLength;
        }
        ds.close();
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief An array of doubles. This corresponds to tine format CF_DOUBLE
   * @param data
   *          is a reference to the double array used to create and bind to the
   *          TDataType object
   */
  public TDataType(double[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of doubles. This corresponds to tine format
   *        CF_DOUBLE
   * @param data
   *          is a reference to the double array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(double[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(double[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_DOUBLE;
      dArrayLength = length;
      dSizeInBytes = length * 8;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 8; i++) 
          ds.write(Swap.Double(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  
  /**
   * @brief An array of (Java) longs. This corresponds to tine format CF_DLONG or CF_INT64
   * @param data
   *          is a reference to the long array used to create and bind to the
   *          TDataType object
   */
  public TDataType(long[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of (java) longs. This corresponds to tine format
   *        CF_DLONG, CF_INT64
   * @param data
   *          is a reference to the long array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(long[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(long[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_DLONG;
      dArrayLength = length;
      dSizeInBytes = length * 8;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 8; i++) 
          ds.writeLong(Swap.DLong(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  
  public void update(byte[] data, int num, int id)
  {
    update(data, 0, data.length, num, id);
  }
  public void update(byte[] data, int off, int len, int num, int id)
  {
    if (TLinkFactory.debugLevel > 1)
      DbgLog.log("TDataType","update: blk "+num+" id "+id+" ("+len+" bytes)");
    blknum = num;
    // check num vs numblks, and id for integrity
    if (id != blkid)
    {
      blkid = id;
      blksin = 0;
      bytesin = 0;
    }
    if (data.length == 0)
    {
      blksin = 1;
      return;
    }
    if (dBuffers == null)
    {
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
      }
    }
    try
    { // possible small mtu server (NIOS) ...
      if (dBuffers[num - 1] == null) dBuffers[num - 1] = new ByteArrayOutputStream(dBufferSegSize);
      dBuffers[num - 1].reset();
      DataOutputStream ds = new DataOutputStream(dBuffers[num - 1]);
      ds.write(data, off, len);
      ds.close();
      blksin++;
      if (TLinkFactory.debugLevel > 1)
      {
        String tag = blksin == numblks ? " transfer complete" : " missing "+(numblks-blksin)+" blocks";
        DbgLog.log("TDataType","update: blks received "+blksin+tag);
      }
      hasBeenUpdated = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public int getData(byte[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(byte[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < len && i < m + b.length; i++) data[i+offset] = ds.readByte();
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      if (dFormat == TFormat.CF_STRUCT)
      {
        TStructDescription tsd = TStructRegistry.get(dTag);
        if (tsd != null  && tsd.getSize() != 0) 
        {
          dCompletionLength /= tsd.getSize();
        }
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int getData(short[] data)
  {
    return getData(data,data != null ? data.length : 0, 0);
  }
  public int getData(short[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readFloat();
            break;
          case TFormat.CF_BITFIELD64:
          case TFormat.CF_DLONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readLong();
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_BITFIELD32:
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (short) ds.readInt();
            break;
          case TFormat.CF_BITFIELD16:
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(short data)
  {
    short[] d = new short[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(short[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(short[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
   if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
    case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)
          ((long[])hDataObject)[i] = (long) data[i+offset];
        break;        
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public void applyBitField()
  {
    if (dBitField == null) return;
    if (hDataObject instanceof TBitfield[]) return;
    switch (dFormat)
    {
      case TFormat.CF_BITFIELD8:
        if (hDataObject instanceof byte[])
        {
          for (int i=0; i<dCompletionLength; i++)
          {
            ((byte[])hDataObject)[i] = 
              (byte)dBitField.getField(dField, ((byte[])hDataObject)[i]);
          }
        }
        break;
      case TFormat.CF_BITFIELD16:
        if (hDataObject instanceof short[])
        {
          for (int i=0; i<dCompletionLength; i++)
          {
            ((short[])hDataObject)[i] = 
              (short)dBitField.getField(dField, ((short[])hDataObject)[i]);
          }
        }
      case TFormat.CF_BITFIELD32:
        if (hDataObject instanceof int[])
        {
          for (int i=0; i<dCompletionLength; i++)
          {
            ((int[])hDataObject)[i] = 
              (int)dBitField.getField(dField, ((int[])hDataObject)[i]);
          }
        }
      case TFormat.CF_BITFIELD64:
        if (hDataObject instanceof long[])
        {
          for (int i=0; i<dCompletionLength; i++)
          {
            //TODO: allow int64 bitfields !
            ((long[])hDataObject)[i] = 
              (long)dBitField.getField(dField, (int)((long[])hDataObject)[i]);
          }
        }
    } 
    return;
  }
  public int getData()
  {
    switch (dFormat)
    {
      case TFormat.CF_STRUCT:
        if (hDataObject != null) getData(hDataObject);
        if (hStructObject != null && hStructObject instanceof TTaggedStructure)
        {
          ((TTaggedStructure)hStructObject).toStruct((byte[])hDataObject);
        }
        else if (hStructObject instanceof TTaggedStructure[])
        {
          TTaggedStructure[] tts = (TTaggedStructure[])hStructObject;
          int dsiz = tts[0].getSizeInBytes();
          byte[] hb = new byte[dsiz];
          for (int i = 0; i < tts.length; i++)
          {
            System.arraycopy((byte[])hDataObject, i * dsiz, hb, 0, dsiz);
            tts[i].toStruct(hb);
          }
        }
        break;
      default:
        getData(getDataObject());
        break;
    }
    return 0;
  }
  public int getData(boolean[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(boolean[] data,int length,int offset)
  {
    if (data == null || data.length == 0) return TErrorList.argument_list_error;
    int idata[] = new int[data.length];
    int cc = getData(idata,length,offset);
    if (cc == 0)
    {
      for (int i=0; i<data.length; i++) data[i] = idata[i] == 0 ? false : true; 
    }
    return cc;
  }
  public int getData(int[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(int[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (int) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (int) ds.readFloat();
            break;
          case TFormat.CF_BITFIELD32:
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
          case TFormat.CF_BITFIELD64:
          case TFormat.CF_DLONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (int) ds.readLong();
            break;  
          case TFormat.CF_BITFIELD16:
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(boolean data)
  {
    boolean[] d = new boolean[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(boolean[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(boolean[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    int idata[] = new int[data.length];
    for (int i=0; i<data.length; i++) idata[i] = data[i] ? 1 : 0;
    return putData(idata,length,offset);
  }  
  public int putData(int data)
  {
    int[] d = new int[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(int[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(int[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (long) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int getData(float[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(float[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (float) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readFloat();
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
         case TFormat.CF_DLONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (float)ds.readLong();
            break;            
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int getData(TTaggedStructure[] data)
  {
    return getData(data,data != null ? data.length : 0, 0);
  }
  public int getData(TTaggedStructure[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data[0].getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    getData(hDataObject);
    int dsiz = data[0].getSizeInBytes();
    byte[] hb = new byte[dsiz];
    for (int i = 0; i < data.length; i++)
    { // copy the byte stream into the referenced data object
      System.arraycopy(hDataObject, i*dsiz, hb, 0, dsiz);
      data[i].toStruct(hb);
    }
    dCompletionLength /= dsiz;
    return 0;
  }  
  public int getData(TCompoundDataObject[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(TCompoundDataObject[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (data instanceof TBitfield[])
    {
      if (((TBitfield[])data)[0] == null) return TErrorList.not_initialized;
    }
    if (data instanceof IMAGE[])
    {
      return getData(((IMAGE[])data)[0]);
    }
    if (data instanceof HISTORY[])
    {
      if (dFormat != TFormat.CF_HISTORY) return TErrorList.illegal_format;
    }
    if (data instanceof SPECTRUM[])
    {
      return getData(((SPECTRUM[])data)[0]);
    }
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of this buffered array
    if (len > length) len = length; // can't get more than is there !
    int to_size = 0;  // size in bytes of the TCompoundDataObject
    int from_size = TFormat.formatSizeOf(dFormat);
    int cpy_size = from_size; // copy size
    try
    {
      int i, n, m, j;
      for (i = 0, n = 0, m = 0, j = 0; n < numblks; n++, m = i)
      {
        if (i+offset >= data.length) 
        {
          MsgLog.log("TDataType","data overflow disgarded",TErrorList.buffer_too_small,null,1);
          break;
        }
        if (data[i+offset] == null) makeNewCompoundDataObject(data,i+offset);
        if (to_size == 0) 
        {
          to_size = data[i+offset].getSizeInBytes();
          if (dFormat == TFormat.CF_TEXT)
          { // this is a TEXT object being sent a NAMExx
            to_size = 1;
          }
          if (to_size < cpy_size) cpy_size = to_size;
        }
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray(); // this object's byte stream
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
         for (; i < len && i < m + b.length /from_size; i++)
        {
          if (data[i+offset] == null) makeNewCompoundDataObject(data,i+offset);
          if (data[i+offset].getByteArray() == null) data[i+offset].toByteArray();
          byte[] ba = data[i+offset].getByteArray();
          for (j = 0; j <cpy_size; j++) ba[j] = ds.readByte();
          data[i+offset].toStruct();
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }  
  public int getData(SPECTRUM data)
  {
    if (dFormat != TFormat.CF_SPECTRUM) return TErrorList.illegal_format;
    if (dArrayLength > data.length) return TErrorList.dimension_error;
    int dSize = dArrayLength + 24; // effective data size for SPECTRUM
    try
    {
      int i, n, m;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < dSize * 4 && i < m + b.length; i++)
        {
          data.hByteArray[i] = ds.readByte();
        }
        data.toStruct();
        ds.close();
        bs.close();
      }
      dCompletionLength = i - 24;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int getData(IMAGE data)
  {
    if (dFormat != TFormat.CF_IMAGE) return TErrorList.illegal_format;
    if (dArrayLength > data.getImageFrameSize()) return TErrorList.dimension_error;
    try
    {
      int i, n, len = 0, off = 0;
      for (i = 0, n = 0; n < numblks; n++)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        len = b.length;
        if (n == 0)
        { // first block contains the image header
          data.setHeaderBytes(ds);
          off = 0;
          len -= TFormat.getFormatHeaderSize(TFormat.CF_IMAGE);
        }
        data.setImageBytes(ds, off, len);
        off += len;
        data.toStruct();
        ds.close();
        bs.close();
      }
      dCompletionLength = i - 24;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(float data)
  {
    float[] d = new float[1];
    d[0] = data;
    return putData(d);
  }
  public int putData()
  {
    if (isLocked) return TErrorList.not_allowed;
    repackageDataForOutput();
    return 0;    
  }
  public int putData(float[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(float[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)
          ((long[])hDataObject)[i] = (long) data[i+offset];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(FLTINT data)
  {
    FLTINT[] d = new FLTINT[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(FLTINT[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(FLTINT[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_LNGINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((INTINT[])hDataObject)[i] != null)
            ((INTINT[])hDataObject)[i].setValues((int) data[i+offset].fval, data[i+offset].ival);
          else
            ((INTINT[])hDataObject)[i] = 
              new INTINT((int) data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          if (((FLTINT[])hDataObject)[i] != null)
            ((FLTINT[])hDataObject)[i].setValues(data[i+offset].fval, data[i+offset].ival);
          else
            ((FLTINT[])hDataObject)[i] = data[i+offset];
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBL[])hDataObject)[i] != null)
            ((DBLDBL[])hDataObject)[i].setValues(data[i+offset].fval, data[i+offset].ival);
          else
            ((DBLDBL[])hDataObject)[i] = new DBLDBL(data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_DBLDBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBLDBL[])hDataObject)[i] != null)
            ((DBLDBLDBL[])hDataObject)[i].setValues(data[i+offset].fval, 0, data[i+offset].ival);
          else
            ((DBLDBLDBL[])hDataObject)[i] = new DBLDBLDBL(data[i+offset].fval, 0, data[i+offset].ival);
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = data[i+offset].fval;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].fval;
        }
        break;
      case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((long[])hDataObject)[i] = data[i+offset].ival;
        }
        break;  
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = data[i+offset].ival;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(NAME16FI[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(NAME16FI[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_NAME16FI:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16FI[])hDataObject)[i] != null)
            ((NAME16FI[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME16FI[])hDataObject)[i] = data[i+offset];
        }
        break;
      case TFormat.CF_NAME16II:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16II[])hDataObject)[i] != null)
            ((NAME16II[])hDataObject)[i].setValues(data[i+offset].name,(int)data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME16II[])hDataObject)[i] = 
              new NAME16II(data[i+offset].name, (int)data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16DBLDBL[])hDataObject)[i] != null)
            ((NAME16DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME16DBLDBL[])hDataObject)[i] =
              new NAME16DBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME32DBLDBL[])hDataObject)[i] != null)
            ((NAME32DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME32DBLDBL[])hDataObject)[i] = 
              new NAME32DBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME64DBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBL[])hDataObject)[i] != null)
            ((NAME64DBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval);
          else
            ((NAME64DBL[])hDataObject)[i] = 
              new NAME64DBL(data[i+offset].name, data[i+offset].fval);
        }
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBLDBL[])hDataObject)[i] != null)
            ((NAME64DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival);
          else
            ((NAME64DBLDBL[])hDataObject)[i] = 
              new NAME64DBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBLDBLDBL[])hDataObject)[i] != null)
            ((NAME64DBLDBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].fval,data[i+offset].ival,0);
          else
            ((NAME64DBLDBLDBL[])hDataObject)[i] = 
              new NAME64DBLDBLDBL(data[i+offset].name, data[i+offset].fval, data[i+offset].ival);
        }
        break;
      case TFormat.CF_USTRING:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i] == null) continue;
          if (((USTRING[])hDataObject)[i] != null)
            ((USTRING[])hDataObject)[i].setValues(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
          else
            ((USTRING[])hDataObject)[i] = 
              new USTRING(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
        }
        break;
      case TFormat.CF_UNAME:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((UNAME[])hDataObject)[i] != null)
            ((UNAME[])hDataObject)[i].setValues(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
          else
            ((UNAME[])hDataObject)[i] = 
              new UNAME(data[i+offset].ival,data[i+offset].fval,0,0,data[i+offset].name);
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(NAME32DBLDBL[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(NAME32DBLDBL[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_NAME16II:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16II[])hDataObject)[i] != null)
            ((NAME16II[])hDataObject)[i].setValues(data[i+offset].name,(int)data[i+offset].d1val,(int)data[i+offset].d2val);
          else
            ((NAME16II[])hDataObject)[i] = 
              new NAME16II(data[i+offset].name, (int)data[i+offset].d1val, (int)data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME16FI:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16FI[])hDataObject)[i] != null)
            ((NAME16FI[])hDataObject)[i].setValues(data[i+offset].name,(float)data[i+offset].d1val,(int)data[i+offset].d2val);
          else
            ((NAME16FI[])hDataObject)[i] = 
              new NAME16FI(data[i].name, (float)data[i+offset].d1val, (int)data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME16DBLDBL[])hDataObject)[i] != null)
            ((NAME16DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val);
          else
            ((NAME16DBLDBL[])hDataObject)[i] = 
              new NAME16DBLDBL(data[i+offset].name, data[i+offset].d1val, data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((NAME32DBLDBL[])hDataObject)[i] != null)
            ((NAME32DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val);
          else
            ((NAME32DBLDBL[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_NAME64DBL:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBL[])hDataObject)[i] != null)
            ((NAME64DBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val);
          else
            ((NAME64DBL[])hDataObject)[i] = 
              new NAME64DBL(data[i+offset].name, data[i+offset].d1val);
        }
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBLDBL[])hDataObject)[i] != null)
            ((NAME64DBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val);
          else
            ((NAME64DBLDBL[])hDataObject)[i] = 
              new NAME64DBLDBL(data[i+offset].name, data[i+offset].d1val, data[i+offset].d2val);
        }
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((NAME64DBLDBLDBL[])hDataObject)[i] != null)
            ((NAME64DBLDBLDBL[])hDataObject)[i].setValues(data[i+offset].name,data[i+offset].d1val,data[i+offset].d2val,0);
          else
            ((NAME64DBLDBLDBL[])hDataObject)[i] = 
              new NAME64DBLDBLDBL(data[i+offset].name, data[i+offset].d1val, data[i+offset].d2val);
        }
        break;
      case TFormat.CF_USTRING:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((USTRING[])hDataObject)[i] != null)
            ((USTRING[])hDataObject)[i].setValues((int)data[i+offset].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
          else
            ((USTRING[])hDataObject)[i] = 
              new USTRING((int)data[i].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
        }
        break;
      case TFormat.CF_UNAME:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((UNAME[])hDataObject)[i] != null)
            ((UNAME[])hDataObject)[i].setValues((int)data[i+offset].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
          else
            ((UNAME[])hDataObject)[i] = 
              new UNAME((int)data[i+offset].d2val,(float)data[i+offset].d1val,0,0,data[i+offset].name);
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(INTINT data)
  {
    INTINT[] d = new INTINT[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(INTINT[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(INTINT[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_LNGINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((INTINT[])hDataObject)[i] != null)
            ((INTINT[])hDataObject)[i].setValues(data[i+offset].i1val, data[i+offset].i2val);
          else
            ((INTINT[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((FLTINT[])hDataObject)[i] != null)
            ((FLTINT[])hDataObject)[i].setValues(data[i+offset].i1val, data[i+offset].i2val);
          else
            ((FLTINT[])hDataObject)[i] = new FLTINT(data[i+offset].i1val, data[i+offset].i2val);
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBL[])hDataObject)[i] != null)
            ((DBLDBL[])hDataObject)[i].setValues(data[i+offset].i1val, data[i+offset].i2val);
          else
            ((DBLDBL[])hDataObject)[i] = new DBLDBL(data[i+offset].i1val, data[i+offset].i2val);
        }
        break;
      case TFormat.CF_DBLDBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBLDBL[])hDataObject)[i] != null)
            ((DBLDBLDBL[])hDataObject)[i].setValues(data[i+offset].i1val, 0, data[i+offset].i2val);
          else
            ((DBLDBLDBL[])hDataObject)[i] = new DBLDBLDBL(data[i+offset].i1val, 0, data[i+offset].i2val);
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;
    case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((long[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;        
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = data[i+offset].i1val;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(DBLDBL data)
  {
    DBLDBL[] d = new DBLDBL[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(DBLDBL[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(DBLDBL[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_LNGINT:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((INTINT[])hDataObject)[i] != null)
            ((INTINT[])hDataObject)[i].setValues((int) data[i+offset].d1val, (int) data[i+offset].d2val);
          else
            ((INTINT[])hDataObject)[i] = 
              new INTINT((int) data[i+offset].d1val, (int) data[i+offset].d2val);
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((FLTINT[])hDataObject)[i] != null)
            ((FLTINT[])hDataObject)[i].setValues((float) data[i+offset].d1val, (int) data[i+offset].d2val);
          else
            ((FLTINT[])hDataObject)[i] = 
              new FLTINT((float) data[i+offset].d1val, (int) data[i+offset].d2val);
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBL[])hDataObject)[i] != null)
            ((DBLDBL[])hDataObject)[i].setValues(data[i+offset].d1val,data[i+offset].d2val);
          else
            ((DBLDBL[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = (float)data[i+offset].d1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].d1val;
        }
        break;
    case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((long[])hDataObject)[i] = (long)data[i+offset].d2val;
        }
        break;           
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = (int)data[i+offset].d2val;
        }
        break;      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(DBLDBLDBL data)
  {
    DBLDBLDBL[] d = new DBLDBLDBL[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(DBLDBLDBL[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(DBLDBLDBL[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_FLTINTINT:
        for (int i = 0; i < dArrayLength; i++) 
        {
          if (data[i+offset] == null) continue;
          if (((FLTINTINT[])hDataObject)[i] != null)
            ((FLTINTINT[])hDataObject)[i].setValues((float) data[i+offset].d1val, (int)data[i+offset].d2val, (int)data[i+offset].d3val);
          else
            ((FLTINTINT[])hDataObject)[i] = 
              new FLTINTINT((float) data[i+offset].d1val, (int) data[i+offset].d2val, (int)data[i+offset].d2val);
        }
        break;
      case TFormat.CF_DBLDBLDBL:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i+offset] == null) continue;
          if (((DBLDBLDBL[])hDataObject)[i] != null)
            ((DBLDBLDBL[])hDataObject)[i].setValues(data[i+offset].d1val,data[i+offset].d2val,data[i+offset].d3val);
          else
            ((DBLDBLDBL[])hDataObject)[i] = data[i+offset]; 
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((float[])hDataObject)[i] = (float)data[i+offset].d1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((double[])hDataObject)[i] = data[i+offset].d1val;
        }
        break;
    case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((long[])hDataObject)[i] = (long)data[i+offset].d3val;
        }
        break;   
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)        
        {
          if (data[i+offset] == null) continue;
          ((int[])hDataObject)[i] = (int)data[i+offset].d3val;
        }
        break;      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(String[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(String[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    int maxlen = dArrayLength;
    if (maxlen > length) maxlen = length;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != getObjectFormat(data)) return TErrorList.illegal_format;

    for (int i = 0; i < maxlen; i++) 
      ((String[])hDataObject)[i] = data[i+offset];
    if (dArrayLength > maxlen) dArrayLength = maxlen;
    dSizeInBytes = 0;
    for (int i=0; i<length; i++)
    { // first see how much string space we need (going out)
      dSizeInBytes += (data[i] != null) ? data[i].length() : 0; 
      dSizeInBytes++; // add terminating 0
    }
    repackageDataForOutput();
    return 0;    
  }
  public int putData(String data)
  {
    int len;
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || (len = data.length()) < 1) return 0;
    switch (dFormat)
    {
      case TFormat.CF_NAME8:
        ((NAME8[])hDataObject)[0] = new NAME8(data);
        break;
      case TFormat.CF_NAME16:
        ((NAME16[])hDataObject)[0] = new NAME16(data);
        break;
      case TFormat.CF_NAME32:
        ((NAME32[])hDataObject)[0] = new NAME32(data);
        break;
      case TFormat.CF_NAME48:
        ((NAME48[])hDataObject)[0] = new NAME48(data);
        break;
      case TFormat.CF_NAME64:
        ((NAME64[])hDataObject)[0] = new NAME64(data);
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        if (len < dArrayLength) dArrayLength = len;
        hDataObject = data.toCharArray();
        break;
      case TFormat.CF_BYTE:
        if (len < dArrayLength) dArrayLength = len;
        hDataObject = data.getBytes();
        break;
      case TFormat.CF_BOOLEAN:
        if (len < dArrayLength) dArrayLength = len;
        ((int[])hDataObject)[0] = (data.compareToIgnoreCase("true") == 0) ||
                                  (data.compareToIgnoreCase("yes") == 0) ? 1 : 0;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(TTaggedStructure data)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data.getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    data.toByteArray();
    System.arraycopy(data.getByteArray(), 0, (byte[])hDataObject, 0, data.getSizeInBytes());
    if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    { // this TDataType references another tagged structure !
      TTaggedStructure tts = (TTaggedStructure)hStructObject;
      tts.toStruct(data.getByteArray());
    }
    else if (hStructObject instanceof TTaggedStructure[])
    { // this TDataType references an array of tagged structures !
      TTaggedStructure tts[] = (TTaggedStructure[])hStructObject;
      tts[0].toStruct(data.getByteArray());
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(TTaggedStructure[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(TTaggedStructure[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data[0].getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    int tsiz = data[0].getSizeInBytes();
    int dsiz = getArrayLength();
    if (dsiz > length) dsiz = length;

    for (int i=0; i<dsiz; i++)
    {
      data[i+offset].toByteArray();
      System.arraycopy(data[i+offset].getByteArray(), 0, (byte[])hDataObject, i*tsiz, tsiz);
    }
    if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    {
      TTaggedStructure tts = (TTaggedStructure)hStructObject;
      tts.toStruct(data[offset].getByteArray());
    }
    else if (hStructObject instanceof TTaggedStructure[])
    {
      TTaggedStructure tts[] = (TTaggedStructure[])hStructObject;
      for (int i = 0; i < dsiz; i++)
      {
        tts[i].toStruct(data[i+offset].getByteArray());
      }
    }
    setArrayLength(dsiz);
    repackageDataForOutput();
    return 0;
  }
  public int putData(byte[] data)
  {
    return putData(data,data != null ? data.length : 0, 0);
  }
  public int putData(byte[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_TEXT:
        if (hDataObject instanceof char[])
        {
          for (int i=0; i<dArrayLength; i++) 
            ((char[])hDataObject)[i] = (char)data[i+offset];
          break;
        }
        if (hDataObject instanceof StringBuffer)
        {
          String s = new String(data);
          ((StringBuffer)hDataObject).append(s, offset, dArrayLength);
          break;
        }
        return TErrorList.illegal_format;
      case TFormat.CF_BITFIELD8:
      case TFormat.CF_BYTE:
        System.arraycopy(data, offset, (byte[])hDataObject, 0, dArrayLength);
        break;
      case TFormat.CF_BITFIELD16:
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_BITFIELD32:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(byte[] data, String tag)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (hDataObject != null)
    {
      if (tag != null && TStructRegistry.getSizeInBytes(tag) < 1) return TErrorList.invalid_structure_tag;
      if (data.length < dArrayLength) dArrayLength = data.length;
      System.arraycopy(data, 0, (byte[])hDataObject, 0, dArrayLength);
    }
    else if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    {
      TTaggedStructure tts = (TTaggedStructure)hStructObject;
      if (data.length != tts.getSizeInBytes()) return TErrorList.invalid_structure_tag;
      System.arraycopy(data, 0, tts.getByteArray(), 0, data.length);
    }
    else if (hStructObject instanceof TTaggedStructure[])
    {
      TTaggedStructure tts[] = (TTaggedStructure[])hStructObject;
      int dsiz = tts[0].getSizeInBytes();
      for (int i = 0; i < tts.length; i++)
      {
        System.arraycopy(data, i * dsiz, tts[i].getByteArray(), 0, dsiz);
      }
    }
    else
    {
      return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(Object data)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data instanceof byte[])
      return putData((byte[]) data);
    else if (data instanceof short[])
      return putData((short[]) data);
    else if (data instanceof int[])
      return putData((int[]) data);
    else if (data instanceof float[])
      return putData((float[]) data);
    else if (data instanceof double[])
      return putData((double[]) data);
    else if (data instanceof long[])
      return putData((long[]) data);
    else if (data instanceof String[])
      return putData((String[])data);
    else if (data instanceof TCompoundDataObject[])
    {
      int cc = putData((TCompoundDataObject[]) data);
      if (dFormat == TFormat.CF_HISTORY)
      { // reset the 'arraylength' to the length in bytes (that's how it's done!)
        dArrayLength *= (TFormat.getHistoryHeaderSize() + TFormat.getCarriedFormatSize(dTag)); 
      }      
      return cc;
    }
    if (dFormat != getObjectFormat(data)) return TErrorList.illegal_format;
    int len = getObjectSize(data);
    if (dFormat != TFormat.CF_TEXT &&
        dFormat != TFormat.CF_XML &&
        len < dArrayLength) dArrayLength = len;
    setObjectHandle(data);
    repackageDataForOutput();
    return 0;
  }
  public int putData(TCompoundDataObject data)
  {
    TCompoundDataObject[] d = new TCompoundDataObject[1];
    d[0] = data;
    return putData(d);    
  }  
  public int putData(TCompoundDataObject[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(TCompoundDataObject[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    int maxlen = dArrayLength;
    if (maxlen > length) maxlen = length;
    if (length <= 0) return TErrorList.argument_list_error;
    boolean formatsEqual = dFormat == getObjectFormat(data);

    if ((data instanceof NAME[] && hDataObject instanceof NAME[]) ||
        (data instanceof NAMEINT[] && hDataObject instanceof NAMEINT[]))
    {
      if (formatsEqual)
      {
        for (int i = 0; i < maxlen; i++) 
          ((TCompoundDataObject[])hDataObject)[i] = data[i+offset];
      }
      else
      {
        try
        {
          Class[] params = new Class[1];
          params[0] = String.class; 
          Object[] objs = new Object[1];
          for (int i = 0; i < maxlen; i++)
          {
            objs[0] = (data[i+offset].toString());
            ((TCompoundDataObject[])hDataObject)[i] = 
              (TCompoundDataObject)hDataObject.getClass().getComponentType().getConstructor(params).newInstance(objs); 
          }
        }
        catch (Exception e)
        {
          MsgLog.log("TDataType","cannot convert " + TFormat.toString(getObjectFormat(data)) + " to " + TFormat.toString(dFormat),TErrorList.illegal_format,e,0);
        }
      }
    }
    else if (data instanceof NAME[] && hDataObject instanceof char[])
    { // TEXT object being passed a NAMExx
      hDataObject = data[offset].toString().toCharArray();
      maxlen *= data[offset].getSizeInBytes();
    }
    else if (data instanceof NAME[] && hDataObject instanceof USTRING[])
    {
      for (int i=0; i< dArrayLength; i++)
      {
        ((USTRING[])hDataObject)[i].setString(data[i+offset].toString());
      }
    }
    else
    {
      if (!formatsEqual) return TErrorList.illegal_format;
      setObjectHandle(data);
    }
    if (dArrayLength > maxlen) dArrayLength = maxlen;
    repackageDataForOutput();
    return 0;
  }
  public int getData(double[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(double[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readFloat();
            break;
          case TFormat.CF_DLONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readLong();
            break;  
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(double data)
  {
    double[] d = new double[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(double[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(double[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)
          ((long[])hDataObject)[i] = (long) data[i+offset];
        break;  
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  
  
   public int getData(long[] data)
  {
    return getData(data,data != null ? data.length : 0, 0); 
  }
  public int getData(long[] data,int length,int offset)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength;  // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (long)ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = (long)ds.readFloat();
            break;
          case TFormat.CF_BITFIELD64:
          case TFormat.CF_DLONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readLong();
            break;  
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_BITFIELD32:
          case TFormat.CF_LONG:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readInt();
            break;
          case TFormat.CF_BITFIELD16:
          case TFormat.CF_SHORT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i+offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return 0;
  }
  public int putData(long data)
  {
    long[] d = new long[1];
    d[0] = data;
    return putData(d);
  }
  public int putData(long[] data)
  {
    return putData(data,data != null ? data.length : 0,0);
  }
  public int putData(long[] data,int length,int offset)
  {
    if (isLocked) return TErrorList.not_allowed;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dArrayLength > length) dArrayLength = length;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < dArrayLength; i++)
          ((double[])hDataObject)[i] = (double) data[i+offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < dArrayLength; i++)
          ((float[])hDataObject)[i] = (float) data[i+offset];
        break;
      case TFormat.CF_SHORT:
        for (int i = 0; i < dArrayLength; i++)
          ((short[])hDataObject)[i] = (short) data[i+offset];
        break;
      case TFormat.CF_DLONG:
        for (int i = 0; i < dArrayLength; i++)
          ((long[])hDataObject)[i] = (long) data[i+offset];
        break;  
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        for (int i = 0; i < dArrayLength; i++)
          ((int[])hDataObject)[i] = (int) data[i+offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public byte[] getDataBuffer()
  {
    try
    {
      if (dBuffers == null) return null;
      if (dBuffer == null)
        dBuffer = new ByteArrayOutputStream(dSizeInBytes);
      else
        dBuffer.reset();
      DataOutputStream dos = new DataOutputStream(dBuffer);
      for (int n = 0; n < numblks; n++)
      {
        if (dBuffers[n] == null) break;
        dos.write(dBuffers[n].toByteArray());
      }
      dos.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return dBuffer.toByteArray();
  }
  public byte[] getDataBuffer(int n)
  {
    try
    {
      if (dBuffer == null)
        dBuffer = new ByteArrayOutputStream();
      else
        dBuffer.reset();
      DataOutputStream dos = new DataOutputStream(dBuffer);
      if (n < numblks && dBuffers[n] != null)
      {
        dos.write(dBuffers[n].toByteArray());
      }
      dos.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return dBuffer.toByteArray();
  }
  /**
   * @brief Returns the data array length of the associated tine Data Type
   *        object
   * @return Returns the data array length of the associated tine Data Type
   *         object
   */
  public int getArrayLength()
  {
    int dlen = dArrayLength;
    if (dFormat == TFormat.CF_STRUCT) 
    { // if tagged structure, then this call returns the number of structures, not bytes !
      int ssiz = TStructRegistry.getSizeInBytes(dTag);
      if (ssiz < 1) ssiz = 1;
      dlen /= ssiz;
    }
    if (dFormat == TFormat.CF_STRING)
    {
      try 
      {
        if (dTag.length() > 0) dlen = Integer.parseInt(dTag);
      } catch (NumberFormatException e) {};
    }
    return dlen;
  }
  /**
   * @brief Returns the data type of the associated tine Data Type object
   * @return Returns the data type of the associated tine Data Type object
   */
  public short getFormat()
  {
    return dFormat >= 0 ? dFormat : (short) (256 + dFormat);
  }
  public void removeTagFromLink()
  {
    removeTagAtLinkTime = true;
  }
  public String getTag()
  {
    return removeTagAtLinkTime ? "" : dTag;
  }
  public int getSizeInBytes()
  {
    return dSizeInBytes;
  }
  public int getDataSize()
  {
    int fmtsiz = 0;
    if ((fmtsiz = TFormat.formatSizeOf(dFormat)) > 0)
    {
      int datasize = fmtsiz * dArrayLength;
      if (dFormat == TFormat.CF_SPECTRUM) datasize += SPECTRUM.hdrSizeInBytes;
      return datasize;
    }
    return 0;
  }
  private void makeNewCompoundDataObject(TCompoundDataObject[] Data, int index)
  {
    try
    {
      Data[index] = (TCompoundDataObject)Data.getClass().getComponentType().newInstance();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }
  private byte[] getStringArrayBytes(String[] data)
  {
    int sptr = 0, slen = 0;
    byte[] bdata = new byte[dSizeInBytes];
    for (int i=0; i<data.length; i++)
    { // now fill in the string space
      if (data[i] != null) 
      {
        slen = data[i].length();
        if (slen + sptr >= dSizeInBytes) break;
        System.arraycopy(data[i].getBytes(), 0, bdata, sptr, slen);
        sptr += slen;
      }
      if (sptr < dSizeInBytes) bdata[sptr++] = 0; 
    }
    return bdata;
  }
  private void setObjectContents(DataOutputStream ds, Object Data, int index)
  {
    try
    {
      if (Data instanceof byte[])
      {
        ds.writeByte(((byte[]) Data)[index]);
      }
      else if (Data instanceof short[])
      {
        ds.writeShort(Swap.Short(((short[]) Data)[index]));
      }
      else if (Data instanceof int[])
      {
        ds.writeInt(Swap.Long(((int[]) Data)[index]));
      }
      else if (Data instanceof long[])
      {
        ds.writeLong(Swap.DLong(((long[]) Data)[index]));
      }
      else if (Data instanceof float[])
      {
        ds.write(Swap.Float(((float[]) Data)[index]));
      }
      else if (Data instanceof double[])
      {
        ds.write(Swap.Double(((double[]) Data)[index]));
      }
      else if (Data instanceof char[])
      {
        ds.writeBytes(new String((char[])Data));
      }
      else if (Data instanceof StringBuffer)
      {
        ds.writeBytes(((StringBuffer)Data).toString());
      }
      else if (Data instanceof String)
      {
        ds.write(((String) Data).getBytes());
      }
      else if (Data instanceof TCompoundDataObject[])
      {
        if (((TCompoundDataObject[])Data)[index] == null) 
          makeNewCompoundDataObject((TCompoundDataObject[])Data,index); 
        ds.write(((TCompoundDataObject[])Data)[index].toByteArray());
      }
      else if (Data instanceof String[])
      {
        ds.write(getStringArrayBytes((String[])Data));
      }
      else if (Data instanceof SPECTRUM)
      {
        ds.write(((SPECTRUM)Data).toByteArray());
      }
      else if (Data instanceof IMAGE)
      {
        ds.write(((IMAGE)Data).toByteArray());
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public void setDataObject(int dataSize, short dataFormat)
  { // only allow this if the data object is null (used for CF_DEFAULT)
    if (hDataObject == null) makeDataObject(dataSize,dataFormat);
  }
  private void setObjectHandle(TCompoundDataObject[] Data)
  {
    hDataObject = Data;
  }  
  private void setObjectHandle(Object Data)
  {
    hDataObject = Data;    
  }
  public static int getObjectSize(Object Data)
  {
    if (Data instanceof TCompoundDataObject[])
      return ((TCompoundDataObject[]) Data).length;
    else if (Data instanceof byte[])
      return ((byte[]) Data).length;
    else if (Data instanceof short[])
      return ((short[]) Data).length;
    else if (Data instanceof int[])
      return ((int[]) Data).length;
    else if (Data instanceof float[])
      return ((float[]) Data).length;
    else if (Data instanceof double[])
      return ((double[]) Data).length;
    else if (Data instanceof SPECTRUM)
      return ((SPECTRUM) Data).hByteArray.length;
    else if (Data instanceof IMAGE)
      return ((IMAGE) Data).getImageFrameSize();
    else if (Data instanceof char[])
      return ((char[]) Data).length;
    else if (Data instanceof String)
      return ((String) Data).length();
    else if (Data instanceof StringBuffer) 
      return ((StringBuffer) Data).toString().length();
    return 0;
  }
  public static short getObjectFormat(Object Data)
  {
    if (Data instanceof TCompoundDataObject[])
    {
      try
      {
        Object tdo = ((TCompoundDataObject[])Data)[0] != null ? 
            ((TCompoundDataObject[])Data)[0] : Data.getClass().getComponentType().newInstance();
        return ((TCompoundDataObject) tdo).getFormat();
      }
      catch (Exception e)
      {
      }
    }
    if (Data instanceof byte[])
      return TFormat.CF_BYTE;
    else if (Data instanceof short[])
      return TFormat.CF_SHORT;
    else if (Data instanceof int[])
      return TFormat.CF_LONG;
    else if (Data instanceof float[])        
      return TFormat.CF_FLOAT;
    else if (Data instanceof long[])
      return TFormat.CF_DLONG;
    else if (Data instanceof double[])
      return TFormat.CF_DOUBLE;
    else if (Data instanceof String)
      return TFormat.CF_TEXT;
    else if (Data instanceof StringBuffer)
      return TFormat.CF_TEXT;
    else if (Data instanceof char[]) 
      return TFormat.CF_TEXT;
    else if (Data instanceof String[])
      return TFormat.CF_STRING;
    return TFormat.CF_NULL;
  }
  public int getData(Object Data)
  {
    if (Data instanceof byte[])
      getData((byte[]) Data);
    else if (Data instanceof short[])
      getData((short[]) Data);
    else if (Data instanceof int[])
      getData((int[]) Data);
    else if (Data instanceof float[])
      getData((float[]) Data);
    else if (Data instanceof double[])
      getData((double[]) Data);
    else if (Data instanceof long[])
      getData((long[]) Data);
    else if (Data instanceof char[])
      getData((char[]) Data);
    else if (Data instanceof StringBuffer)
      getData((StringBuffer) Data);
    else if (Data instanceof TCompoundDataObject[])
      getData((TCompoundDataObject[]) Data);
    else if (Data instanceof SPECTRUM) 
      getData((SPECTRUM) Data);
    else if (Data instanceof IMAGE) 
      getData((IMAGE) Data);
    else if (Data instanceof String[]) 
      getData((String[]) Data);
    return 0;
  }
  public int dataFill(String fromStr)
  {
    // TODO: add the new formats
    int n = 0;
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        for (n = 0; n < dArrayLength && ((NAME16FI[])hDataObject)[n] != null; n++)
          ((NAME16FI[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16II:
        for (n = 0; n < dArrayLength && ((NAME16II[])hDataObject)[n] != null; n++)
          ((NAME16II[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME16DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME16DBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME32DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME32DBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBL:
        for (n = 0; n < dArrayLength && ((NAME64DBL[])hDataObject)[n] != null; n++)
          ((NAME64DBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME64DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME64DBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        for (n = 0; n < dArrayLength && ((NAME64DBLDBLDBL[])hDataObject)[n] != null; n++)
          ((NAME64DBLDBLDBL[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME48:
      case TFormat.CF_NAME64:
        for (n = 0; n < dArrayLength && ((TCompoundDataObject[])hDataObject)[n] != null; n++)
          ((NAME[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16I:
      case TFormat.CF_NAME32I:
      case TFormat.CF_NAME48I:
      case TFormat.CF_NAME64I:
        for (n = 0; n < dArrayLength && ((TCompoundDataObject[])hDataObject)[n] != null; n++)
          ((NAMEINT[])hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_USTRING:
        for (n = 0; n < dArrayLength && ((USTRING[])hDataObject)[n] != null; n++)
          ((USTRING[])hDataObject)[n].setString(fromStr);
        break;
      case TFormat.CF_UNAME:
        for (n = 0; n < dArrayLength && ((UNAME[])hDataObject)[n] != null; n++)
          ((UNAME[])hDataObject)[n].setString(fromStr);
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        int size = dArrayLength < fromStr.length() ? dArrayLength : fromStr.length();
        char[] s = fromStr.toCharArray();
        for (n = 0; n < size && hDataObject != null; n++)
          ((char[])hDataObject)[n] = s[n];
        break;
    }
    return n;
  }
  public int dataFill(double fromVal)
  {
    return dataFill(fromVal, 0);
  }
  public int dataFill(double fromVal, int offset)
  {
    // TODO: add the new formats
    int n = 0, len;
    if (offset >= dArrayLength) return 0;
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        if ((len = ((NAME16FI[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16FI[])hDataObject)[n] != null; n++)
          ((NAME16FI[])hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_NAME16II:
        if ((len = ((NAME16II[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16II[])hDataObject)[n] != null; n++)
          ((NAME16II[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_NAME16DBLDBL:
        if ((len = ((NAME16DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME16DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME32DBLDBL:
        if ((len = ((NAME32DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME32DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME32DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME64DBL:
        if ((len = ((NAME64DBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBL[])hDataObject)[n] != null; n++)
          ((NAME64DBL[])hDataObject)[n].dval = fromVal;
        break;
      case TFormat.CF_NAME64DBLDBL:
        if ((len = ((NAME64DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBLDBL[])hDataObject)[n] != null; n++)
          ((NAME64DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        if ((len = ((NAME64DBLDBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBLDBLDBL[])hDataObject)[n] != null; n++)
          ((NAME64DBLDBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        if ((len = ((byte[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((byte[])hDataObject) != null; n++)
          ((byte[])hDataObject)[n] = (byte) fromVal;
        break;
      case TFormat.CF_SHORT:
        if ((len = ((short[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((short[])hDataObject) != null; n++)
          ((short[])hDataObject)[n] = (short) fromVal;
        break;
      case TFormat.CF_LONG:
        if ((len = ((int[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((int[])hDataObject) != null; n++)
          ((int[])hDataObject)[n] = (int) fromVal;
        break;
      case TFormat.CF_DLONG:
        if ((len = ((long[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((long[])hDataObject) != null; n++)
          ((long[])hDataObject)[n] = (long)fromVal;
        break;  
      case TFormat.CF_FLOAT:
        if ((len = ((float[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((float[])hDataObject) != null; n++)
          ((float[])hDataObject)[n] = (float) fromVal;
        break;
      case TFormat.CF_DOUBLE:
        if ((len = ((double[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((double[])hDataObject) != null; n++)
          ((double[])hDataObject)[n] = fromVal;
        break;
      case TFormat.CF_LNGINT:
        if ((len = ((INTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINT[])hDataObject)[n] != null; n++)
          ((INTINT[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FLTINT:
        if ((len = ((FLTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINT[])hDataObject)[n] != null; n++)
          ((FLTINT[])hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_FLTFLT:
        if ((len = ((FLTFLT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTFLT[])hDataObject)[n] != null; n++)
          ((FLTFLT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_DBLDBL:
        if ((len = ((DBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((DBLDBL[])hDataObject)[n] != null; n++)
          ((DBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_DBLDBLDBL:
        if ((len = ((DBLDBLDBL[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((DBLDBLDBL[])hDataObject)[n] != null; n++)
          ((DBLDBLDBL[])hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_INTFLTINT:
        if ((len = ((INTFLTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTFLTINT[])hDataObject)[n] != null; n++)
          ((INTFLTINT[])hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_FLTFLTINT:
        if ((len = ((FLTFLTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTFLTINT[])hDataObject)[n] != null; n++)
          ((FLTFLTINT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_FLTINTINT:
        if ((len = ((FLTINTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINTINT[])hDataObject)[n] != null; n++)
          ((FLTINTINT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_INTINTINT:
        if ((len = ((INTINTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTINT[])hDataObject)[n] != null; n++)
          ((INTINTINT[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_ADDRESS:
        if ((len = ((INTINTINTINT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTINTINT[])hDataObject)[n] != null; n++)
          ((INTINTINTINT[])hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FILTER:
        if ((len = ((INTFLTFLTFLT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTFLTFLTFLT[])hDataObject)[n] != null; n++)
          ((INTFLTFLTFLT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_FWINDOW:
        if ((len = ((INTINTFLTFLT[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTFLTFLT[])hDataObject)[n] != null; n++)
          ((INTINTFLTFLT[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_USTRING:
        if ((len = ((USTRING[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((USTRING[])hDataObject)[n] != null; n++)
          ((USTRING[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_UNAME:
        if ((len = ((UNAME[])hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((UNAME[])hDataObject)[n] != null; n++)
          ((UNAME[])hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_SPECTRUM:
        for (n = offset; n < dArrayLength && ((SPECTRUM)hDataObject) != null; n++)
          ((SPECTRUM)hDataObject).d_spect_array[n] = (float) fromVal;
        break;
    }
    return n - offset;
  }
  public int dataCopy(TDataType fromData)
  {
    if (fromData == null) return 0;
    return dataCopy(fromData,0);
  }
  public int dataCopy(TDataType fromData,int offset)
  {
    boolean assignBuffers = false;
    if (dFormat != fromData.dFormat) 
    {
      if (dFormat == TFormat.CF_DEFAULT) synchronized (fromData)
      { // just for jDDD ...
        dFormat = fromData.dFormat;
        dArrayLength = fromData.dArrayLength;
        dCompletionLength = fromData.dCompletionLength;
        if (hDataObject == null && hStructObject == null)
          setDataObject(dArrayLength,dFormat);
        assignBuffers = true;
      }
      else 
      { // no can do -> jump out
        return 0;
      }
    }
    if (fromData.hasBeenUpdated && !hasBeenUpdated) 
    {
      assignBuffers = true;
    }
    if (assignBuffers)
    {
      if (dBuffers == null) dBuffers = fromData.dBuffers;
      hasBeenUpdated = true;
    }
    numblks = fromData.numblks;
    int n = 0;
    int size = dArrayLength < fromData.dArrayLength ? dArrayLength : fromData.dArrayLength;

    // copy the innards ...
    copy(fromData);

    if (fromData.hDataObject == null || hDataObject == null) return 0;
    // copy into the referenced object ...
    if (hDataObject instanceof TCompoundDataObject[])
    {
      System.arraycopy((TCompoundDataObject[])fromData.hDataObject,offset,(TCompoundDataObject[])hDataObject,0,size);
      n = size;
    }
    // TODO: catch SPECTRUM and any other special cases ?
    // TODO: LNGINT and INTINT both have the same format (also TDS and INTFLTINT)
    switch (dFormat)
    {
      case TFormat.CF_DEFAULT:
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_TEXT:
        if (hDataObject instanceof char[])
        {
          System.arraycopy((char[])fromData.hDataObject, offset, (char[])hDataObject, 0, size);
          n = size;
          break;          
        }
        else if (hDataObject instanceof StringBuffer)
        {
          String s = ((StringBuffer)fromData.hDataObject).substring(0);
          n = s.length();
          ((StringBuffer)hDataObject).ensureCapacity(n);
          ((StringBuffer)hDataObject).insert(0,s);
          break;
        }
      case TFormat.CF_STRING:
        for (int i=0; i<size; i++) ((String[])hDataObject)[i] = ((String[])fromData.hDataObject)[i]; 
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        System.arraycopy((byte[])fromData.hDataObject, offset, (byte[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_SHORT:
        System.arraycopy((short[])fromData.hDataObject, offset, (short[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_BOOLEAN:
      case TFormat.CF_LONG:
        System.arraycopy((int[])fromData.hDataObject, offset, (int[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_DLONG:
        System.arraycopy((long[])fromData.hDataObject, offset, (long[])hDataObject, 0, size);
        n = size;
        break;        
      case TFormat.CF_FLOAT:
        System.arraycopy((float[])fromData.hDataObject, offset, (float[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_DOUBLE:
        System.arraycopy((double[])fromData.hDataObject, offset, (double[])hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_SPECTRUM:
        System.arraycopy(((SPECTRUM)fromData.hDataObject).d_spect_array, offset, ((SPECTRUM)hDataObject).d_spect_array, 0, size);
        n = size;
        break;
    }
    return n;
  }
public Object getStructObject() { return hStructObject; }
  /**
   * @brief Retrieves the returned data as an Object container. This method is
   *        useful for obtaining the unbound data when the data type is unknown
   *        at link time. A call to getFormat() can be made to ascertain the
   *        tine data type and hence the array nature of the object returned.
   */
  public Object getDataObject() { return hDataObject; }
  /**
   * @brief Retrieves the returned data as a Vector. This method is useful for
   *        obtaining the unbound data when the data type is unknown at link
   *        time. A call to getFormat() can be made to ascertain the tine data
   *        type and hence the contents of the Vector.
   */
  public Vector<Object> getVector()
  {
    // TODO: clean up the individual type allocations and follow the NAME8 example
    // TODO: add the new formats
    // We still have to have the long switch on the format so that we know 
    // what we have to create!
    int i = 0;
    if (dVector == null) dVector = new Vector<Object>();
    Vector<Object> v = dVector;
    v.clear();
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME8:
      {
        NAME8 d[] = (NAME8[])hDataObject;
        if (d == null)
        {
          d = new NAME8[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;        
      }
      case TFormat.CF_NAME16:
      {
        NAME16 d[] = (NAME16[])hDataObject;
        if (d == null)
        {
          d = new NAME16[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME32:
      {
        NAME32 d[] = (NAME32[])hDataObject;
        if (d == null)
        {
          d = new NAME32[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME48:
      {
        NAME48 d[] = (NAME48[])hDataObject;
        if (d == null)
        {
          d = new NAME48[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64:
      {
        NAME64 d[] = (NAME64[])hDataObject;
        if (d == null)
        {
          d = new NAME64[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16I:
      {
        NAME16I d[] = (NAME16I[])hDataObject;
        if (d == null)
        {
          d = new NAME16I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16FI:
      {
        NAME16FI d[] = (NAME16FI[])hDataObject;
        if (d == null)
        {
          d = new NAME16FI[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16II:
      {
        NAME16II d[] = (NAME16II[])hDataObject;
        if (d == null)
        {
          d = new NAME16II[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16DBLDBL:
      {
        NAME16DBLDBL d[] = (NAME16DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME16DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME32DBLDBL:
      {
        NAME32DBLDBL d[] = (NAME32DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME32DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64DBL:
      {
        NAME64DBL d[] = (NAME64DBL[])hDataObject;
        if (d == null)
        {
          d = new NAME64DBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64DBLDBL:
      {
        NAME64DBLDBL d[] = (NAME64DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME64DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64DBLDBLDBL:
      {
        NAME64DBLDBLDBL d[] = (NAME64DBLDBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME64DBLDBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME32I:
      {
        NAME32I d[] = (NAME32I[])hDataObject;
        if (d == null)
        {
          d = new NAME32I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME48I:
      {
        NAME48I d[] = (NAME48I[])hDataObject;
        if (d == null)
        {
          d = new NAME48I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64I:
      {
        NAME64I d[] = (NAME64I[])hDataObject;
        if (d == null)
        {
          d = new NAME64I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_BYTE:
      {
        byte d[] = (byte[])hDataObject;
        if (d == null)
        {
          d = new byte[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(new Byte(d[i]));
        break;
      }
      case TFormat.CF_HISTORY:
      case TFormat.CF_STRUCT:
        // TODO: could check the tag and make a vector of the struct object
        break;
      case TFormat.CF_SHORT:
      {
        short d[] = (short[])hDataObject;
        if (d == null)
        {
          d = new short[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(new Short(d[i]));
        break;
      }
      case TFormat.CF_LONG:
      {
        int d[] = (int[])hDataObject;
        if (d == null)
        {
          d = new int[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(new Integer(d[i]));
        break;
      }
      case TFormat.CF_DLONG:
      {
        long d[] = (long[])hDataObject;
        if (d == null)
        {
          d = new long[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(new Long(d[i]));
        break;
      }         
      case TFormat.CF_FLOAT:
      {
        float d[] = (float[])hDataObject;
        if (d == null)
        {
          d = new float[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(new Float(d[i]));
        break;
      }
      case TFormat.CF_DOUBLE:
      {
        double d[] = (double[])hDataObject;
        if (d == null)
        {
          d = new double[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(new Double(d[i]));
        break;
      }
      case TFormat.CF_LNGINT:
      {
        INTINT d[] = (INTINT[])hDataObject;
        if (d == null)
        {
          d = new INTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTINT:
      {
        FLTINT d[] = (FLTINT[])hDataObject;
        if (d == null)
        {
          d = new FLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTFLT:
      {
        FLTFLT d[] = (FLTFLT[])hDataObject;
        if (d == null)
        {
          d = new FLTFLT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DBLINT:
      {
        DBLINT d[] = (DBLINT[])hDataObject;
        if (d == null)
        {
          d = new DBLINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DBLDBL:
      {
        DBLDBL d[] = (DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DBLDBLDBL:
      {
        DBLDBLDBL d[] = (DBLDBLDBL[])hDataObject;
        if (d == null)
        {
          d = new DBLDBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INTFLTINT:
      {
        INTFLTINT d[] = (INTFLTINT[])hDataObject;
        if (d == null)
        {
          d = new INTFLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTFLTINT:
      {
        FLTFLTINT d[] = (FLTFLTINT[])hDataObject;
        if (d == null)
        {
          d = new FLTFLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTINTINT:
      {
        FLTINTINT d[] = (FLTINTINT[])hDataObject;
        if (d == null)
        {
          d = new FLTINTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INTINTINT:
      {
        INTINTINT d[] = (INTINTINT[])hDataObject;
        if (d == null)
        {
          d = new INTINTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_ADDRESS:
      {
        INTINTINTINT d[] = (INTINTINTINT[])hDataObject;
        if (d == null)
        {
          d = new INTINTINTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FILTER:
      {
        INTFLTFLTFLT d[] = (INTFLTFLTFLT[])hDataObject;
        if (d == null)
        {
          d = new INTFLTFLTFLT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FWINDOW:
      {
        INTINTFLTFLT d[] = (INTINTFLTFLT[])hDataObject;
        if (d == null)
        {
          d = new INTINTFLTFLT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_USTRING:
      {
        USTRING d[] = (USTRING[])hDataObject;;
        if (d == null)
        {
          d = new USTRING[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++) v.addElement(d[i]);
        break;
      }
      case TFormat.CF_UNAME:
      {
        UNAME undata[] = (UNAME[])hDataObject;;
        if (undata == null)
        {
          undata = new UNAME[dArrayLength];
          getData(undata);
        }
        for (i = 0; i < undata.length; i++) v.addElement(undata[i]);
        break;
      }
    }
    dCompletionLength = i;
    return v;
  }
  private String pushToString(TCompoundDataObject[] cdo)
  {
    if (dCompletionLength == 0) return "";
    StringBuffer sb = new StringBuffer(cdo[0].getSizeInBytes() * dCompletionLength);
    for (int i = 0; i < dCompletionLength && i < cdo.length; i++)
    {
      if (cdo[i] == null) break;
      sb.append(cdo[i].toString());
      sb.append(arrayDelimiter);
    }
    return sb.toString();
  }
  /**
   * @brief Outputs the returned data as a String. This method is useful for
   *        displaying the unbound data when the data type is unknown at link
   *        time.
   */
    @Override
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    // TODO: clean up, add the new formats
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        return new String("");
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
      {
        if (hDataObject instanceof char[]) 
          return new String((char[])hDataObject).trim();
        if (hDataObject instanceof StringBuffer) 
          return ((StringBuffer)hDataObject).toString().trim();
        char[] t = new char[dArrayLength];
        getData(t);
        return new String(t).trim();
      }
      case TFormat.CF_STRING:
      {
        String[] d = (String[])hDataObject;
        if (d == null)
        {
          d = new String[dArrayLength];
          getData(d);
        }
        String ds = new String();
        for (int i=0; i<d.length; i++) ds += d[i]+"\n";
        return ds;
      }
      case TFormat.CF_USTRING:
      {
        USTRING d[] = (USTRING[])hDataObject;
        if (d == null)
        {
          d = new USTRING[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_UNAME:
      {
        UNAME d[] = (UNAME[])hDataObject;
        if (d == null)
        {
          d = new UNAME[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_INTINTINT:
      {
        INTINTINT d[] = (INTINTINT[])hDataObject;
        if (d == null)
        {
          d = new INTINTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_ADDRESS:
      {
        INTINTINTINT d[] = (INTINTINTINT[])hDataObject;
        if (d == null)
        {
          d = new INTINTINTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_FILTER:
      {
        INTFLTFLTFLT d[] = (INTFLTFLTFLT[])hDataObject;
        if (d == null)
        {
          d = new INTFLTFLTFLT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_FWINDOW:
      {
        INTINTFLTFLT d[] = (INTINTFLTFLT[])hDataObject;
        if (d == null)
        {
          d = new INTINTFLTFLT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_INTFLTINT:
      {
        INTFLTINT d[] = (INTFLTINT[])hDataObject;
        if (d == null)
        {
          d = new INTFLTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_FLTFLTINT:
      {
        FLTFLTINT d[] = (FLTFLTINT[])hDataObject;
        if (d == null)
        {
          d = new FLTFLTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_FLTINTINT:
      {
        FLTINTINT d[] = (FLTINTINT[])hDataObject;
        if (d == null)
        {
          d = new FLTINTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME8:
      {
        NAME8 d[] = (NAME8[])hDataObject;
        if (d == null)
        {
          d = new NAME8[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME16:
      {
        NAME16 d[] = (NAME16[])hDataObject;
        if (d == null)
        {
          d = new NAME16[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME32:
      {
        NAME32 d[] = (NAME32[])hDataObject;
        if (d == null)
        {
          d = new NAME32[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME48:
      {
        NAME48 d[] = (NAME48[])hDataObject;
        if (d == null)
        {
          d = new NAME48[dArrayLength];
          getData(d);
        }
        getData(d);
        return pushToString(d);
      }
      case TFormat.CF_NAME64:
      {
        NAME64 d[] = (NAME64[])hDataObject;
        if (d == null)
        {
          d = new NAME64[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME16I:
      {
        NAME16I d[] = (NAME16I[])hDataObject;
        if (d == null)
        {
          d = new NAME16I[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME16FI:
      {
        NAME16FI d[] = (NAME16FI[])hDataObject;
        if (d == null)
        {
          d = new NAME16FI[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME16II:
      {
        NAME16II d[] = (NAME16II[])hDataObject;
        if (d == null)
        {
          d = new NAME16II[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME16DBLDBL:
      {
        NAME16DBLDBL d[] = (NAME16DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME16DBLDBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME32DBLDBL:
      {
        NAME32DBLDBL d[] = (NAME32DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME32DBLDBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME64DBL:
      {
        NAME64DBL d[] = (NAME64DBL[])hDataObject;
        if (d == null)
        {
          d = new NAME64DBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME64DBLDBL:
      {
        NAME64DBLDBL d[] = (NAME64DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME64DBLDBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME64DBLDBLDBL:
      {
        NAME64DBLDBLDBL d[] = (NAME64DBLDBLDBL[])hDataObject;
        if (d == null)
        {
          d = new NAME64DBLDBLDBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME32I:
      {
        NAME32I d[] = (NAME32I[])hDataObject;
        if (d == null)
        {
          d = new NAME32I[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME48I:
      {
        NAME48I d[] = (NAME48I[])hDataObject;
        if (d == null)
        {
          d = new NAME48I[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_NAME64I:
      {
        NAME64I d[] = (NAME64I[])hDataObject;
        if (d == null)
        {
          d = new NAME64I[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_BITFIELD8:
      case TFormat.CF_BYTE:
      {
        byte d[] = (byte[])hDataObject;
        if (d == null)
        {
          d = new byte[dArrayLength];
          getData(d);
        }
        String ds;
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          if (dFormat == TFormat.CF_BITFIELD8 && dBitField == null)
          {
            ds = TBitfieldRegistry.getBitfield(dTag).toString();
          }
          else
          {
            ds = new String(""+d[i]);
          }
          sb.append(ds);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }
      case TFormat.CF_HISTORY:
      {
        HISTORY d[] = (HISTORY[])hDataObject;
        if (d == null)
        {
          d = new HISTORY[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_STRUCT:
      {
        if (hStructObject != null && hStructObject instanceof TTaggedStructure)
        {
          return ((TTaggedStructure)hStructObject).toString();
        }
        if (hStructObject instanceof TTaggedStructure[])
        {
          for (int i = 0; i < ((TTaggedStructure[])hStructObject).length; i++)
          {
            sb.append(((TTaggedStructure[])hStructObject)[i].toString());
            sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        if (hDataObject != null)
        {
          int len = TStructRegistry.getSizeInBytes(dTag) * dCompletionLength;
          return TStructRegistry.toString(dTag, (byte[])hDataObject, len);
        }
        return new String("user defined structure");
      }
      case TFormat.CF_BITFIELD16:
      case TFormat.CF_SHORT:
      {
        short d[] = (short[])hDataObject;
        if (d == null)
        {
          d = new short[dArrayLength];
          getData(d);
        }
        String ds;
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          if (dFormat == TFormat.CF_BITFIELD16 && dBitField == null)
          {
            ds = TBitfieldRegistry.getBitfield(dTag).toString(d[i]);
          }
          else
          {
            ds = new String(""+d[i]);
          }
          sb.append(ds);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }
      case TFormat.CF_BITFIELD32:
      case TFormat.CF_LONG:
      {
        int d[] = (int[])hDataObject;
        if (d == null)
        {
          d = new int[dArrayLength];
          getData(d);
        }
        String ds;
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          if (dFormat == TFormat.CF_BITFIELD32 && dBitField == null)
          {
            ds = TBitfieldRegistry.getBitfield(dTag).toString();
          }
          else
          {
            ds = new String(""+d[i]);
          }
          sb.append(ds);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }      
      case TFormat.CF_BITFIELD64:
      case TFormat.CF_DLONG:
      {
        long d[] = (long[])hDataObject;
        if (d == null)
        {
          d = new long[dArrayLength];
          getData(d);
        }
        String ds;
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          if (dFormat == TFormat.CF_BITFIELD64 && dBitField == null)
          {
            ds = TBitfieldRegistry.getBitfield(dTag).toString();
          }
          else
          {
            ds = new String(""+d[i]);
          }
          sb.append(ds);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }            
      case TFormat.CF_BOOLEAN:
      {
        int d[] = (int[])hDataObject;
        if (d == null)
        {
          d = new int[dArrayLength];
          getData(d);
        }
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          sb.append(d[i] == 0 ? "false" : "true");
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }
      case TFormat.CF_LNGINT:
      {
        INTINT d[] = (INTINT[])hDataObject;
        if (d == null)
        {
          d = new INTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_FLOAT:
      {
        float d[] = (float[])hDataObject;
        if (d == null)
        {
          d = new float[dArrayLength];
          getData(d);
        }
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          sb.append(d[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }
      case TFormat.CF_FLTINT:
      {
        FLTINT d[] = (FLTINT[])hDataObject;
        if (d == null)
        {
          d = new FLTINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_FLTFLT:
      {
        FLTFLT d[] = (FLTFLT[])hDataObject;
        if (d == null)
        {
          d = new FLTFLT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_SPECTRUM:
      {
        SPECTRUM d = (SPECTRUM)hDataObject;
        if (d == null)
        {
          d = new SPECTRUM();
          getData(d);
        }
        return d.toString();
      }
      case TFormat.CF_DOUBLE:
      {
        double d[] = (double[])hDataObject;
        if (d == null)
        {
          d = new double[dArrayLength];
          getData(d);
        }
        for (int i = 0; i < dCompletionLength && i < d.length; i++)
        {
          sb.append(d[i]);
          sb.append(arrayDelimiter);
        }
        return sb.toString();
      }
      case TFormat.CF_DBLINT:
      {
        DBLINT d[] = (DBLINT[])hDataObject;
        if (d == null)
        {
          d = new DBLINT[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_DBLDBL:
      {
        DBLDBL d[] = (DBLDBL[])hDataObject;
        if (d == null)
        {
          d = new DBLDBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
      case TFormat.CF_DBLDBLDBL:
      {
        DBLDBLDBL d[] = (DBLDBLDBL[])hDataObject;
        if (d == null)
        {
          d = new DBLDBLDBL[dArrayLength];
          getData(d);
        }
        return pushToString(d);
      }
    }
    return new String("");
  }
  /**
   * @brief An array of chars (to be interpretted as a text string)
   * @param data
   *          is a reference to the char[] array used to create and bind to the
   *          TDataType object
   */
  public TDataType(char[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of chars (to be interpretted as a text string)
   * @param data
   *          is a reference to the char[] array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(char[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(char[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        byte bdata;
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
        {
          bdata = (byte) data[i];
          ds.writeByte(bdata);
        }
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A text string buffer
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   */
  public TDataType(StringBuffer data)
  {
    initTDataType(data, data.capacity());
  }
  /**
   * @brief A subset of a text string buffer
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset length to use.
   */
  public TDataType(StringBuffer data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(StringBuffer data, int length)
  {
    try
    {
      if (length > data.capacity()) length = data.capacity();
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = dArrayLength;
      byte[] bdata = data.toString().getBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A subset of an array of Strings. This corresponds to
   *        tine format CF_STRING
   * @param data
   *          is a reference to the String array used to create and bind to the
   *          TDataType object
   */
  public TDataType(String[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * @brief A subset of an array of Strings. This corresponds to
   *        tine format CF_STRING
   * @param data
   *          is a reference to the String array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(String[] data,int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(String[] data, int length)
  {
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_STRING;
      dArrayLength = length;
      dSizeInBytes = 0;
      for (int i=0; i<length; i++)
      { // first see how much string space we need (going out)
        dSizeInBytes += (data[i] != null) ? data[i].length() : 0; 
        dSizeInBytes++; // add terminating 0
      }
      byte[] bdata = getStringArrayBytes(data);
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  /**
   * @brief A text string
   * @param data
   *          is a reference to the String used to create and bind to the
   *          TDataType object
   */
  public TDataType(String data)
  {
    initTDataType(data, data.length());
  }
  /**
   * @brief A text string
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset length to use.
   */
  public TDataType(String data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(String data, int length)
  {
    try
    {
      if (length > data.length()) length = data.length();
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = dArrayLength;
      byte[] bdata = data.getBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = new char[dArrayLength];
      if (dArrayLength > 0) data.getChars(0, dArrayLength, (char[])hDataObject, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }
  public int getData(char[] data)
  {
    if (dArrayLength > data.length) return TErrorList.string_too_long;
    int i, n, m, cplen;
    byte bdata[] = new byte[data.length];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      cplen = b.length; 
      if (cplen > data.length - m) cplen = data.length - m;
      if (cplen > 0)
      {
        System.arraycopy(b, 0, bdata, m, cplen);
        i += cplen;
      }
    }
    String sdata = new String(bdata,0,i);
    sdata.getChars(0, sdata.length(), data, 0);
//    if (hDataObject instanceof StringBuffer) 
//      ((StringBuffer)hDataObject).replace(0,sdata.length(),sdata);
//    else if (hDataObject instanceof char[])
//      System.arraycopy(data,0,(char[])hDataObject,0,sdata.length());
    dCompletionLength = i;
    return 0;
  }
  public int getData(StringBuffer data)
  {
    if (dArrayLength > data.capacity()) return TErrorList.string_too_long;
    int i, n, m;
    byte bdata[] = new byte[data.capacity()];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      int cplen = b.length; 
      if (cplen > data.capacity() - m) cplen = data.capacity() - m;
      if (cplen > 0)
      {
        System.arraycopy(b, 0, bdata, m, b.length);
        i += cplen;
      }
    }
    String sdata = new String(bdata,0,i);
    data.replace(0,sdata.length(),sdata);
    if (hDataObject instanceof StringBuffer) 
      ((StringBuffer)hDataObject).replace(0,sdata.length(),sdata);
    else if (hDataObject instanceof char[])
      System.arraycopy(data,0,(char[])hDataObject,0,sdata.length());
    dCompletionLength = i;
    return 0;
  }
  public int getData(String[] data)
  {
    int dlen = dArrayLength;
    if (dArrayLength > data.length) dlen = data.length;
    int i, n, m;
    dSizeInBytes = bytesin;
    byte[] bdata = new byte[dSizeInBytes];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      System.arraycopy(b, 0, bdata, m, b.length);
      i += b.length;
    }
    int off=0,len=0;
    for (i=0; i<dlen; i++)
    {
      for (len=0; len+off<dSizeInBytes; len++)
        if (bdata[off+len] == 0) break;
      if (len >=dSizeInBytes-off) break;
      data[i] = new String(bdata,off,len);
      if (hDataObject instanceof StringBuffer)
      {
        data[0] = ((StringBuffer)hDataObject).toString().trim();
        break;
      }
      else if (hDataObject instanceof char[])
      {
        data[0] = new String((char[])hDataObject).trim();
        break;
      }
      else if (hDataObject instanceof String[])
      {
        data[i] = ((String[])hDataObject)[i];
      }
      off += len + 1;
    }
    dCompletionLength = i;
    return 0;
  }
  public void setTag(String tag)
  {
    dTag = tag;
  }
  /**
   * @brief returns the last data completion length.
   * @return the last data completion length.  This can be different
   * (smaller) than the data length used to construct the object.
   */
  public int getCompletionLength()
  {
    return dCompletionLength;
  }
  public void setArrayLength(int arrayLength)
  {
    int dlen = arrayLength;
    if (dFormat == TFormat.CF_STRUCT) 
    { // if tagged structure, then this call takes the number of structures, not bytes !
      int ssiz = TStructRegistry.getSizeInBytes(dTag);
      if (ssiz < 1) ssiz = 1;
      dlen *= ssiz;
    }
    dArrayLength = dlen;
  }
  public String getField()
  {
    return dField;
  }
  public void setField(String field)
  {
    dField = field;
  }
  public TBitfield getBitField()
  {
    return dBitField;
  }
  public void setBitField(TBitfield bitfield)
  {
    dBitField = bitfield;
  }
}