package de.desy.tine.definitions;

/**
 * @brief data access flags
 * 
 * The access word in a TLink will contain information pertaining
 * to the type of access requested by the caller or by the subsystem. A
 * client can specify 'CA_READ' or 'CA_WRITE' and in addition ask for a
 * data link over a connected (TCP/IP) socket by specifying
 * 'CA_CONNECT'. A server can examine the access word when handling the
 * call. A server can also determine whether the call is coming from the
 * local alarm server or local history server. Likewise, a server can
 * examine the scope of the call, that is whether the link is starting
 * up (the 'CA_FIRST' bit is set) or going out of scope (the 'CA_LAST'
 * bit is set).
 */
public final class TAccess
{
  private short mode;
  public static final short CA_NULL = 0x00;
  public static final short CA_READ = 0x01;
  public static final short CA_WRITE = 0x02;
  public static final short CA_REPEAT = 0x04;
  public static final short CA_SYNC = 0x08;
  public static final short CA_ALARM = 0x10;
  public static final short CA_HIST = 0x20;
  public static final short CA_FIRST = 0x40;
  public static final short CA_LAST = 0x80;
  public static final short CA_RETRY = 0x100;
  public static final short CA_NETWORK = 0x200;
  public static final short CA_QUERY = 0x400;
  public static final short CA_LOCKED = 0x800;
  public static final short CA_NORETRY = 0x1000;
  public static final short CA_CONNECT = 0x2000;
  public static final short CA_MUTABLE = 0x4000;
//  public static final short CA_SYNCNOTIFY = 0x8000; <doesn't work as a short in java !
  /**
   * Constructor. Used for when CF-constants are or-ed
   * 
   * @param newMode
   */
  public TAccess(int newMode)
  {
    mode = (short) newMode;
  }
  public TAccess()
  {
    this(CA_NULL);
  }
  /**
   * @brief sets the access mode to CA_NULL
   */
  public void clear()
  {
    mode = CA_NULL;
  }
  /**
   * @param access is the access mode for which the base mode is desired
   * @return the base access mode 
   * (one of CA_NULL, CA_READ, CA_WRITE, or CA_READ|CA_WRITE)
   */
  public static short toBase(short access)
  {
    return (short) (access & 0x03);
  }
  
  /**
   * @param mode is the access mode for which a TAccess object is desired
   * @return a TAccess Object representative of the input mode
   */
  public static TAccess valueOf(short mode) 
  {
	return new TAccess(mode);
  }
  /** 
   * @brief returns a string representation of the access mode
   * @return the string representation of the access mode 
   */
  public String toString()
  {
    return TAccess.toString(mode);
  }
  /** 
   * @brief returns a string representation of the access mode given
   * @param access is the access mode for which the string representation is desired
   * @return the string representation of the access mode 
   */
  public static String toString(short access)
  {
    String add = "";
    if ((access & CA_SYNC) == CA_SYNC) add = "sync ";
    switch (access & 0x03)
    {
      case CA_NULL:
        return "none";
      case CA_READ:
        return add + "read only";
      case CA_WRITE:
        return add + "write";
      default:
        return add + "read/write";
    }
  }
  /**
   * Returns the access mode bits as a short integer.
   * 
   * @return
   */
  public short toShort()
  {
    return this.mode;
  }
  /**
   * Returns true if all bits which are set in pattern are also set in this
   * object.
   * 
   * @param pattern
   *          A Bit pattern.
   * @return
   */
  public boolean includes(short pattern)
  {
    return ((this.mode & pattern) == pattern);
  }
  /**
   * @brief Returns true if read bit is set.
   * @return true if read bit is set.
   */
  public boolean isRead()
  {
    return ((this.mode & CA_READ) != 0);
  }
  /**
   * @brief Returns true if read bit is set.
   * @param access is the access byte to test
   * @return true if read bit is set.
   */
  public static boolean isRead(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_READ) != 0;
  }
  /**
   * @brief returns true if write bit is set.
   * @return true if write bit is set.
   */
  public boolean isWrite()
  {
    return ((this.mode & CA_WRITE) != 0);
  }
  /**
   * @brief returns true if write bit is set.
   * @param access is the access byte to test
   * @return true if write bit is set.
   */
  public static boolean isWrite(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_WRITE) != 0;
  }
  /**
   * @brief returns true if FIRST bit is set.
   * @return true if FIRST bit is set.
   */
  public boolean isFirst()
  {
    return ((this.mode & CA_FIRST) != 0);
  }
  /**
   * @brief returns true if FIRST bit is set.
   * @param access is the access byte to test
   * @return true if FIRST bit is set.
   */
  public static boolean isFirst(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_FIRST) != 0;
  }
  /**
   * @brief returns true if LAST bit is set.
   * @return true if LAST bit is set.
   */
  public boolean isLast()
  {
    return ((this.mode & CA_LAST) != 0);
  }
  /**
   * @brief returns true if LAST bit is set.
   * @param access is the access byte to test
   * @return true if LAST bit is set.
   */
  public static boolean isLast(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_LAST) != 0;
  }
  /**
   * @brief returns the input access byte with the appropriate bits removed
   * @param access is the access byte from which to remove the bit
   * @param bits are the access bits to remove
   * @return the input access byte with the appropriate bit removed
   */
  public static byte removeBits(byte access,short bits)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    uaccess &= ~(bits);
    return (byte)uaccess;
  }
  /**
   * @brief returns true if both the FIRST and LAST bits are set.
   * @param access is the access byte to test
   * @return true if both the FIRST and LAST bits are set.
   */
  public static boolean isSynchronous(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & (CA_LAST|CA_FIRST)) == (CA_FIRST|CA_LAST);
  }
  /**
   * Returns true if no bit is set.
   * 
   * @return
   */
  public boolean isNull()
  {
    return (this.mode == CA_NULL);
  }
  /**
   * @return true if the access contains the CA_RETRY flag
   */
  public boolean canRetry()
  {
    return ((this.mode & CA_RETRY) != 0);
  }
  /**
   * Returns true if all bits are equal.
   * 
   * @param acc
   * @return
   */
  public boolean equals(TAccess acc)
  {
    return (this.mode == acc.mode);
  }
  /**
   * Returns true if all mode bits which are set in another object are also set
   * in this object.
   * 
   * @param acc
   * @return
   */
  public boolean includes(TAccess acc)
  {
    return includes(acc.mode);
  }
  /**
   * @brief appends the input access mode flag
   */
  public void add(short mode)
  {
    this.mode |= mode;
  }
  /**
   * @brief appends the input access mode flag
   */
  public void add(TAccess acc)
  {
    this.mode |= acc.mode;
  }
  // TODO More methods (getSync etc)
}