/*
 * Created on Feb 10, 2005
 *
 * CHANGES
 * 2005-02-15 JW * Better overload support.
 * 13.7.05 PD added setHandler() method.
 */
package de.desy.tine.server.properties;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.TreeMap;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.*;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.logger.DbgLog;

/**
 * \internal
 * 
 * A Container for property data. For every property name, TPropertyList
 * contains a list of property data. Each element in that list represents an
 * instance of the property. Multiple properties can have the same name. Every
 * property has a TPropertyHandler.
 */
public class TPropertyList
{
  private int debugLevel = 0;
  private TEquipmentModule eqm = null; // parent container
  //private HashMap<String, LinkedList<PropertyData>> myPropertyData = new HashMap<String, LinkedList<PropertyData>>();
  private TreeMap<String, LinkedList<PropertyData>> myPropertyData = new TreeMap<String, LinkedList<PropertyData>>(String.CASE_INSENSITIVE_ORDER);
  private boolean deviceOriented = false;
  public boolean isDeviceOriented() { return deviceOriented; }
  public void setDeviceOriented(boolean value) { deviceOriented = value; }
  /**
   * \internal
   * 
   * This class holds the data contained in a TPropertyList.
   * 
   * @author jwlg 
   */
  public class PropertyData
  {
    private TExportProperty property;
    private TPropertyHandler handler = null;
    public PropertyData(TExportProperty property, TPropertyHandler propertyHandler)
    {
      this.property = property;
      this.handler = propertyHandler;
    }
    /**
     * @return Returns the property handler.
     */
    public TPropertyHandler getHandler()
    {
      return handler;
    }
    /**
     * @return Returns the property.
     */
    public TExportProperty getProperty()
    {
      return property;
    }
    /**
     * @param handler
     *          The handler to set.
     */
    public void setHandler(TPropertyHandler handler)
    {
      this.handler = handler;
    }
  }
  /**
   * Returns a linked list of all properties which share the same name.
   * 
   * @param propertyName
   * @return
   */
  public LinkedList<TExportProperty> getEqualProperties(String propertyName)
  {
    LinkedList<PropertyData> datalist;
    LinkedList<TExportProperty> rv = new LinkedList<TExportProperty>();
    datalist = (LinkedList<PropertyData>) myPropertyData.get(propertyName);
    if (datalist != null)
    {
      for (Iterator<PropertyData> it = datalist.iterator(); it.hasNext();)
      {
        rv.add(((PropertyData) it.next()).getProperty());
      }
    }
    return rv;
  }
  /**
   * Returns the number of properties which share the same name.
   * 
   * @param propertyName
   * @return
   */
  public int getNumInstances(String propertyName)
  {
    int rv = 0;
    LinkedList<PropertyData> propList = (LinkedList<PropertyData>) myPropertyData.get(propertyName);
    if (propList != null)
    {
      rv = propList.size();
    }
    return rv;
  }
  /**
   * Returns the number of overloads. Provided for compatibility with
   * TExportPropertyTable.
   * 
   * @param propertyName
   * @return
   */
  public int getNumOverloads(String propertyName)
  {
    return (getNumInstances(propertyName) - 1);
  }
  /**
   * Returns the number of different property names. Number of properties, not
   * counting overloads.
   * 
   * @return
   */
  public int countUniqueProperties()
  {
    return myPropertyData.size();
  }
  /**
   * Returns the total number of properties, including overloads
   * 
   * @return
   */
  public int countAllProperties()
  {
    int rv = 0;
    Iterator<LinkedList<PropertyData>> itProps;
    LinkedList<PropertyData> currentData;
    itProps = myPropertyData.values().iterator();
    while (itProps.hasNext())
    {
      currentData = itProps.next();
      if (currentData != null) // null impossible
          rv += currentData.size();
    }
    return rv;
  }
  /**
   * Return a list of PropertyData which contains all properties with a
   * specified name.
   * 
   * @param propertyName
   * @return
   */
  public LinkedList<PropertyData> getPropertyDataList(String propertyName)
  {
    return myPropertyData.get(propertyName);
  }
  /**
   * Returns the property with the specified name
   * 
   * @param propertyName
   * @return
   */
  public TExportProperty getFirstProperty(String propertyName)
  {
    LinkedList<PropertyData> lst;
    TExportProperty rv = null;
    PropertyData theData;
    lst = myPropertyData.get(propertyName);
    if (lst != null)
    {
      theData = (PropertyData) lst.getFirst();
      if (theData != null)
      { // not possible!
        rv = (TExportProperty) theData.property;
      }
    }
    return rv;
  }
  public int setHandler(String propertyName,TPropertyHandler handler)
  {
    LinkedList<PropertyData> lst;
    PropertyData pData;
    if ((lst=myPropertyData.get(propertyName)) == null) return TErrorList.illegal_property;
    if ((pData=(PropertyData)lst.getFirst()) == null) return TErrorList.code_failure;
    pData.setHandler(handler);
    return 0;
  }
  /**
   * Returns the handler of a specified property.
   * 
   * @param propertyName
   * @return
   */
  public TPropertyHandler getFirstPropertyHandler(String propertyName)
  {
    TPropertyHandler rv = null;
    LinkedList<PropertyData> lst = getPropertyDataList(propertyName);
    PropertyData theData = (PropertyData) lst.getFirst();
    if (theData != null)
    {
      rv = (TPropertyHandler) theData.handler;
    }
    return rv;
  }
  /**
   * Find a property handler for a given input and output format.
   * 
   * @param propertyName
   *          Name of the property
   * @param outputFormat
   *          Output format. TFormat.CF_NULL means any format.
   * @see TFormat
   * @return A property handler or null if no property or no suitable property
   *         handler could be found.
   */
  public TPropertyHandler findPropertyHandler(String propertyName, short outputFormat, short inputFormat)
  {
    TPropertyHandler rv = null;
    boolean outputOK, inputOK;
    PropertyData currentData;
    LinkedList<PropertyData> propList = getPropertyDataList(propertyName);
    if (propList != null)
    { // property found
      for (Iterator<PropertyData> it = propList.iterator(); it.hasNext() && (rv == null);)
      {
        currentData = (PropertyData) it.next();
        outputOK = ((outputFormat == TFormat.CF_NULL) || 
                    TFormat.formatsCanMap(currentData.getProperty().getOutputFormat(),outputFormat,false));
        inputOK = ((inputFormat == TFormat.CF_NULL) || 
                   TFormat.formatsCanMap(currentData.getProperty().getInputFormat(),inputFormat,true));
        
        if (outputOK && inputOK)
        {
          rv = currentData.getHandler();
          if (debugLevel > 1)
              DbgLog.log("findPropertyHandler","Output Format: " + currentData.getProperty().getOutputFormat()
                  + "==" + outputFormat);
          break;
        }
        else
        {
          if (debugLevel > 1)
              DbgLog.log("findPropertyHandler","Output Format: " + currentData.getProperty().getOutputFormat()
                  + "!=" + outputFormat);
        }
      }
    }
    else
    {
      DbgLog.log("findPropertyHandler","Property not found");
    }
    return rv;
  }
  /**
   * Returns true if the container contains the specified property.
   * 
   * @param propertyName
   * @return
   */
  public boolean hasProperty(String propertyName)
  {
    return myPropertyData.containsKey(propertyName);
  }
  /**
   * Returns a Set of all contained property names.
   * 
   * @return
   */
  public Set<String> getPropertyNames()
  {
    return myPropertyData.keySet();
  }
  /**
   * Adds or overloads a property.
   * 
   * @param newProperty
   * @param newHandler
   */
  public void addProperty(TExportProperty newProperty, TPropertyHandler newHandler)
  {
    addProperty(null, newProperty, newHandler);
  }
  public void addProperty(TEquipmentModule parentEqm, TExportProperty newProperty, TPropertyHandler newHandler)
  {
    if (eqm == null) eqm = parentEqm;
    String propertyName = newProperty.getName();
    LinkedList<PropertyData> propertyList;
    if (myPropertyData.containsKey(propertyName))
    { // This property already exists; overload.
      propertyList = getPropertyDataList(propertyName);
      if (propertyList != null)
      { // null impossible
        propertyList.add(new PropertyData(newProperty, newHandler));
      }
    }
    else
    { // New property
      propertyList = new LinkedList<PropertyData>();
      propertyList.add(new PropertyData(newProperty, newHandler));
      myPropertyData.put(propertyName, propertyList);
      //if ((newProperty.getDescription().getArrayType() & (TArrayType.AT_CHANNEL|TArrayType.AT_STRUCTURED)) != 0)
      { // a CHANNEL or STRUCTURE property => check for specific device names list ...
        if (eqm != null) eqm.getRegisteredNamesFromPropertyFile(newProperty.getName());
      }
    }
  }
  public short callPropertyHandler(String propertyName, String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    short cc = 0;
    short outputFormat = (dout == null) ? TFormat.CF_NULL : dout.getFormat();
    short inputFormat = (din == null) ? TFormat.CF_NULL : din.getFormat();
    TPropertyHandler theHandler = findPropertyHandler(propertyName, outputFormat, inputFormat);
    if (theHandler != null)
    { // Call handler
      if (debugLevel > 1) DbgLog.log("callPropertyHandler","Calling Property Handler for [" + propertyName + "]");
      try
      {
        cc = (short)theHandler.callHandler(devName, dout, din, devAccess);
      }
      catch (Exception e)
      {
        cc = TErrorList.code_failure;
      }
    }
    else
    {
      if (this.hasProperty(propertyName))
      { 
        if (debugLevel > 1) DbgLog.log("callPropertyHandler","handler not implemented for [" + propertyName + "]");
        cc = TErrorList.not_implemented;
      }
      else
      {
        if (debugLevel > 1) DbgLog.log("callPropertyHandler","Illegal Property [" + propertyName + "]");
        cc = TErrorList.illegal_property;
      }
    }
    return cc;
  }
}