/*
 * Created on May 14, 2003
 *
 * To change this generated comment go to 
 * Window>Preferences>Java>Code Generation>Code Template
 */
package de.desy.tine.structUtils;

import java.util.HashMap;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.types.*;

/**
 * \internal
 * 
 * @brief Global struct registry This class provides static methods which give
 *        access to a global tag structure registry.
 */
public final class TStructRegistry
{
  private static HashMap<String, TStructDescription> myRegistry = new HashMap<String, TStructDescription>();
  public static int numRegisteredStructs = 0;
  public static NAME16II[] getRegisteredStructure(String tag)
  {
    NAME16II[] rv = null;
    TStructDescription dsc = (TStructDescription) myRegistry.get(tag);
    if (dsc != null)
    {
      rv = dsc.toNameIntIntArray();
    }
    return rv;
  }
  /**
   * @brief Registers an element of a tagged structure If a data acquisition
   *        needs to make use of user-defined structures, it is strongly
   *        suggested to register such structures at both the server and client
   *        ends. This entails passing the structure information at
   *        initialization time and giving the structure a 'tag' (i.e. an
   *        8-character name). Structure information is passed by calling the
   *        fill() method for each element of the structure in succession. When
   *        the entire structure has been registered the fill() method should be
   *        call once more with a a TFormat.CF_NULL data type, which effectively
   *        closes the structure registration.
   * @param tag
   *          is the (up to) eight character structure tag name
   * @param fmt
   *          is the de.desy.tine.client Data type appropriate to the structure element
   * @param siz
   *          is the array size of the structure element \b Example: \include
   *          eg_struct1.java
   */
  public static void fill(String tag, short fmt, int siz)
  {
    TStructDescription sd = (TStructDescription) myRegistry.get(tag);
    if (sd == null)
    {
      sd = new TStructDescription(tag);
      myRegistry.put(sd.getTagName(), sd);
      sd.beginDefinition();
    }
    if ((fmt % 512) == TFormat.CF_NULL) // registration finished
    {
      sd.setArraySize((siz));
      sd.endDefinition();
    }
    else
    {
      sd.addField(fmt, siz);
    }
  }
  public static void fill(String tag, String name, short fmt, int siz)
  {
    TStructDescription sd = (TStructDescription) myRegistry.get(tag);
    if (sd == null)
    {
      sd = new TStructDescription(tag);
      myRegistry.put(sd.getTagName(), sd);
      sd.beginDefinition();
    }
    if ((fmt % 512) == TFormat.CF_NULL) // registration finished
    {
      sd.setArraySize((siz));
      sd.endDefinition();
    }
    else
    {
      sd.addField(name, fmt, siz);
    }
  }
  /**
   * Converts a byte array to a string. This method uses the description of a
   * tagged structure to interpret the contents of a given byte array. It
   * assumes that the byte array contains an arbitrary number of elements of the
   * type described by the given structure.
   * 
   * @param tag
   *          is the (up to) eight character structure tag name
   * @param data
   *          is a byte array reference to the tagged structure's data.
   * @see de.desy.tine.client.TStructDescription.converteBytesToString();
   */
  public static String toString(String tag, byte[] data)
  {
    if (data == null) return "";
    return toString(tag,data,data.length);
  }
  public static String toString(String tag, byte[] data,int length)
  {
    String rv = "";
    StringBuffer buffer = new StringBuffer();
    int nStructs;
    TStructDescription sd = (TStructDescription) myRegistry.get(tag);
    if (sd == null)
    {
      rv = new String("Structure " + tag + " not registered");
    }
    else if ((data == null) || (data.length == 0))
    {
      rv = "";
    }
    else
    { // Tag exists, data valid
      nStructs = length / sd.getRawLength();
      for (int i = 0; i < nStructs; i++)
      {
        buffer.append(sd.convertBytesToString(data, i * sd.getRawLength()));
      }
      rv = buffer.toString();
    }
    return rv;
  }
  public static int getSizeInBytes(String tag)
  {
    TStructDescription s = (TStructDescription) myRegistry.get(tag);
    return (s == null) ? -1 : s.getRawLength();
  }
  /**
   * @see TStructRegistry.add()
   */
  public static synchronized void add(TStructDescription newStruct)
  {
    if (!contains(newStruct.getTagName()))
    {
      myRegistry.put(newStruct.getTagName(), newStruct);
    }
  }
  /**
   * @see TStructRegistry.get()
   */
  public static TStructDescription get(String structName)
  {
    return (TStructDescription) myRegistry.get(structName);
  }
  /**
   * @see TStructRegistry.contains()
   */
  public static boolean contains(String structName)
  {
    return myRegistry.containsKey(structName);
  }
}