/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package de.desy.tine.structUtils;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.types.TCompoundDataObject;
import java.util.*;
import java.io.*;

/**
 * @brief Tagged Structure 
 *
 * The transfer of 'user-defined' data structures is allowed in TINE. 
 * In order to avoid nasty problems concerning alignment and byte swapping
 * it is strongly suggested to extend the TTaggedStructure class when
 * creating a data structure.
 */
public class TTaggedStructure implements Cloneable
{
  private String structName;
  private boolean registered = false;
  private int sizeInBytes;   
  private ArrayList<structField> fieldArrayList = null; 
  static private int indent = 0;  // used by toString()
  
  @Override
  public Object clone()
  {
    Object o = null;
    try { o = super.clone(); }
    catch (CloneNotSupportedException e)
    {
      System.err.println("TTaggedStructure cannot clone: " + e);
    }
    return o;
  }
  
  /**
   * @brief Creates a tagged structure with the given name.
   * 
   * @param name is the name to be used as the structure tag.
   * Note that only the first 16 characters of the given name will
   * be used as the tag name!
   * 
   * @note The addField() method must be called for each element of
   * the structure data element.
   */
  public TTaggedStructure(String name) 
  {
    structName = new String(name);
    fieldArrayList = new ArrayList<structField>();
    
    if (TStructRegistry.contains(structName))
    {
      registered = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);    
    }
    else sizeInBytes = 0;   
  }
  /**
   * @brief Creates a tagged structure using the class name as the 
   * structure tag.
   * Note that only the first 16 characters of the class name will
   * be used as the tag name!
   * 
   * @note The addField() method must be called for each element of
   * the structure data element.
   * 
   * \b Example:
   * @include eg_tineTaggedStructure1.java
   */
  public TTaggedStructure() 
  {
    String name = this.getClass().getName();
    StringTokenizer t = new StringTokenizer(name,".$");
    String lastToken = name;
    while (t.hasMoreTokens())
    {
      lastToken  = t.nextToken();
    }
    structName = new String(lastToken);   
    fieldArrayList = new ArrayList<structField>();
    if (TStructRegistry.contains(structName))
    {
      registered = true;
      sizeInBytes = TStructRegistry.getSizeInBytes(structName);
    }    
    else sizeInBytes = 0;
  }
  
  /**
   * @brief Returns the structure's assigned tag name. 
   * 
   * @return The structure's assigned tag name as a string.
   */
  public String getName() { return structName; }
  
  /**
   * @brief Returns the structure's data size in bytes. 
   * 
   * @return The structure's data size in bytes.
   */
  public int getSizeInBytes() { return sizeInBytes; }
  
  
  public byte[] toByteArray()
  {
    ByteArrayOutputStream boas = new ByteArrayOutputStream(getSizeInBytes());
    boas.reset();
    DataOutputStream ds = new DataOutputStream(boas);
    
    try
    {
        for( int i=0; i<fieldArrayList.size(); i++)
        {
            fieldArrayList.get(i).writeToOutputStream(ds);            
        }
        ds.close();
    }
    catch( IOException e)
    {
        MsgLog.log("TTaggedStructure", "---Internal error: can not convert TaggedStructure: " + getName() + " to byte array ",TErrorList.code_failure,e,0);
        e.printStackTrace();
        return null;
    }
    return boas.toByteArray();       
  }  
  
  public byte[] getByteArray() { return toByteArray(); }
  
  public void toStruct(byte[] data)
  {
    ByteArrayInputStream dis = new ByteArrayInputStream(data);
    DataInputStream ds = new DataInputStream(dis);
   
    try
    {         
      for (int i=0; i<fieldArrayList.size(); i++)
      {
        fieldArrayList.get(i).readFromInputStream(ds);      
      }
      ds.close(); 
      dis.close();
    }
    catch (IOException e)
    {
     MsgLog.log("TTaggedStructure", "---Internal error: can not convert input data stream to TaggedStructure: " + getName(),TErrorList.code_failure,e,0);
     e.printStackTrace();
    }
  }
  
  /* @brief returns a string representation of this tagged structure
   */
  @Override
  synchronized public String toString()
  {
    StringBuffer sb = new StringBuffer("");
    structField sf;
    String indentString = new String("");
    indent += 3;
    for( int i = 0; i < indent; i++ ) indentString += " ";
    
    sb.append( "\n" + indentString + "Struct: " + getName() + "\n" );
    sb.append(indentString + "{\n");
    for (int i=0; i<fieldArrayList.size(); i++)
    {
      sf = (structField)fieldArrayList.get(i);
      sb.append( indentString +" "+ sf.getName() + " ->" );
      sb.append( sf.toString() );
      sb.append("\n");
    }
    sb.append( indentString + "}");
    indent -= 3;
    return sb.toString();
  } 
   
    /**
   * @brief Terminates the tagged structure definition and signals that
   * the tagged structure is ready for use. 
   * 
   * This method plays an integral role in defining the tagged
   * structure.  It should be called following all calls to addField().
   */
  public void initDone() 
  {
    if (!registered)
    {
      TStructRegistry.fill(structName,TFormat.CF_NULL,1);
      registered = true;
    }
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(byte[] field, String fieldName)
  {  
    fieldArrayList.add( new byteArrayField( field, fieldName ));    
   
    if( !registered ) 
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_BYTE, field.length );    
        sizeInBytes += field.length;
    }          
    
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
 public void addField(char[] field, String fieldName)
 {    
    fieldArrayList.add( new charArrayField( field, fieldName ));                
    if( !registered )
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.length  );       
        sizeInBytes += field.length;
    }      
  }
  
 /**
  * @param field is the structure field to add to the tagged structure
  * @param fieldName is the name of the structure field
  */
  public void addField(short[] field, String fieldName)
  { 
    fieldArrayList.add( new shortArrayField( field, fieldName ));            
          
    if( !registered )
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_SHORT, field.length );       
        sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_SHORT);
    }    
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(int[] field, String fieldName)
  {       
    fieldArrayList.add( new intArrayField( field, fieldName ));  
    if( registered ) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_LONG, field.length );       
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_LONG);               
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(float[] field, String fieldName)
  { 
    fieldArrayList.add( new floatArrayField( field, fieldName ));   
    if( registered ) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_FLOAT, field.length );       
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_FLOAT);    
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(double[] field, String fieldName)
  { 
    fieldArrayList.add( new doubleArrayField( field, fieldName ));            
    if( registered ) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_DOUBLE, field.length );       
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_DOUBLE); 
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(long[] field, String fieldName)
  { 
    fieldArrayList.add( new longArrayField( field, fieldName ));            
    if( registered ) return;
    TStructRegistry.fill(structName, fieldName, TFormat.CF_DLONG, field.length );       
    sizeInBytes += field.length * TFormat.formatSizeOf(TFormat.CF_DLONG); 
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(TString field, String fieldName)
  {       
    fieldArrayList.add( new TStringField( field, fieldName ));            
    
    if( !registered ) 
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.getCapacity() );       
        sizeInBytes += field.getCapacity();     
    }       
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField(StringBuffer field, String fieldName)
  {    
    fieldArrayList.add( new StringBufferField( field, fieldName ));            
    if( !registered )
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_TEXT, field.capacity());       
        sizeInBytes += field.capacity();  
    }        
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField( TCompoundDataObject field, String fieldName )
  {  
    fieldArrayList.add( new TCompoundDataField( field, fieldName ));                
    if( !registered )
    {
        TStructRegistry.fill(structName, fieldName, field.getFormat(), 1 );       
        sizeInBytes += field.getSizeInBytes(); 
    }        
  }
 
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField( TCompoundDataObject[] field, String fieldName )
  {  
    fieldArrayList.add( new TCompoundDataArrayField( field, fieldName ));                
    
    if( !registered ) 
    {
        TStructRegistry.fill(structName, fieldName, field[0].getFormat(), field.length );       
        sizeInBytes += field[0].getSizeInBytes() * field.length; 
    }    
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField( TTaggedStructure field, String fieldName )
  { 
    int size = TStructRegistry.getSizeInBytes( field.getName());      
    if( size < 1 )  
    {
        MsgLog.log("TTaggedStructure","---Error: attempt to register a field: " + fieldName + " of unknown data type: " + field.getName(),TErrorList.code_failure,null,0);
        return;
    }
    fieldArrayList.add( new TaggedStructureField( field, fieldName ));  
    if( !registered )
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_STRUCT, size );       
        sizeInBytes += size;     
    }      
  }
  
  /**
   * @param field is the structure field to add to the tagged structure
   * @param fieldName is the name of the structure field
   */
  public void addField( TTaggedStructure[] field, String fieldName )
  { 
    int size = TStructRegistry.getSizeInBytes( field[0].getName());  
    if( size < 1 )  
    {
        MsgLog.log("TTaggedStructure","attempt to register a field: " + fieldName + " being an array of unknown data type: " + field[0].getName(),TErrorList.code_failure,null,0);
        return;
    }
    size = size * field.length;
    fieldArrayList.add( new TaggedStructureArrayField( field, fieldName )); 
    if( !registered )
    {
        TStructRegistry.fill(structName, fieldName, TFormat.CF_STRUCT, size );       
        sizeInBytes += size;     
    }        
  }
  

  /*
   *General Byte Alignment Rules
    -------Taken from:  http://www.eventhelix.com/RealtimeMantra/ByteAlignmentAndOrdering.htm

    The following byte padding rules will generally work with most 32 bit processor. You should consult your compiler and microprocessor manuals to see if you can relax any of these rules.

    - Single byte numbers can be aligned at any address
    - Two byte numbers should be aligned to a two byte boundary
    - Four byte numbers should be aligned to a four byte boundary
    - Structures between 1 and 4 bytes of data should be padded so that the total structure is 4 bytes.
    - Structures between 5 and 8 bytes of data should be padded so that the total structure is 8 bytes.
    - Structures between 9 and 16 bytes of data should be padded so that the total structure is 16 bytes.
    - Structures greater than 16 bytes should be padded to 16 byte boundary. 
   * 
   * Karol's comment: it seems not to work for Linux GCC...
   * For now I decide to limit the padding to 4 bytes
   * 
  */

//private void checkPadding()
//{
//    return;
////    if( offset < 1 ) return;
////    
////    int n = offset % 4;
////    if( n == 0 ) return;
////    addField( new byte[4-n], "reserved");
//}

/*private void checkPadding_not_working()
{
    if( offset < 1 ) return;
    if( offset < 4 )
    {
        addField( new byte[ 4 - offset ], "reserved");
        return;
    }
    if( offset < 8 )
    {
        addField( new byte[ 8 - offset ], "reserved");
        return;
    }
    if( offset < 16 )
    {
        addField( new byte[ 16 - offset ], "reserved");
        return;
    }
    
    int n = offset % 16;
    if( n == 0 ) return;
    addField( new byte[16-n], "reserved");
}
*/

  
  
/*
  old version:
  private int padding = 0;
  private void checkAligment(int size)
  {
    //check the size for the correct allignment. Add some placeholders if needed
    int d = size % 4;
    int fieldPadding = 4 - d;
    if( d > 0 )  
    { // there are extra padding bytes
      if (padding > 0 && d > padding)
      { // too many bytes to append to current padding set so clear it               
        padding = fieldPadding; 
        addField( new byte[padding], "reserved");        
      }
      if (padding == 0) padding = 4;
      padding -= d; 
    }
    else
    { // new field is aligned; are there padding bytes?
      if (padding > 0 && padding < 4)
      {
        padding = fieldPadding;             
        addField( new byte[padding], "reserved");        
      }
    }
  }
 */
  
  //----------------------------------------------
    private abstract class structField
    {
        private String name;
        public String getName()
        {
            return name;
        }
        public void setName(String name)
        {
            this.name = name;
        }
        abstract void writeToOutputStream( DataOutputStream ds ) throws IOException;
        abstract void readFromInputStream(DataInputStream ds) throws IOException;
        @Override
        abstract public String toString();
    }

    private class byteArrayField extends structField
    {
        byte[] data;
        byteArrayField(byte[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        {
             ds.read(data,0,data.length);
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            ds.write(data, 0, data.length );
        }

        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");    
            int ibyte;
            for (int i = 0; i < data.length; i++)
            {
                if ((ibyte = data[i]) < 0) ibyte += 256; 
                sb.append(" " + ibyte);
            } 
            return sb.toString();
        }
    }
    
    private class charArrayField extends structField
    {
        char[] data;
        charArrayField(char[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        {            
            for (int i = 0; i < data.length; i++)
            {
               data[i] = (char)(ds.readByte());              
            }             
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for (int i = 0; i < data.length; i++)
            {
              ds.writeByte((byte)(data[i]));
            }
        }
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length && data[i] != 0; i++)
            {   
                sb.append(data[i]);
            } 
            return sb.toString().trim();
        }
    }
    private class shortArrayField extends structField
    {
        short[] data;
        shortArrayField(short[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        {            
            for (int i = 0; i < data.length; i++)
            {
               data[i] = Swap.Short(ds.readShort());              
            }             
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for (int i = 0; i < data.length; i++)
            {
              ds.writeShort(Swap.Short(data[i]));
            }
        }
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++)
            {                
                sb.append(" " + data[i]);
            } 
            return sb.toString();
        }
    }
    private class intArrayField extends structField
    {
        int[] data;
        intArrayField(int[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        {            
            for (int i = 0; i < data.length; i++)
            {
               data[i] = Swap.Long(ds.readInt());              
            }             
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for (int i = 0; i < data.length; i++)
            {
              ds.writeInt(Swap.Long(data[i]));
            }
        }
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++)
            {                
                sb.append(" " + data[i]);
            } 
            return sb.toString();
        }
    }
    private class floatArrayField extends structField
    {
        float[] data;
        floatArrayField(float[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        { 
            byte[] buffer = new byte[4];            
            for( int i=0; i<data.length; i++ )
            {      
              ds.read(buffer);
              data[i] = Swap.Float(buffer);              
            }                        
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for (int i = 0; i < data.length; i++)
            {
              ds.write(Swap.Float(data[i]));
            }
        }
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++)
            {                
                sb.append(" " + data[i]);
            } 
            return sb.toString();
        }
    }
    private class doubleArrayField extends structField
    {
        double[] data;
        doubleArrayField(double[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        { 
            byte[] buffer = new byte[8];            
            for( int i=0; i<data.length; i++ )
            {      
              ds.read(buffer);
              data[i] = Swap.Double(buffer);              
            }                        
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for (int i = 0; i < data.length; i++)
            {
              ds.write(Swap.Double(data[i]));
            }
        }
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++)
            {                
                sb.append(" " + data[i]);
            } 
            return sb.toString();
        }
    }
    
    private class longArrayField extends structField
    {
        long[] data;
        longArrayField(long[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        { 
            byte[] buffer = new byte[8];            
            for( int i=0; i<data.length; i++ )
            {      
              ds.read(buffer);
              data[i] = Swap.DLong(buffer);              
            }                        
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for (int i = 0; i < data.length; i++)
            {
              ds.writeLong(Swap.DLong(data[i]));
            }
        }
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++)
            {                
                sb.append(" " + data[i]);
            } 
            return sb.toString();
        }
    }
    
    private class TStringField extends structField
    {
        TString data;
        TStringField(TString data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        {             
            byte[] b = new byte[data.getCapacity()];
            ds.read(b);
            data.set(b);
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            ds.write(data.getBytes());            
        }
        @Override
        public String toString()
        {            
            return data.toString();
        }
    }
    private class StringBufferField extends structField
    {
        StringBuffer data;
        StringBufferField(StringBuffer data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream( DataInputStream ds) throws IOException
        { 
            for( int i = 0; i < data.capacity(); i++ ) data.setCharAt(i, (char)ds.readByte());   
        }
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {           
            for( int i = 0; i < data.capacity(); i++ ) ds.writeByte((byte)(data.charAt(i)));
        }
        @Override
        public String toString()
        {            
            return data.toString();
        }
    }
    private class TCompoundDataField extends structField
    {
        TCompoundDataObject data;
        TCompoundDataField(TCompoundDataObject data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream(DataInputStream ds) throws IOException
        {
            byte[] b = data.getByteArray();
            ds.read(b);
            data.toStruct();
        }
        
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            ds.write(data.toByteArray());
        }

        @Override
        public String toString()
        {
            return data.toString();
        }        
    }  
    
    private class TCompoundDataArrayField extends structField
    {
        TCompoundDataObject[] data;
        TCompoundDataArrayField(TCompoundDataObject[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream(DataInputStream ds) throws IOException
        {
            for( int i = 0; i < data.length; i++ )
            {
                byte[] b = data[i].getByteArray();
                ds.read(b);
                data[i].toStruct();
            }
        }
        
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for( int i = 0; i < data.length; i++ ) ds.write(data[i].toByteArray());
        }

        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++) sb.append(" " + data[i]);
            return sb.toString();
        }        
    } 
    
    private class TaggedStructureField extends structField
    {
        TTaggedStructure data;
        TaggedStructureField(TTaggedStructure data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream(DataInputStream ds) throws IOException
        {
            byte[] buffer = new byte[data.getSizeInBytes()];
            ds.read(buffer);
            data.toStruct(buffer);
        }        
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            ds.write( data.toByteArray() );
        }

        @Override
        public String toString()
        {
            return data.toString();
        }        
    } 
    
    private class TaggedStructureArrayField extends structField
    {
        TTaggedStructure[] data;
        TaggedStructureArrayField(TTaggedStructure[] data, String name)
        {
            setName(name);
            this.data = data;
        }
        @Override
        void readFromInputStream(DataInputStream ds) throws IOException
        {            
            byte[] buffer = new byte[data[0].getSizeInBytes()];
            for( int i = 0; i < data.length; i++ )
            {                
                ds.read(buffer);
                data[i].toStruct(buffer);
            }
        }
        
        @Override
        void writeToOutputStream(DataOutputStream ds) throws IOException
        {
            for( int i = 0; i < data.length; i++) ds.write( data[i].toByteArray() );
        }

        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("");                
            for (int i = 0; i < data.length; i++) sb.append(" " + data[i]);
            return sb.toString();            
        }        
    } 
}
