package de.desy.tine.types;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;

/**
 * TINE type CF_IMAGE java class implementation. Documentation mainly done by
 * Stefan Weisse. Class design and implementation by Philip Duval, based on
 * negotiations and agreement (between DESY Zeuthen-based controls group (now
 * ERS), MST (now MCS 1) and MVP (now MCS 4)) on a defined control system type
 * for video image data transport.
 * 
 * @author Philip Duval (pd)
 * @author Stefan Weisse (sw)
 * @version 1.1 Jan 28, 2009
 * @version last documented Jan 28, 2009
 */
public class IMAGE implements TCompoundDataObject
{
  private static long UnsignedInt(int ival)
  {
    long lval = (long) ival;
    if (lval < 0) lval += 4294967296L;
    return lval;
  }
  /**
   * Subclass of first part of header (subclass used for logical grouping of
   * class member variables which map almost 1:1 to header structures
   * transferred in binary form.
   * 
   * @author Philip Duval (pd)
   * @author Stefan Weisse (sw)
   */
  public class SourceHeader
  {
    /**
     * Magic ID in front of binary data block. Value is fixed to four character
     * code 'VSV3' (little endian: 33565356 hex). VSV3 stands for VideoSystem
     * Version 3.
     * <p>
     * In C/C++ world, member is an unsigned int32.
     */
    public int baseTag;
    /**
     * Unique ID that identifies one single location of image creation, usually
     * named Image Source in VSV3. There must not be two image sources sharing
     * the same camera port Id. If cameraPortId is set to 0 it means that no
     * camera port id is attached to the image buffer so Image Source (origin of
     * image) is undefined.
     * <p>
     * In C/C++ world, member is an unsigned int32.
     */
    public long cameraPortId; // uint32->long sw!
    /**
     * Defines the version of the IMAGE header. At the moment only version 1 is
     * defined so this tag is fixed to '1'. Applications are asked to check this
     * version number whether it meets their minimum requirements. It is
     * foreseen that future header versions will be backward compatible. New
     * applications should be able to work with old headers and new headers. Old
     * applications can only work with old headers. Checking is performed based
     * on integer comparison.
     * <p>
     * <code>if (versionTag == 0x1) { good_header_version } else
     *  { bad header }</code>
     * <p>
     * In C/C++ world, this header member is an unsigned int32.
     */
    public long versionTag; // uint32->long sw!
    /**
     * Contains the whole length of the data block as it would be transferred
     * bytewise on network level. This value contains physical amount of bytes
     * necessary for sourceHeader, for frameHeader and the net size of appended
     * frame bits. If no frame bits are appended, the size of sourceHeader plus
     * frameHeader must be contained within.
     * <p>
     * For practical reason, current C/C++ implementations of the VideoSystem V3
     * are limited to 6 Megabytes of frame bits size plus header size
     * (sourceHeader + frameHeader). But the general design rule is that an
     * application should never expect to receive image frames that are less
     * than or equal to 6 megabytes in size.
     * <p>
     * In C/C++ world, this header member is an unsigned int32.
     */
    public long totalLength; // uint32->long sw!
    /**
     * Contains the time the image was originally taken at image source, UTC
     * seconds part. The design rule is that this timestamp should be acquired
     * as early as possible after image acquire from hardware sensor. Some
     * digital cameras are even capable of tagging image data with a proper
     * timestamp. This timestamp, as it is most precise one can get, should be
     * transformed and copied here. The aim of the game is not to take a second
     * timestamp in source code once the image is received in software, except if
     * no other possibility exist.
     * <p>
     * In C/C++ world, this header member is an unsigned int32.
     */
    public int timestampSeconds; 
    /**
     * Contains the time the image was originally taken at image source,
     * microseconds offset to UTC time. The same design rules as for
     * timestampSeconds member applies. If the granularity of system clock or
     * camera clock (if image bits are time-tagged at hardware) is not good
     * enough to deliver microseconds, close-by values like milliseconds or
     * tenths of a second are also fine.
     * <p>
     * In C/C++ world, this header member is an signed int32.
     */
    public int timestampMicroseconds;
    /** < image timestamp (micro seconds) */
    /**
     * A human-readable string giving imaginative representation of image
     * source. Image source here does not only mean Image Source in VSv3 terms
     * (position of image acquire location), but might also contain a string
     * describing file name where the image was taken from (if loaded).
     * <p>
     * In C/C++ world, this header member is a 64 char array (64 bytes). It may
     * not contain a null byte at the end, but 64 truly useful characters!
     */
    public String cameraPortName = "unknown"; // [64];
    /**
     * Sets the totalLength member. This function never fails.
     * 
     * @param lengthInBytes
     *          passed totalLength value
     * @see #totalLength
     */
    public void setTotalLength(int lengthInBytes)
    {
      totalLength = lengthInBytes;
    }
    public void setTotalLength(long lengthInBytes)
    {
      totalLength = lengthInBytes;
    }
    /**
     * Sets the timestamp member variables (UTC seconds and microseconds). This
     * function never fails.
     * 
     * @param time
     *          Contains UTC time*1000 plus fractional millisecond offset*1000.
     * @see #timestampSeconds
     * @see #timestampMicroseconds
     */
    public void setTimeStamp(long time)
    {
      timestampSeconds = (int) (time / 1000);
      timestampMicroseconds = (int) ((time % 1000) * 1000);
    }
    public void setStaticValues(int baseId, int specId, int version, String cameraPort)
    {
      this.baseTag = baseId;
      this.cameraPortId = specId;
      versionTag = version;
      cameraPortName = cameraPort;
    }
    public void setStaticValues(int baseId, long specId, long version, String cameraPort)
    {
      this.baseTag = baseId;
      this.cameraPortId = specId;
      versionTag = version;
      cameraPortName = cameraPort;
    }
    public String toString()
    {
      String s = new String("ha !");
      // double v = 0;
      // String s = "";
      // Field[] flds = this.getClass().getFields();
      // for (Field f : flds)
      // {
      // try
      // {
      // s += f.getName() + " : " + f.get(v) + "\n";
      // }
      // catch (Exception e)
      // {
      // e.printStackTrace();
      // }
      // }
      return s;
    }
  }
  /**
   * Subclass of second (and last) part of header (subclass used for logical
   * grouping of class member variables which map almost 1:1 to header
   * structures physically transferred in binary form.
   * <p>
   * This class public member variables are mainly a clone of the
   * TTF2_IMAGE_STRUCT defined and agreed in 2008 between ERS, MCS 1 and MCS 4.
   * 
   * @author Philip Duval (pd)
   * @author Stefan Weisse (sw)
   */
  public class FrameHeader
  {
    /**
     * Width in pixels of the source image (if aoi is specified) or the raw
     * appended image size. The variable is unnecessary if an aoi is set and
     * very necessary otherwise. See aoiWidth, aoiHeight, xStart, yStart and
     * sourceWidth.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely, but possible), value is '-1'.
     */
    public int sourceWidth;
    /**
     * Width in pixels of the source image (if aoi is specified) or the raw
     * appended image size. The variable is unnecessary if an aoi is set and
     * very necessary otherwise. See aoiWidth, aoiHeight, xStart, yStart and
     * sourceHeight.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely, but possible), value is '-1'.
     */
    public int sourceHeight;
    /**
     * Width in pixels of the area of interest. A special mechanism is
     * implemented in a CF_IMAGE. If at the image source it can be decided that
     * only a part of the sensor is read out or contains significant information
     * (invaluable surroundings (mounting plates etc.) does not need to get
     * transferred to remote peers), not the whole image may be transferred but
     * only a rectangular region out of it. Even this rectangular piece is break
     * down into two pieces: horizontal truncation (sourceWidth, aoiWidth and
     * xStart) and vertical truncation (sourceHeight, aoiHeight and yStart).
     * <p>
     * If aoiWidth is specified (bigger than zero), it defines the number of
     * horizontal pixels that are attached to the header as frame. The
     * sourceWidth member in that case should be bigger than aoiWidth and
     * defines the source width in pixels (at source level) before an aoi was
     * cut out. The whole mechanism is referred to sometimes as 'aoi cutout'.
     * <p>
     * If aoiWidth is set to a value of 0 or less, it is treated as NOT set. If
     * xStart is bigger than zero the aoi is considered to be set in horizontal
     * dimension. If aoiWidth is zero or less in that case, the header is
     * considered to be NOT valid!
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very possible), value is less than or equal to zero. A recommendation
     * for unset value is '-1', though.
     */
    public int aoiWidth;
    /**
     * Height in pixels of the area of interest. A special mechanism is
     * implemented in a CF_IMAGE. If at the image source it can be decided that
     * only a part of the sensor is read out or contains significant information
     * (invaluable surroundings (mounting plates etc.) does not need to get
     * transferred to remote peers), not the whole image may be transferred but
     * only a rectangular region out of it. Even this rectangular piece is break
     * down into two pieces: horizontal truncation (sourceWidth, aoiWidth and
     * xStart) and vertical truncation (sourceHeight, aoiHeight and yStart).
     * <p>
     * If aoiHeight is specified (bigger than zero), it defines the number of
     * vertical pixels that are attached to the header as frame. The
     * sourceHeight member in that case should be bigger than aoiHeight and
     * defines the source width in pixels (at source level) before an aoi was
     * cut out. The whole mechanism is referred to sometimes as 'aoi cutout'.
     * <p>
     * If aoiHeight is set to a value of 0 or less, it is treated as NOT set. If
     * yStart is bigger than zero the aoi is considered to be set in horizontal
     * dimension. If aoiHeight is zero or less in that case, the header is
     * considered to be NOT valid!
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very possible), value is less than or equal to zero. A recommendation
     * for unset value is '-1', though.
     */
    public int aoiHeight;
    /**
     * Pixel index location (counting starts from zero) of left index from
     * top-left position if an aoi is set. If no aoi is set (here: aoiWidth <=
     * 0); xStart is defined to be exactly zero. If it is not, the header is
     * invalid. xStart plus aoiWidth must be less than or equal to sourceWidth.
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very possible), value must be zero. A value of less than zero is
     * invalid!
     * 
     * @see #aoiWidth
     * @see #sourceWidth
     * @see #aoiHeight
     * @see #yStart
     * @see #sourceHeight
     */
    public int xStart;
    /**
     * Pixel index location (counting starts from zero) of top index from
     * top-left position if an aoi is set. If no aoi is set (here: aoiHeight <=
     * 0); xStart is defined to be exactly zero. If it is not, the header is
     * invalid. yStart plus aoiHeight must be less than or equal to
     * sourceHeight. In C/C++ world, this member variable is a signed int32. If
     * it is unset (very possible), value must be zero. A value of less than
     * zero is invalid!
     * 
     * @see #aoiHeight
     * @see #sourceHeight
     * @see #aoiWidth
     * @see #xStart
     * @see #sourceWidth
     */
    public int yStart;
    /**
     * Specifies amount of bytes per pixel. Currently defined values are 1
     * (exactly one byte per pixel), 2 (exactly two bytes per pixel), 3 (three
     * bytes per pixel) and 4 (four bytes per pixel). Fractional bytes per pixel
     * (as with packed 12 bit grayscale values) are not supported at the moment.
     * It is foreseen to use the '0' to specify that.
     * <p>
     * Generally speaking, header version 1 supports raw grayscale images of 1,2
     * and 3 bytes per pixel and raw RGB colour images of 3 or 4 bytes per pixel
     * (24 bits or 24 bits padded to 32 bits).
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely), value must be '-1'. Values of less than one or bigger than
     * four are treated as illegal for header version 1.
     */
    public int bytesPerPixel;
    /**
     * Specifies amount of effective bits per pixel of appended image bits. For
     * performance reasons, physical resolution of 12 bits per pixel (in example
     * Prosilica GE / GC-type monochrome cameras) is wrapped into 16 bit
     * integers for improving handling by computers. For such scheme effective
     * bits per pixel value was invented to distinguish valid from invalid pixel
     * values and to have more input for image processing algorithms. The value
     * should be set appropriate. If it is not set, it is believed to be set to
     * the maximum value the current data format is capable of. If the value is
     * bigger than theoretically possible for a given bytesPerPixel member, the
     * header is considered to be invalid.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely), value must be '-1'. A value of zero is treated as 'unset' for
     * header version 1.
     */
    public int effectiveBitsPerPixel;
    /**
     * Pure informal value describes whether integer binning in horizontal
     * direction was performed on the image. It is not said whether this binning
     * was performed by hardware or software.
     * <ul>
     * <li>0 - binning status unknown
     * <li>1 - no binning (known status!)
     * <li>2 - 1 pixel was created out of 2 pixels (2:1)
     * <li>3 - 1 pixel was created out of 3 pixels (3:1)
     * <li>4 - 4:1
     * <li>...
     * </ul>
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very likely), value must be '0'. If value is less than zero header is
     * considered invalid.
     */
    public int horizontalBinning;
    /**
     * Pure informal value describes whether integer binning in vertical
     * direction was performed on the image. It is not said whether this binning
     * was performed by hardware or software.
     * <ul>
     * <li>0 - binning status unknown
     * <li>1 - no binning (known status!)
     * <li>2 - 1 pixel was created out of 2 pixels (2:1)
     * <li>3 - 1 pixel was created out of 3 pixels (3:1)
     * <li>4 - 4:1
     * <li>...
     * </ul>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very likely), value must be '0'. If value is less than zero, header is
     * considered invalid.
     */
    public int verticalBinning;
    /**
     * Rather informal value declaring the image format taken by camera hardware
     * at image source. Once set at image acquire level, this member should not
     * be changed. One practical case is that colour information was originally
     * delivered by sensor as bayer pattern but decoded by image acquire
     * software to 24 bit RGB data, which was subsequently be appended to this
     * image header. The image source format should be bayer in that case.
     * Certain formats are implemented at the moment, see {@link #imageFormat}.
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (unlikely), value must be '-1'. If value is less than -1, header is
     * considered invalid.
     */
    public int sourceFormat;
    /**
     * Very important member declaring the binary format in which image bits,
     * appended to this header taken by camera, are encoded. In general, its
     * value can be very different from sourceFormat value. It should always
     * reflect encoding type of image bits. Various formats can be set:
     * <ul>
     * <li>0 - Grayscale
     * <li>1 - RGB (24bpp)
     * <li>2 - RGB (32bpp)
     * <li>3 - RGB (15bpp)*
     * <li>4 - RGB (16bpp)*
     * <li>5 - JPEG
     * <li>6 - TIFF*
     * <li>7 - YUV411*
     * <li>8 - YUV422*
     * <li>9 - YUV444*
     * <li>10 - HUFFYUV
     * <li>11 - BMP (Windows BMP)*
     * <li>11 - BAYER*
     * </ul>
     * <p>
     * - not specified in more detail up to now, should be used as informal
     * values at {@link #sourceFormat} only
     * <p>
     * In C/C++ world, this member variable is a signed int32. If it is unset
     * (very unlikely), value must be '-1'. If value is less than -1, header is
     * considered invalid.
     */
    public int imageFormat;
    /**
     * Specifies unique frame identification number. This number is not really
     * unique. Starting at one, for each frame (as early as possible after
     * acquiring, maybe even done inside camera hardware) the image is tagged
     * with this number. The next image (no losses in between!) should be tagged
     * with number+1 and so on. By performing this method, it can be checked
     * later whether a sequence contains really subsequent frames as they were
     * acquired at source level and no video frame is lost in between.
     * <p>
     * In C/C++ world, this member variable is an unsigned int32. If it is
     * unset, value must be '0'.
     */
    public long frameNumber; // uint32->long sw!
    /**
     * Specifies event number when frame was taken. In some modern timing
     * systems, together with the main repetition rate (triggered accelerator
     * like PITZ) an information is modulated on the trigger signal and properly
     * decoded to computer registers by timing hardware. This information can be
     * helpful in joining or synchronization cases. One use-case is DAQ
     * correlation ("where did the lightning hit first before a lot of systems
     * went crazy").
     * <p>
     * The header members eventNumber, frameNumber, timeStamp and cameraPortId
     * build up a primary key which should identify an image throughout a whole
     * facility for the lifetime of the facility.
     * <p>
     * In C/C++ world, this member variable is an unsigned int32. If it is
     * unset, value must be '0'. Any positive value is considered a valid event
     * number and usually event number are incremented like frame numbers, but
     * for each event instead of each subsequent frame.
     */
    public long eventNumber; // uint32->long sw!
    /**
     * Ratio between pixels and millimetres for x-pixel dimension. If xScale is
     * 3.0, one pixel is 3 mm "wide". If unset or unspecified, value must be set
     * to ~ -1.0. Any negative value except ~ -1.0 is treated as faulty. A value
     * of ~ 0.0 is considered to be illegal.
     * <p>
     * In C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float xScale;
    /**
     * Ratio between pixels and millimetres for x-pixel dimension. If yScale is
     * 2.0, one pixel is 2 mm "tall". If unset or unspecified, value must be set
     * to ~ -1.0. Any negative value except ~ -1.0 is treated as faulty. A value
     * of ~ 0.0 is considered to be illegal.
     * <p>
     * In C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float yScale;
    /**
     * Specifies the rotation balance (compare: account balance, saldo in
     * German) that was applied since image acquire at source level. For example
     * lossless orientation changes can be performed in CoreProvider TINE Server
     * video component to undo perspective changes introduced because of awkward
     * or improper mounting of camera hardware at readout position. This member
     * is always initialized to 0.0 (no rotation performed). Positive values
     * specify clockwise rotation, negative values specify counterclockwise
     * rotation. Regarding value of this field, a general agreement is to keep
     * it in range of -360.0&nbsp;<&nbsp;imageRotation&nbsp;<&nbsp;360.0 .
     * <p>
     * In C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float imageRotation;
    /**
     * First out of three spare floats eserved for future extensions. Current
     * value of this float should be '-1.0' (uninitialized). In
     * <p>
     * C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float fspare1;
    /**
     * Second out of three spare floats reserved for future extensions. Current
     * value of this float should be '-1.0' (uninitialized). In
     * <p>
     * C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float fspare2;
    /**
     * Third out of three spare floats reserved for future extensions. Current
     * value of this float should be '-1.0' (uninitialized). In
     * <p>
     * C/C++ world, this member is a standard 32 bit floating point number.
     */
    public float fspare3;
    /**
     * Contains flags that describe boolean 'states' of an image. Each state is
     * encoded into one bit. Up to now, four bits are already used (LSB 1, MSB
     * 32):
     * <ul>
     * <li>0: if set, image bits are appended in little endian byte order, big
     * endian otherwise
     * <li>1: if set, image acquired from source is possible to restore out of
     * data contained within losslessly. If unset, a lossy change was applied on
     * the way...
     * <li>2: if set, image was flipped along horizontal axis on its way. If bit
     * 2 and 3 are set, horizontal flipping takes precedence over vertical
     * flipping. Any rotation degree, if specified, is performed before any of
     * the flipping.
     * <li>3: if set, image was flipped along vertical axis on its way. If bit 2
     * and 3 are set, horizontal flipping takes precedence over vertical
     * flipping. Any rotation degree, if specified, is performed before any of
     * the flipping.
     * <li>4 to 32: reserved, should be unset
     * </ul>
     * <p>
     * In C/C++ world, this member is a standard 32 bit signed integer.
     */
    public int imageFlags;
    /**
     * First out of three spare integers reserved for future extensions. Current
     * value should be '-1' (uninitialized).
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int ispare1;
    /**
     * Second out of three spare integers reserved for future extensions.
     * Current value should be '-1' (uninitialized).
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int ispare2;
    /**
     * Third out of three spare integers reserved for future extensions. Current
     * value should be '-1' (uninitialized).
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int ispare3;
    /**
     * Contains number of bytes of 'image bits' data that follows the CF_IMAGE
     * header. A value of -1 means that the member is not yet initialized. A
     * value of 0 means that no bytes are appended at all. A positive integer
     * value is equal to the number of bytes that are appended. A value of less
     * than -1 is treated as invalid!
     * <p>
     * In C/C++ world, this member is a signed 32 bit integer.
     */
    public int appendedFrameSize;
  }
  private SourceHeader sourceHeader = new SourceHeader();
  private FrameHeader frameHeader = new FrameHeader();
  private byte[] imageFrame = null;
  private byte[] hHeaderByteArray = new byte[TFormat.getFormatHeaderSize(TFormat.CF_IMAGE)];
  private int sizeInBytes = TFormat.getFormatHeaderSize(TFormat.CF_IMAGE);
  /**
   * Sets the internal image bits buffer reference to the passed byte buffer
   * argument.
   * 
   * @param buffer
   *          reference to image bits buffer
   * @see TFormat
   * @see TFormat#getFormatHeaderSize(short)
   */
  public void setImageFrameBuffer(byte[] buffer)
  {
    imageFrame = buffer;
    sizeInBytes = TFormat.getFormatHeaderSize(TFormat.CF_IMAGE) + imageFrame.length;
  }
  /**
   * Returns a reference to the appended image bits buffer. One should note that
   * this buffer might only be filled to a certain extend, because buffer can be
   * bigger than contained data is.
   * 
   * @return reference to attached frame buffer (byte array)
   * @see FrameHeader#appendedFrameSize
   */
  public byte[] getImageFrameBuffer()
  {
    return imageFrame;
  }
  public void setHeaderBytes(DataInputStream dis) throws IOException
  {
    dis.read(hHeaderByteArray);
  }
  public void setImageBytes(DataInputStream dis) throws IOException
  {
    dis.read(imageFrame);
  }
  public void setImageBytes(DataInputStream dis, int off, int len) throws IOException
  {
    dis.read(imageFrame, off, len);
  }
  /**
   * Returns gross length of the appended image bits buffer in bytes. Value
   * might be significantly bigger than appended frame size, because buffer must
   * be bigger or equal to appended frame size.
   * 
   * @return size of the attached buffer in bytes
   */
  public int getImageFrameSize()
  {
    return imageFrame.length;
  }
  /**
   * Returns the header size by questioning TFormat.
   * 
   * @return size of the header in bytes, if serialized
   * @see TFormat
   * @see TFormat#getFormatHeaderSize(short)
   */
  public int getHeaderSize()
  {
    return TFormat.getFormatHeaderSize(TFormat.CF_IMAGE);
  }
  private void readImageSourceHeader(DataInputStream dis) throws IOException
  {
    sourceHeader.baseTag = Swap.Long(dis.readInt());
    sourceHeader.cameraPortId = UnsignedInt(Swap.Long(dis.readInt()));
    sourceHeader.versionTag = UnsignedInt(Swap.Long(dis.readInt()));
    sourceHeader.totalLength = UnsignedInt(Swap.Long(dis.readInt()));
    sourceHeader.timestampSeconds = Swap.Long(dis.readInt());
    sourceHeader.timestampMicroseconds = Swap.Long(dis.readInt());
    byte[] b = new byte[64];
    dis.read(b, 0, 64);
    sourceHeader.cameraPortName = (new String(b)).trim();
  }
  private void writeImageSourceHeader(DataOutputStream ds) throws IOException
  {
    ds.writeInt(Swap.Long(sourceHeader.baseTag));
    ds.writeInt(Swap.Long(sourceHeader.cameraPortId));
    ds.writeInt(Swap.Long(sourceHeader.versionTag));
    ds.writeInt(Swap.Long(sourceHeader.totalLength));
    ds.writeInt(Swap.Long(sourceHeader.timestampSeconds));
    ds.writeInt(Swap.Long(sourceHeader.timestampMicroseconds));
    byte[] b = new byte[64];
    ds.write(sourceHeader.cameraPortName.getBytes());
    ds.write(b, 0, 64 - sourceHeader.cameraPortName.length());
  }
  private void readImageFrameHeader(DataInputStream dis) throws IOException
  {
    frameHeader.sourceWidth = Swap.Long(dis.readInt());
    frameHeader.sourceHeight = Swap.Long(dis.readInt());
    frameHeader.aoiWidth = Swap.Long(dis.readInt());
    frameHeader.aoiHeight = Swap.Long(dis.readInt());
    frameHeader.xStart = Swap.Long(dis.readInt());
    frameHeader.yStart = Swap.Long(dis.readInt());
    frameHeader.bytesPerPixel = Swap.Long(dis.readInt());
    frameHeader.effectiveBitsPerPixel = Swap.Long(dis.readInt());
    frameHeader.horizontalBinning = Swap.Long(dis.readInt());
    frameHeader.verticalBinning = Swap.Long(dis.readInt());
    frameHeader.sourceFormat = Swap.Long(dis.readInt());
    frameHeader.imageFormat = Swap.Long(dis.readInt());
    frameHeader.frameNumber = UnsignedInt(Swap.Long(dis.readInt()));
    frameHeader.eventNumber = UnsignedInt(Swap.Long(dis.readInt()));
    byte[] b = new byte[4];
    dis.read(b, 0, 4);
    frameHeader.xScale = Swap.Float(b);
    dis.read(b, 0, 4);
    frameHeader.yScale = Swap.Float(b);
    dis.read(b, 0, 4);
    frameHeader.imageRotation = Swap.Float(b);
    dis.read(b, 0, 4);
    frameHeader.fspare1 = Swap.Float(b);
    dis.read(b, 0, 4);
    frameHeader.fspare2 = Swap.Float(b);
    dis.read(b, 0, 4);
    frameHeader.fspare3 = Swap.Float(b);
    frameHeader.imageFlags = Swap.Long(dis.readInt());
    frameHeader.ispare1 = Swap.Long(dis.readInt());
    frameHeader.ispare2 = Swap.Long(dis.readInt());
    frameHeader.ispare3 = Swap.Long(dis.readInt());
    frameHeader.appendedFrameSize = Swap.Long(dis.readInt());
  }
  private void writeImageFrameHeader(DataOutputStream ds) throws IOException
  {
    ds.writeInt(Swap.Long(frameHeader.sourceWidth));
    ds.writeInt(Swap.Long(frameHeader.sourceHeight));
    ds.writeInt(Swap.Long(frameHeader.aoiWidth));
    ds.writeInt(Swap.Long(frameHeader.aoiHeight));
    ds.writeInt(Swap.Long(frameHeader.xStart));
    ds.writeInt(Swap.Long(frameHeader.yStart));
    ds.writeInt(Swap.Long(frameHeader.bytesPerPixel));
    ds.writeInt(Swap.Long(frameHeader.effectiveBitsPerPixel));
    ds.writeInt(Swap.Long(frameHeader.horizontalBinning));
    ds.writeInt(Swap.Long(frameHeader.verticalBinning));
    ds.writeInt(Swap.Long(frameHeader.sourceFormat));
    ds.writeInt(Swap.Long(frameHeader.imageFormat));
    ds.writeInt(Swap.Long(frameHeader.frameNumber));
    ds.writeInt(Swap.Long(frameHeader.eventNumber));
    ds.write(Swap.Float(frameHeader.xScale));
    ds.write(Swap.Float(frameHeader.yScale));
    ds.write(Swap.Float(frameHeader.imageRotation));
    ds.write(Swap.Float(frameHeader.fspare1));
    ds.write(Swap.Float(frameHeader.fspare2));
    ds.write(Swap.Float(frameHeader.fspare3));
    ds.writeInt(Swap.Long(frameHeader.imageFlags));
    ds.writeInt(Swap.Long(frameHeader.ispare1));
    ds.writeInt(Swap.Long(frameHeader.ispare2));
    ds.writeInt(Swap.Long(frameHeader.ispare3));
    ds.writeInt(Swap.Long(frameHeader.appendedFrameSize));
  }
  /**
   * Returns reference to subclass FrameHeader. This method never fails.
   * 
   * @return reference to subclass FrameHeader
   */
  public FrameHeader getFrameHeader()
  {
    return frameHeader;
  }
  /**
   * Returns reference to subclass SourceHeader. This method never fails.
   * 
   * @return reference to subclass SourceHeader
   */
  public SourceHeader getSourceHeader()
  {
    return sourceHeader;
  }
  // implementation of TCompoundDataObject interface further on
  ByteArrayOutputStream dBuffer;
  public byte[] toByteArray()
  {
    try
    {
      int thisSizeInBytes = TFormat.getFormatHeaderSize(TFormat.CF_IMAGE);
      int frameSizeInBytes = imageFrame.length;
      // TODO: if java video server, may need to worry about limiting 
      // the byte array length to the net size of the frame
//      if (frameHeader.appendedFrameSize < imageFrame.length)
//      {
//        frameSizeInBytes = frameHeader.appendedFrameSize;
//      }
      thisSizeInBytes += frameSizeInBytes;

      dBuffer = new ByteArrayOutputStream(thisSizeInBytes);
      DataOutputStream ds = new DataOutputStream(dBuffer);
      writeImageSourceHeader(ds);
      writeImageFrameHeader(ds);
      if (imageFrame != null) ds.write(imageFrame,0,frameSizeInBytes);
      ds.close();
      return dBuffer.toByteArray();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("IMAGE",e.getMessage(),TErrorList.code_failure,e,0);        
    }
    return null;
  }
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream bais = new ByteArrayInputStream(hHeaderByteArray);
      DataInputStream dis = new DataInputStream(bais);
      readImageSourceHeader(dis);
      readImageFrameHeader(dis);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("IMAGE",e.getMessage(),TErrorList.code_failure,e,0);        
    }
  }
  public byte[] getByteArray()
  {
    return hHeaderByteArray;
//    byte[] ba;
//    int ba_len = sizeInBytes;
//    ba = new byte[ba_len];
//    System.arraycopy(hHeaderByteArray,0,ba,0,hHeaderByteArray.length);
//    if (imageFrame != null)
//    {
//      System.arraycopy(imageFrame,0,ba,hHeaderByteArray.length,imageFrame.length);
//    }
//    return ba;
  }
  public void setByteArray(byte[] byteArray)
  {
    if (byteArray == null) return;
    if (hHeaderByteArray == null) 
      hHeaderByteArray = new byte[TFormat.getFormatHeaderSize(TFormat.CF_IMAGE)];
    try
    {
      System.arraycopy(byteArray,0,hHeaderByteArray,0,hHeaderByteArray.length);
      byte[] fb = new byte[byteArray.length-hHeaderByteArray.length];
      System.arraycopy(byteArray,hHeaderByteArray.length,fb,0,byteArray.length-hHeaderByteArray.length);
      setImageFrameBuffer(fb);
    }
    catch (Exception e)
    {
      MsgLog.log("IMAGE.setByteArray", e.getMessage(),TErrorList.code_failure,e,0);
    }
  }
  public short getFormat()
  {
    return TFormat.CF_IMAGE;
  }
  public int getSizeInBytes()
  {
    return sizeInBytes;
  }
  public double getDoubleValue()
  {
    return (double) 0;
  }
  public int getIntegerValue()
  {
    return (int) 0;
  }
}
